/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@gmail.com)
 *  (C) Copyright 2009 ZeXx86 (zexx86@gmail.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifdef __WIN32__
 #include <windows.h>
#endif
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "proto.h"
#include "defaults.h"
#include "ban.h"

extern user_t user_list;
extern ban_t ban_list;

extern long long int sec;

/*************************************************************\
|		  	CREATE NEW BANS			      |
\*************************************************************/

/* create global ban */
ban_t *ban_add (char *data)
{
	ban_t *ctx;

	// alloc and init context
	ctx = (ban_t*) malloc (sizeof (ban_t));

	ctx->nick[0] = '\0';
	ctx->ip[0] = '\0';

	ctx->banlong = atol (data);
	if (ctx->banlong)
		ctx->bantime = sec;
	else
		ctx->bantime = 0;

	ctx->next = &ban_list;
	ctx->prev = ban_list.prev;
	ctx->prev->next = ctx;
	ctx->next->prev = ctx;

	return ctx;
}


unsigned short ban_add_bynick (user_t *u, char *data)
{
	/* ban for this user already exists */
	if (ban_find_bynick (u->nick))
		return 0;

	ban_t *ban;
	ban = ban_add (data);

	strcpy(ban->nick, u->nick);

	return 1;
}

unsigned short ban_add_byip (user_t *u, char *data)
{
	/* ban for this user already exists */
	if (ban_find_byip (u->nick))
		return 0;

	ban_t *ban;
	ban = ban_add (data);

	strcpy(ban->ip, u->ip);

	return 1;
}

/*************************************************************\
|		  	  REMOVE BANS			      |
\*************************************************************/

void ban_del (ban_t *ban)
{
	ban->next->prev = ban->prev;
	ban->prev->next = ban->next;

	free (ban);
}

/*************************************************************\
|		  	FIND OF BANS		     	      |
\*************************************************************/

unsigned long ban_find_bynick (char *nick)
{
	ban_t *ban;

	for (ban = ban_list.next; ban != &ban_list; ban = ban->next) {
		if (!strcmp (ban->nick, nick)) {
			unsigned long left = sec - ban->banlong;
			long l = ban->bantime - left;

			if (l <= 0) {
				ban_del (ban);
				return 0;
			}

			return (ban->bantime - left);
		}
	}

	return 0;
}

unsigned long ban_find_byip (char *ip)
{
	ban_t *ban;

	for (ban = ban_list.next; ban != &ban_list; ban = ban->next) {
		if (!strcmp (ban->ip, ip)) {
			if (!ban->bantime)	// perm ban
				return 1;

			unsigned long left = sec - ban->banlong;
			long l = ban->bantime - left;

			if (l <= 0) {
				ban_del (ban);
				return 0;
			}

			return (ban->bantime - left);
		}
	}

	return 0;
}

unsigned short ban_save ()
{

  FILE *fp;
  ban_t *ban;
  char *filename = strdup (DEFAULT_BANFILE);

  fp = fopen (filename, "w+");
  if (!fp)
    return 0;
	
  for (ban = ban_list.next; ban != &ban_list; ban = ban->next) {
	if (ban->banlong == 0)
		fprintf (fp, "%s\n", ban->ip);
  }

  fclose (fp);

  free (filename);

  return 1;
}

unsigned short ban_load (char *filename)
{
  FILE *fp;
  fp = fopen (filename, "r");

  if (!fp)
    return 0;

  char buffer[4097];

  while (!feof (fp)) {
	buffer[0] = '\0';
	fgets (buffer, 4096, fp);
	short len = strlen (buffer);

	if (!len)
		break;

	buffer[len-1] = '\0';
	if (len > 6) {
		unsigned short i = 0;
		ban_t *ban;
		for (ban = ban_list.next; ban != &ban_list; ban = ban->next) {
			if (!strcmp (ban->ip, buffer))
				i ++;
		}

		if (!i) {
			ban_t *b;
			b = ban_add ("0");
			
			strcpy(b->ip, buffer);
		}
	}
  }

  fclose (fp);

  return 1;

}

/*************************************************************\
|		  INITIALIZATION OF BAN			      |
\*************************************************************/

int init_ban ()
{
	ban_list.next = &ban_list;
	ban_list.prev = &ban_list;

	char *filename = strdup (DEFAULT_BANFILE);
	ban_load (filename);
	free (filename);

	return 1;
}
