/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2009 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2010 ZeXx86 (zexx86@zexos.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */ 


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "chatroom.h"
#include "defaults.h"
#include "commands.h"
#include "config.h"
#include "proto.h"
#include "utils.h"
#include "user.h"

extern chatroom_public_t chatroom_data;
extern chatroom_t chatroom_list;
extern user_t user_list;
extern config_t config;

char roomfilename [] = DEFAULT_CHATROOMFILE;

bool handler_addchatroom (user_t *u, char *data);
int chatroom_rebuild ();

/*************************************************************\
|		   CHATROOM MESSAGE HANDLER		      |
\*************************************************************/

int chatroom_msghandler (user_t *u, char *data, unsigned int len, chatroom_t *room)
{
	/* first we need check, that user have rights for this chatroom */
	if (u->iclass < room->minclass) 
		return 0;

	unsigned int x = 0, s = 0, e = 0;

	while (x <= len) {
		if (data[x] == '$')
			if (data [x+1] == '<')
				s = x;

		if (data[x] == '|')
			e = x;

		x ++;
	}

	if (s == 0 || e == 0)
		return 0;

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
		if (user->iclass >= room->minclass && user != u) {
			char *msg = (char *) calloc (len+u->nick_len+32, sizeof (char));

			if (!msg)
				return 0;
			
			unsigned int l = 12+room->name_len;
			
			memcpy (msg, data, l);
			memcpy (msg+l, room->name, l-12);
			memcpy (msg+l+l-12, " ", 1);
			memcpy (msg+l+l-11, data+s, e+1-s);

			unsigned int m = l+l-10+e-s;
			msg[m] = '\0';

			send_to_user_len (user, msg, m);

			free (msg);
		}
	}

	return 1;
}

/*************************************************************\
|	       CHATROOM SAVE AND LOAD FROM FILES	      |
\*************************************************************/

int chatroom_save ()
{
	FILE *fp;

	fp = fopen (roomfilename, "w+");

	if (!fp)
		return 0;

	chatroom_t *room;
	for (room = chatroom_list.next; room != &chatroom_list; room = room->next)
		fprintf (fp, "%d %s %s\n", room->minclass, room->name, room->desc);
	
	fclose (fp);

	return 1;
}

int chatroom_load ()
{
	FILE *fp = fopen (roomfilename, "r");

	if (!fp)
		return 0;
	
	char buffer[1025], name[80];
	int minclass = 0;

	while (!feof (fp)) {
		buffer [0] = '\0';
		char *ret = fgets (buffer, 1024, fp);

		int len = strlen (buffer);

		if (len < 2)
			break;

		sscanf (buffer, "%d %s", &minclass, name);

		if (minclass < 1)
			minclass = 0;
		if (minclass > 5)
			minclass = 5;
	
		unsigned int x = 2;
		while (buffer [x] != ' ')
			x ++;

		x -= 2;

		char desc[320];
		desc[0] = '\0';
	
		unsigned int d = len-x-5;

		if (!d || strlen (buffer) < (x+3))
			continue;

		memcpy (desc, buffer+x+3, d);
		desc[d] = '\0';

		chatroom_t *ctx;
	
		// alloc and init context
		ctx = (chatroom_t *) malloc (sizeof (chatroom_t));
		
		if (!ctx) {
			fclose (fp);
			return 0;
		}
		
		strcpy (ctx->name, name);
		name[x] = '\0';
		ctx->name_len = x;
		strcpy (ctx->desc, desc);
		ctx->minclass = minclass;
		ctx->myinfo_len = 27+x+d;
	
		ctx->myinfo = (char *) calloc (28+x+d, sizeof(char));
	
		if (!ctx->myinfo) {
			free (ctx);
			fclose (fp);
			return 0;
		}
		
		/* built myinfo */
		memcpy(ctx->myinfo, "$MyINFO $ALL ", 13);
		memcpy(ctx->myinfo+13, name, x);
		memcpy(ctx->myinfo+13+x, " ", 1);
		memcpy(ctx->myinfo+14+x, desc, d);
		memcpy(ctx->myinfo+14+x+d, " $ $ROOM $$$|", 13);
	
		ctx->next = &chatroom_list;
		ctx->prev = chatroom_list.prev;
		ctx->prev->next = ctx;
		ctx->next->prev = ctx;
	}

	fclose (fp);

	chatroom_rebuild ();	/* now we can rebuild chatroom myinfos */

	return 1;
}

/*************************************************************\
|		   CHATROOM FIND FUNCTION		      |
\*************************************************************/

chatroom_t *chatroom_find (char *name)
{
	chatroom_t *room;
	for (room = chatroom_list.next; room != &chatroom_list; room = room->next) {
		if (!strcmp(name, room->name))
			return room;
	}

	return 0;
}

/*************************************************************\
|		   CHATROOM REBUILD BUFFER		      |
\*************************************************************/

int chatroom_rebuild ()
{
	unsigned long iRoomLen = 0;
	unsigned int  iListLen = 0;

	chatroom_t *room;

	/* first we need calculate long of myinfos for allocation of buffer */
	for (room = chatroom_list.next; room != &chatroom_list; room = room->next) {
		iRoomLen += room->myinfo_len;
		iListLen += room->name_len+2;
	}

	/* save length of all chatroom myinfo and list */
	chatroom_data.len = iRoomLen;
	chatroom_data.l = iListLen;

	/* allocation of buffer */
	chatroom_data.myinfos = (char *) calloc (iRoomLen+1, sizeof(char));

	if (!chatroom_data.myinfos)
		return 0;
	
	chatroom_data.lists = (char *) calloc (iListLen+1, sizeof(char));
	
	if (!chatroom_data.lists) {
		free (chatroom_data.myinfos);
		return 0;
	}

	iRoomLen = 0;
	iListLen = 0;
	/* we can now add each myinfo and list of chatroom to allocated buffer */
	for (room = chatroom_list.next; room != &chatroom_list; room = room->next) {
		memcpy (chatroom_data.myinfos+iRoomLen, room->myinfo, room->myinfo_len);
		iRoomLen += room->myinfo_len;

		unsigned int iRoomNameLen = room->name_len;
		memcpy (chatroom_data.lists+iListLen, room->name, iRoomNameLen);
		memcpy (chatroom_data.lists+iListLen+iRoomNameLen, "$$", 2);
		iListLen += iRoomNameLen+2;
	}

	return 1;
}

/*************************************************************\
|		       CHATROOM CREATE		     	      |
\*************************************************************/

bool handler_addchatroom (user_t *u, char *data)
{
	char arg[192];
	unsigned int data_len = strlen (data);

	strcpy (arg, strarg (3, data, data_len));

	if (arg[0] != '0') {
		char use[128];
		sprintf (use, "Use:\n%s <min.class> <name> <description>", arg);

		bot_to_user (u, use);
		return 0;
	}

	int minclass;
	char cmd[64], name[256];
	name[0] = '\0';

	sscanf (data, "%s %d %s", cmd, &minclass, name);

	if (minclass < 0)
		minclass = 0;
	else if (minclass > 5)
		minclass = 5;

	unsigned int l = strlen (name);

	if (l > DEFAULT_MAX_NICKLENGTH && u) {
		bot_to_user (u, "Chatroom name is too long.");
		return 0;
	}

	char desc[320];
	desc[0] = '\0';

	unsigned int d = data_len-l-4-strlen (cmd);
	memcpy (desc, data+strlen (cmd)+l+4, d);
	desc[d] = '\0';

	if (d > 192) {
		bot_to_user (u, "Chatroom description is too long.");
		return 0;
	}

	// alloc and init context
	chatroom_t *ctx = (chatroom_t *) malloc (sizeof (chatroom_t));

	if (!ctx)
		return 0;
	
	unsigned int x = strlen (name);
	
	strcpy (ctx->name, name);
	ctx->name[x] = '\0';
	ctx->name_len = x;
	strcpy (ctx->desc, desc);
	ctx->minclass = minclass;
	ctx->myinfo_len = 27+l+d;

	ctx->myinfo = (char *) calloc (28+l+d, sizeof(char));

	if (!ctx->myinfo) {
		free (ctx);
		return 0;
	}

	/* built myinfo */
	memcpy (ctx->myinfo, "$MyINFO $ALL ", 13);
	memcpy (ctx->myinfo+13, name, l);
	memcpy (ctx->myinfo+13+l, " ", 1);
	memcpy (ctx->myinfo+14+l, desc, d);
	memcpy (ctx->myinfo+14+l+d, " $ $ROOM $$$|", 13);

	ctx->next = &chatroom_list;
	ctx->prev = chatroom_list.prev;
	ctx->prev->next = ctx;
	ctx->next->prev = ctx;

	send_to_all_len (ctx->myinfo, 27+l+d);
	
	char str[256];

	sprintf (str, "Chatroom '%s' for users with min. class (%d) was succefully created.", name, minclass);

	bot_to_user (u, str);

	chatroom_rebuild ();	/* now we can rebuild chatroom myinfos */

	return 0;
}

/*************************************************************\
|		       CHATROOM DELETE		     	      |
\*************************************************************/

bool handler_remchatroom (user_t *u, char *data)
{
	char cmd[32], arg[192];
	sscanf (data, "%s %s", cmd, arg);

	if (strlen (arg) < 2) {
		char use[80];
		sprintf (use, "Use:\n%s <name>", arg);
		bot_to_user (u, use);
		return 0;
	}

	chatroom_t *room;
	room = chatroom_find (arg);

	if (!room) {
		char err[128];

		sprintf (err, "Chatroom with name '%s' does not exist.", arg);

		bot_to_user (u, err);
		return 0;
	}
	
	/* remove a chatroom from context */
 	room->next->prev = room->prev;
 	room->prev->next = room->next; 

	free (room);

	char str[80];

	sprintf (str, "Chatroom '%s' was deleted.", arg);

	bot_to_user (u, str);

	chatroom_rebuild ();	/* now we can rebuild chatroom myinfos */

	return 0;
}

/*************************************************************\
|		   	 CHATROOM LIST		      	      |
\*************************************************************/

bool handler_listchatroom (user_t *u, char *data)
{
	char *str = strdup ("Chatroom List\nName\tDescription\tMin.Class\n\n");

	if (!str)
		return 0;

	unsigned l = strlen (str);

	chatroom_t *room;
	for (room = chatroom_list.next; room != &chatroom_list; room = room->next) {
		char c[80];

		sprintf (c, "%s\t%s\t%d\n", room->name, room->desc, room->minclass);
		unsigned m = strlen (c);

		memcpy (str+l, c, m);

		l += m;
	}

	bot_to_user_len (u, str, l);

	return 0;
}

/*************************************************************\
|		   CHATROOM INITIALIZE			      |
\*************************************************************/

int init_chatroom ()
{
	chatroom_list.next = &chatroom_list;
	chatroom_list.prev = &chatroom_list;

	chatroom_list.minclass = 1;

	chatroom_load ();

	command_register ("chatroomadd", "Create chat room for discusion.", &handler_addchatroom , 5);
	command_register ("chatroomdel", "Delete chat room for discusion.", &handler_remchatroom , 5);
	command_register ("chatroomlist", "List of chat rooms.", &handler_listchatroom , 3);

	return 1;
}
