/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2009 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2010 ZeXx86 (zexx86@zexos.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#ifdef __WIN32__
 #include <windows.h>
#else
 #ifndef __zexos__
  #include <sys/resource.h>
 #endif
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include "user.h"
#include "commands.h"
#include "utils.h"
#include "config.h"
#include "defaults.h"
#include "proto.h"
#include "ban.h"
#include "user_rights.h"

extern command_t command_list;
extern user_t user_list;
extern config_elem_t config_list;
extern config_t config;
extern stats_t	stats;
extern account_t accounts_list;
extern ban_t ban_list;

extern long long int sec;	// uptime

#define PROCSTAT_LEVELS 3
#define PROCSTAT_MEASUREMENTS 60

#ifndef __WIN32__
#ifndef __zexos__
typedef struct procstat {
  struct rusage ps;
  struct timeval tv;
} procstat_t;

procstat_t procstat_boot;
procstat_t procstats[PROCSTAT_LEVELS][PROCSTAT_MEASUREMENTS + 1];
unsigned int current[PROCSTAT_LEVELS];
#endif
#endif

#define TV_TO_MSEC(tv) ((tv.tv_sec*1000)+(tv.tv_usec/1000))

char *filename = (char *) DEFAULT_HUBFILE;


/*************************************************************\
|		      COMMAND's HANDLERS		      |
\*************************************************************/

bool handler_drop (user_t *u, char *data)
{
	char a[16], b[64];
	a[0] = '\0'; b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		int l = strlen (a);

		if (l >= 60)
			return 0;

		char str[80];
		sprintf (str, "Use:\n%s <nick/ip>", a);

		bot_to_user (u, str);

		return 0;
	}

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
		if (user->socket == 0)
			continue;

		if (user->state == PROTO_STATE_ONLINE)	/* socket == 0 is because memleak with virtual user */
		{
			if (!strncmp (b, user->nick, user->nick_len) || !strcmp (b, user->ip)) {
				char str[user->nick_len + 32];
				sprintf (str, "User or ip '%s' was dropped.", b);
			
				bot_to_user (u, str);

				proto_nmdc_state_disconnect (user);

				return 1;
			}
		}
	}

	char str[128];
	sprintf (str, "User or ip '%s' is'nt online.", b);
			
	bot_to_user (u, str);

	return 0;
}

bool handler_help (user_t *u, char *data)
{
	char str[2048];

	sprintf (str, "\n--------------------- [ HELP ] ---------------------");

	command_t *cmd;
	for (cmd = command_list.next; cmd != &command_list; cmd = cmd->next) {
		if (u->iclass >= cmd->iclass) {
			strcat (str, "\n"); strcat (str, cmd->name); strcat (str, " : "); strcat (str, cmd->desc);
		}
	}
	strcat (str, "\n----------------------------------------------------");

	bot_to_user (u, str);

	return 1;
}

bool handler_myinfo (user_t *u, char *data)
{
	char str[512];
	char tag[128];

	unsigned int r = 0, tags = u->myinfo_len, tage = 0;

	while (r != (unsigned) u->myinfo_len) {
		if (u->myinfo[r] == '<')
			tags = r;

		if (u->myinfo[r] == '>')
			tage = r;

		r ++;
	}

	if (tags < tage) {
		memcpy (tag, u->myinfo+tags+1, tage-tags-1);
		tag[tage-tags-1] = '\0';
	}
	
	sprintf (str, "\n---- [ Your INFO ] ----\nNick: %s\nIP Address: %s\nSocket: %d\nBuffer: %lu Bytes\nShare: %s\nTag: %s\nMode: %s\nSlots: %d", u->nick, u->ip, u->socket, u->buf_len, share_tostring (u->share), tag, u->mode == 'A' ? "Active" : "Passive", u->slots);

	if (u->flag & NMDC_SUPPORTS_ZPipe)
		strcat (str,"\nSupports: ZPipe");

	bot_to_user (u, str);

	return 1;
}

bool handler_getinfo (user_t *u, char *data)
{
	unsigned int x = 0;
	char a[16], b[64];
	a[0] = '\0'; b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick>", a);
		bot_to_user (u, use);
		return 0;
	}

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
	    if (user->state != PROTO_STATE_ONLINE || user->socket == 0)	/* socket == 0 is because memleak with virtual user */
		continue;

    	    if (!strcmp (b, user->nick)) {
		char str[512];
		char tag[128];

		unsigned int r = 0, tags = user->myinfo_len, tage = 0;

		while (r != (unsigned) user->myinfo_len) {
			if (user->myinfo[r] == '<')
				tags = r;

			if (user->myinfo[r] == '>')
				tage = r;

			r ++;
		}

		if (tags < tage) {
			memcpy (tag, user->myinfo+tags+1, tage-tags-1);
			tag[tage-tags-1] = '\0';
		}
		
		sprintf (str, "\n---- [ INFO ] ----\nNick: %s\nIP Address: %s\nSocket: %d\nShare: %s\nTag: %s\nMode: %s\nSlots: %d\nSupports: %s", user->nick, user->ip, user->socket, share_tostring (user->share), tag, user->mode == 'A' ? "Active" : "Passive", user->slots, user->flag & NMDC_SUPPORTS_ZPipe ? "ZPipe" : "nothing");

		bot_to_user (u, str);

		x ++;
		break;
	    }
 	}

	/* user is on hub ? */
	if (!x) {
		char err[80];
		sprintf (err, "User '%s' is offline.", b);

		bot_to_user (u, err);
	}

	return 0;
}

bool handler_whoip (user_t *u, char *data)
{
	unsigned int x = 0;
	char a[16], b[64];
	a[0] = '\0'; b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <ip>", a);
		bot_to_user (u, use);
		return 0;
	}

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
		if (user->socket == 0)
			continue;

		if (user->state == PROTO_STATE_ONLINE)	/* socket == 0 is because memleak with virtual user */
		{
			if (!strcmp (b, user->ip)) {
				char str[512];
				sprintf (str, "\n---- [ INFO ] ----\nNick: %s\nIP Address: %s\nSocket: %d\nSupports: %s", user->nick, user->ip, user->socket, user->flag & NMDC_SUPPORTS_ZPipe ? "ZPipe" : "nothing");
			
				bot_to_user (u, str);
		
				x ++;
			}

		} else {
			if (!strcmp (b, user->ip)) {
				/*char state[200];
				
				switch (user->state) {		// FIXME memleak with strlen
					case PROTO_STATE_INIT:
						strncpy (state, "1. Init", 7);
						break;
					case PROTO_STATE_SENDLOCK:
						strncpy (state, "2. Lock", 7);
						break;
					case PROTO_STATE_WAITNICK:
						strncpy (state, "3. WaitNick", 11);
						break;
					case PROTO_STATE_WAITPASS:
						strncpy (state, "4. WaitPass", 11);
						break;
					case PROTO_STATE_HELLO:
						strncpy (state, "5. Hello", 8);
						break;
					case PROTO_STATE_ONLINE:
						strncpy (state, "6. OnLine", 9);
						break;
					case PROTO_STATE_DISCONNECTED:
						strncpy (state, "7. Disconnected", 15);
						break;
					case PROTO_STATE_OUT:
						strncpy (state, "8. Out", 6);
						break;
				}
				state[16] = '\0';*/
				

				char str[60];
				sprintf (str, "\nState: %d\nIP Address: %s\nSocket: %d", user->state, user->ip, user->socket);
			
				bot_to_user (u, str);
		
				x ++;
			}
		}
 	}

	/* user is on hub ? */
	if (!x) {
		char err[80];
		sprintf (err, "User with IP address '%s' is offline.", b);

		bot_to_user (u, err);
	}

	return 0;
}


bool handler_regnew (user_t *u, char *data)
{
	int c = 1;
	char a[64], b[64], d[64]; 
	b[0] = '\0'; d[0] = '\0';
	
	sscanf (data, "%s %s %d %s", a, b, &c, d);

	if (c < 1 || strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick> <class> [<passwd>]", a);
		bot_to_user (u, use);
		return 0;
	}

	if (strlen (d) < 1)
		strcpy (d, "none");	// default password

	char str[80];
	sprintf (str, "\nAdded a user '%s' with class (%d).", b, c);

	char err[80];
	sprintf (err, "User '%s' already exist.", b);

	unsigned short i = 0;
	user_t *t;

	for (t = user_list.next; t != &user_list; t = t->next) {
    	    if (!strcmp (b, t->nick)) {

		int r = account_add (t, d, c);

		char reg[150];
		reg [0] = '\0';
		sprintf (reg, "You have been registered by %s with class '%d'. Please set your password with command '!passwd'. Default password is 'none'.", u->nick, c);

		bot_to_user (t, reg);

		if (r)	
			bot_to_user (u, str);
		else
			bot_to_user (u, err);

		i ++;
	    }
	}

	if (!i) {
		sprintf (str, "User '%s' not logged.", b);
		bot_to_user (u, str);
		return 0;
	}

	return 1;
}

bool handler_regdel (user_t *u, char *data)
{
	char a[64], b[64]; 
	b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick>", a);
		bot_to_user (u, use);
		return 0;
	}
	
	char str[128];
	sprintf (str, "User account '%s' was deleted.", b);
	
	char err[128];
	sprintf (err, "User account '%s' does not exist.", b);

	int r = account_del (u, b);

	if (r)
		bot_to_user (u, str);
	else
		bot_to_user (u, err);

	return 1;
}

bool handler_regclass (user_t *u, char *data)
{
	int c = 1;
	char a[64], b[64];
	b[0] = '\0';
	
	sscanf (data, "%s %s %d", a, b, &c);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick> <class>", a);
		bot_to_user (u, use);
		return 0;
	}

	if (c < 1 || c > 5) {
		bot_to_user (u, "Invalid class. You can choose between 1 to 5.");
		return 0;
	}
	
	char x[6]; sprintf (x, "(%d)", c);
	
	int cl = account_find (b);

	if (u->iclass < c || u->iclass < cl) {
		char err0[80];
		sprintf (err0, "You have'nt rights to change user with class (%d) on class (%d).", cl, c);

		bot_to_user (u, err0);

		return 0;
	}

	char str[128];
	sprintf (str, "Changed a class of user '%s'.", b);
	
	char err1[128];
	sprintf (err1, "User account '%s' does not exist.", b);

	char err2[128];
	sprintf (err2, "User class '%s' does not valid.", x);

	int r = account_class (b, c);

	if (r) {
		bot_to_user (u, str);
		
		user_t *tgt = user_findbynick (b);
		if (tgt) {
			char info[80];
			sprintf (info, "Your class was changed to %d by user %s.", c, u->nick);

			bot_to_user (tgt, info);
		}
	} else if (r == 0)
		bot_to_user (u, err1);
	else if (r == -1)
		bot_to_user (u, err2);

	return 1;
}


bool handler_reglist (user_t *u, char *data)
{
	unsigned int i = 0;
	unsigned str_len = 45;
	
	account_t *a;
	for (a = accounts_list.next; a != &accounts_list; a = a->next) {
		i ++;

		char line[96];
		sprintf (line, "\n%d.\t%s\t\t%d", i, a->nick, a->iclass);

		str_len += strlen (line);
	}
	
	if (!i) {
		bot_to_user (u, "No registered user found.");
		return 0;
	}
	
	char *str = (char *) malloc (str_len + 1);
	
	if (!str)
		return 0;

	i = 0;
	str_len = 0;
	
	for (a = accounts_list.next; a != &accounts_list; a = a->next) {
		i ++;

		char line[96];
		sprintf (line, "\n%d.\t%s\t\t%d", i, a->nick, a->iclass);
		
		unsigned line_len = strlen (line);
		
		memcpy (str + str_len, line, line_len);
		
		str_len += line_len;
	}
	
	str[str_len] = '\0';

	bot_to_user_len (u, str, str_len);

	free (str);
	
	return 0;
}

bool handler_regme (user_t *u, char *data)
{
	unsigned int i = account_add (u, (char *) "none", 1);

	if (!i) {
		bot_to_user (u, "You are already registered.");
		return 0;
	}

	bot_to_user (u, "You have been registered with class '1'. Please set your password with command '!passwd'. Default password is 'none'.");

	return 0;
}

bool handler_passwd (user_t *u, char *data)
{
	char a[64], b[64]; 
	b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <passwd>", a);
		bot_to_user (u, use);
		return 0;
	}

	int i = account_passwd (u, b);

	if (!i) {
		bot_to_user (u, "You are'nt registered user.");

		return 0;
	}

	char str[128];
	str[0] = '\0';
		
	sprintf (str, "Your password was set to '%s'.", b);

	bot_to_user (u, str);

	return 0;
}

bool handler_set (user_t *u, char *data)
{
	unsigned int data_len = strlen (data);

	char a[64], b[64], c[512]; 
	a[0] = '\0'; b[0] = '\0'; c[0] = '\0';
	
	sscanf (data, "%s %s %s", a, b, c);

	unsigned int a_len    = strlen (a);
	unsigned int b_len    = strlen (b);
	unsigned int c_len    = strlen (c);

	if (b_len < 1 || c_len < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <variable> <value>", a);
		bot_to_user (u, use);
		return 0;
	}

	config_elem_t *cfg;

	int i = 0;

	for (cfg = config_list.next; cfg != &config_list; cfg = cfg->next) {
		if (!strcmp (b, cfg->name)) {
			char str[512];
			sprintf (str, "Changed:\n Old '%s' = '%s'", b, cfg->value);

			if (cfg->value && *cfg->value)
				free (cfg->value);

			cfg->value = (char *) calloc (data_len-a_len-b_len + 1, sizeof (char));

			if (!cfg->value)
				break;
			
			cfg->value[0] = '\0';
			memcpy (cfg->value, c, c_len);

			cfg->value[c_len] = '\0';
			cfg->value[c_len+1] = '\0';

			if (!strcmp (b, "bot.name")) {
  				config.botname[0]='\0';strcpy (config.botname, cfg->value); }	// parse bot name
			if (!strcmp (b, "hubname")) {	// parse whole name
				memcpy (cfg->value, data+b_len+a_len+2, data_len-b_len+a_len-2); cfg->value[data_len-b_len+a_len-2]='\0';}
			if (!strcmp (b, "hubdesc")) {	// parse whole description
				memcpy (cfg->value, data+b_len+a_len+2, data_len-b_len+a_len-2); cfg->value[data_len-b_len+a_len-2]='\0';}
			if (!strcmp (b, "hublist.hosts")) {	// parse all adress
				memcpy (cfg->value, data+b_len+a_len+2, data_len-b_len+a_len-2); cfg->value[data_len-b_len+a_len-2]='\0';}
			if (!strcmp (b, "redirect")) {	// parse all adress
				memcpy (cfg->value, data+b_len+a_len+2, data_len-b_len+a_len-2); cfg->value[data_len-b_len+a_len-2]='\0';}

			if (!strncmp (b, "minshare", 8)) {	// parse minshare limits
				if (c_len >= 64)
					continue;

				unsigned long long minshare = atoll (cfg->value);
				char limit[64];
				char type;
				short sqrtx = 0;
				type = data[data_len-1];	

				switch (type) {
					case 'b':
						sqrtx = 0;
						break;
					case 'k':
						sqrtx = 1;
						break;
					case 'm':
						sqrtx = 2;
						break;
					case 'g':
						sqrtx = 3;
						break;
				}

				while (sqrtx) {
					minshare = minshare * 1024;
					sqrtx --;
				}

				sprintf (limit, "%llu", minshare);

				unsigned int len = strlen (limit);
				memcpy (cfg->value, limit, len);
				cfg->value[len] = '\0';
			}
/*#ifdef SELECT
			if (!strcmp (b, "userlimit") && atoi (c) > 1024)
				bot_to_user (u, "Warning -> Hub may be unstable with socket Select function, for break more users count compile hubsoft with another func. (comment line with #define SELECT in file defaults.h)");
#endif*/

#ifndef __WIN32
			if (!strcmp (b, "userlimit") && atoi (c) > 1024)
				bot_to_user (u, "Warning -> Dont forgot on system value of max. connection count, which is default setted to 1024. Please set this value to higher by console/terminal and restart a hub. Example of change this value: ulimit -n 10240");
#endif

			/* send to op info about changed variable */
			strcat (str, "'\nNew '"); strcat (str, b); strcat (str, "' = '"); strcat (str, cfg->value); strcat (str, "'");

			bot_to_user (u, str);

			cache_rebuild ();

			i ++;
		}
	}

	if (i == 0) {
		char err[128];
		sprintf (err, "Variable '%s' not found.", b);

		bot_to_user (u, err);
	}

	return 1;
}

bool handler_getconfig (user_t *u, char *data)
{
	char str[4096];

	strcpy (str, "\n--------------------- [ CONFIG ] ---------------------");

	config_elem_t *ctx;
	for (ctx = config_list.next; ctx != &config_list; ctx = ctx->next) {
		strcat (str, "\n"); strcat (str, ctx->name); strcat (str, " "); strcat (str, ctx->value);

	}
	
	strcat (str, "\n------------------------------------------------------");

	bot_to_user (u, str);

	return 1;
}

#ifndef __WIN32__
#ifndef __zexos__
/************************* CPU FUNCTIONS ******************************************/
char cpuinfo[512];
short cpucount = 1;

void cpu_parse ()
{
	FILE *fp;
	char buf[1024];
	
	fp = fopen ("/proc/cpuinfo", "r");

	if (!fp) {
		strcpy (cpuinfo, "CPU Unknown");
		return;
	}
	
	while (!feof (fp)) {
		char *ret = fgets (buf, 1024, fp);

		if (!strncasecmp (buf, "model name", 10) || !strncmp (buf, "Processor", 9) || !strncmp (buf, "cpu model", 9)) {
			sprintfa (cpuinfo, "\nCPU %.1d:\n Model: %.*s\n", cpucount ++, strlen (buf) - 13, buf + 12);
		} else if (!strncasecmp (buf, "cpu MHz", 7)) {
			sprintfa (cpuinfo, " Clock: %.*s MHz\n", strlen (buf) - 12, buf + 11);
		} else if (!strncasecmp (buf, "bogomips", 8)) {
			sprintfa (cpuinfo, " BogoMIPS: %.*s\n", strlen (buf) - 12, buf + 11);
		}
	}

	fclose (fp);
}

void cpu_init ()
{
	cpu_parse ();
	
	memset (&procstats, 0, sizeof (procstats));
	memset (&current, 0, sizeof (current));
	memset (&procstat_boot, 0, sizeof (procstat_t));
	getrusage (RUSAGE_SELF, &(procstats[0][0].ps));
	gettimeofday (&(procstats[0][0].tv), NULL);
	procstat_boot = procstats[0][0];
	procstats[2][0] = procstats[1][0] = procstats[0][0];
}

unsigned int cpu_measure ()
{
	int i;

	current[0]++;
	getrusage (RUSAGE_SELF, &(procstats[0][current[0]].ps));
	gettimeofday (&(procstats[0][current[0]].tv), NULL);
	
	for (i = 0; (current[i] == (PROCSTAT_MEASUREMENTS)) && (i < PROCSTAT_LEVELS); i++) {
		if (i < (PROCSTAT_LEVELS - 1)) {
			current[i + 1]++;
			procstats[i + 1][current[i + 1]] = procstats[i][PROCSTAT_MEASUREMENTS];
		}
		
		current[i] = 0;
		procstats[i][0] = procstats[i][PROCSTAT_MEASUREMENTS];
	}

	return 0;
}

float cpu_calc (procstat_t * old, procstat_t * nef)
{
	long used, real;

	/* we aren't up that long yet */
	if (!old->tv.tv_sec)
	  return 0;

	used = TV_TO_MSEC (nef->ps.ru_utime);
	used += TV_TO_MSEC (nef->ps.ru_stime);
	used -= TV_TO_MSEC (old->ps.ru_utime);
	used -= TV_TO_MSEC (old->ps.ru_stime);

	real = TV_TO_MSEC (nef->tv);
	real -= TV_TO_MSEC (old->tv);

	return (100.0 * (float) used) / (float) real;
}

float cpu_calculate (int seconds)
{
	int level = seconds % PROCSTAT_MEASUREMENTS;

	if (procstats[level][(current[level] - 1) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		return cpu_calc (&procstats[level][(current[level] - 1) % PROCSTAT_MEASUREMENTS], &procstats[0][current[0]]);
	else
		return 0.0;
}

bool handler_cpuinfo (user_t *u, char *data)
{
	char str[512];
	char hour[64], minute[64], seconds[64], boot[128];
	
	hour[0] = '\0'; minute[0] = '\0'; seconds[0] = '\0'; boot[0] = '\0';	// put NULL character into char
	
	if (procstats[2][(current[2] - 1) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		sprintf (hour, " last hour %2.2f%%\n",
			cpu_calc (&procstats[2][(current[2] - 1) % PROCSTAT_MEASUREMENTS],
					&procstats[0][current[0]]));
	if (procstats[1][(current[1] - 1) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		sprintf (minute, " last minute %2.2f%%\n",
			cpu_calc (&procstats[1][(current[1] - 1) % PROCSTAT_MEASUREMENTS],
					&procstats[0][current[0]]));
	if (procstats[0][(current[0] - 5) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		sprintf (seconds, " last 5 seconds %2.2f%%\n\n",
			cpu_calc (&procstats[0][(current[0] - 5) % PROCSTAT_MEASUREMENTS],
					&procstats[0][current[0]]));
	
	sprintf (boot, "Since boot %2.2f%%\n", cpu_calc (&procstat_boot, &procstats[0][current[0]]));
	
	sprintf (str, "\nCPU Usage:\n%s%s%s%s\nCPU Statistics:%s", hour, minute, seconds, boot, cpuinfo);
	
	bot_to_user (u, str);

  return 0;
}

bool handler_linuxinfo (user_t *u, char *data)
{
	char *buffer;
	
	FILE *fp = fopen ("/proc/version", "r");

	if (!fp)
		return 0;
	
	buffer = (char *) calloc (1024, sizeof (char));
	
	if (!buffer)
		return 0;

	char *ret = fgets (buffer, 1024, fp);

	fclose (fp);

	bot_to_user (u, buffer);

	free (buffer);

	return 0;
}
#endif
#else	// win32

unsigned int cpucount = 0;

#define FT_TO_MSEC(time)	( (((unsigned long long)time.dwLowDateTime) + (((unsigned long long)time.dwHighDateTime) << 32)) / 10000LL)

void ft_add (FILETIME * target, FILETIME * t1, FILETIME * t2)
{
	unsigned long long tmp;

	target->dwHighDateTime = t1->dwHighDateTime + t2->dwHighDateTime;
	tmp = t1->dwLowDateTime + t2->dwLowDateTime;
	
	if (tmp > 0xffffffff)
		target->dwHighDateTime ++;
	
	target->dwLowDateTime = tmp & 0xffffffff;
}

typedef struct procstat {
	FILETIME ps;
	struct timeval tv;
} procstat_t;

procstat_t procstat_boot;
procstat_t procstats[PROCSTAT_LEVELS][PROCSTAT_MEASUREMENTS + 1];
unsigned int current[PROCSTAT_LEVELS];

HANDLE myProcess;

#define CPU_REG_KEY             	HKEY_LOCAL_MACHINE
#define CPU_REG_SUBKEY  "		HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\%d"

#define CPU_SPEED                       "~MHz"
#define CPU_IDENTIFIER          	"Identifier"
#define CPU_NAME                        "ProcessorNameString"
#define CPU_VENDOR                      "VendorIdentifier"

float cpu_calc (procstat_t * old, procstat_t * snew);

bool handler_cpuinfo (user_t *u, char *data)
{
	HKEY hKey;
	LONG ret;
	DWORD dwSize, dwType;
	unsigned int i;
	char buf[256], *c, cpuinfo[1024];
	SYSTEM_INFO sysinfo;
	
	GetSystemInfo (&sysinfo);
	
	cpucount = sysinfo.dwNumberOfProcessors;

	for (i = 0; i < cpucount; i++) {
		//sprintf (cpuinfo, "CPU %d: ", i+1);
		sprintf (buf, CPU_REG_SUBKEY, i);
		
		ret = RegOpenKeyEx (CPU_REG_KEY, buf, 0, KEY_READ, &hKey);
		
		if (ret != ERROR_SUCCESS)
			continue;
		
		dwSize = 0;
		buf[0] = 0;
		ret = RegQueryValueEx (hKey, CPU_NAME, NULL, &dwType, NULL, &dwSize);
		ret = RegQueryValueEx (hKey, CPU_NAME, NULL, &dwType, (BYTE *) buf, &dwSize);
		if (ret != ERROR_SUCCESS) {
			RegCloseKey (hKey);

			return -1;
		}
		
		c = buf;
		while (*c == ' ')
			c++;
		
		//sprintfa (cpuinfo, "%s\n", c);
 	}


	char str[512];
	char hour[64], minute[64], seconds[64], boot[128];
	
	hour[0] = '\0'; minute[0] = '\0'; seconds[0] = '\0'; boot[0] = '\0';	// put NULL character into char

	if (procstats[2][(current[2] - 1) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		sprintf (hour, " last hour %2.2f%%\n",
			cpu_calc (&procstats[2][(current[2] - 1) % PROCSTAT_MEASUREMENTS],
					&procstats[0][current[0]]));
	if (procstats[1][(current[1] - 1) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		sprintf (minute, " last minute %2.2f%%\n",
			cpu_calc (&procstats[1][(current[1] - 1) % PROCSTAT_MEASUREMENTS],
					&procstats[0][current[0]]));
	if (procstats[0][(current[0] - 5) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		sprintf (seconds, " last 5 seconds %2.2f%%\n\n",
			cpu_calc (&procstats[0][(current[0] - 5) % PROCSTAT_MEASUREMENTS],
					&procstats[0][current[0]]));
	
	sprintf (boot, "Since boot %2.2f%%\n", cpu_calc (&procstat_boot, &procstats[0][current[0]]));


	/*time_t tnow = now.tv_sec;
	string str;
	
	bf_printf (buf, _("\nCpu Statistics at %s"), ctime (&tnow));
	
	if (procstats[2][(current[2] - 1) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		bf_printf (buf, _(" last hour %2.2f%%\n"),
		cpu_calc (&procstats[2][(current[2] - 1) % PROCSTAT_MEASUREMENTS],
				&procstats[0][current[0]]));
	if (procstats[1][(current[1] - 1) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		bf_printf (buf, _(" last minute %2.2f%%\n"),
		cpu_calc (&procstats[1][(current[1] - 1) % PROCSTAT_MEASUREMENTS],
				&procstats[0][current[0]]));
	if (procstats[0][(current[0] - 5) % PROCSTAT_MEASUREMENTS].tv.tv_sec > 0)
		bf_printf (buf, _(" last 5 seconds %2.2f%%\n\n"),
		cpu_calc (&procstats[0][(current[0] - 5) % PROCSTAT_MEASUREMENTS],
				&procstats[0][current[0]]));
	
	bf_printf (buf, _("Since boot %2.2f%%\n"), cpu_calc (&procstat_boot, &procstats[0][current[0]]));*/
	
	//  if (cpuinfo)
	//   bf_printf (buf, "\n%.*s", bf_used (cpuinfo), cpuinfo->s);
	
	sprintf (str, "\n----------- [ CPU info ] -----------\n\n%s%s%s%s", hour, minute, seconds, boot);
	//str += "\n"; str += cpuinfo;
	
	bot_to_user (u, str);

	return 0;

}

void cpu_init ()
{
	FILETIME d1, d2, KernelTime, UserTime;
	DWORD pid = GetCurrentProcessId ();

	myProcess = OpenProcess (PROCESS_QUERY_INFORMATION, FALSE, pid);

	procstats[0][0].tv.tv_sec = GetTickCount ();

	GetProcessTimes (myProcess, &d1, &d2, &KernelTime, &UserTime);
	ft_add (&(procstats[0][0].ps), &KernelTime, &UserTime);

	procstat_boot = procstats[0][0];
}

unsigned int cpu_measure ()
{
	int i;
	FILETIME d1, d2, KernelTime, UserTime;

	GetProcessTimes (myProcess, &d1, &d2, &KernelTime, &UserTime);

	current[0]++;
	procstats[0][current[0]].tv.tv_sec = GetTickCount ();

	ft_add (&(procstats[0][current[0]].ps), &KernelTime, &UserTime);

	for (i = 0; (current[i] == (PROCSTAT_MEASUREMENTS)) && (i < PROCSTAT_LEVELS); i++) {
		if (i < (PROCSTAT_LEVELS - 1)) {
			current[i + 1]++;
			procstats[i + 1][current[i + 1]] = procstats[i][PROCSTAT_MEASUREMENTS];
		}
		
		current[i] = 0;
		procstats[i][0] = procstats[i][PROCSTAT_MEASUREMENTS];
	}

}

float cpu_calc (procstat_t * old, procstat_t * snew)
{
	long long used, real;

	/* we aren't up that long yet */
	if (!old->tv.tv_sec)
		return 0;

	used = FT_TO_MSEC (snew->ps);
	used -= FT_TO_MSEC (old->ps);

	real = TV_TO_MSEC (snew->tv);
	real -= TV_TO_MSEC (old->tv);

	return ((100.0 * (float) used) / (float) real) / (float) cpucount;
}

float cpu_calculate (int seconds)
{
	int level = seconds % PROCSTAT_MEASUREMENTS;

	if (procstats[level][(current[level] - 1) % PROCSTAT_MEASUREMENTS].ps.dwHighDateTime > 0)
		return cpu_calc (&procstats[level][(current[level] - 1) % PROCSTAT_MEASUREMENTS], &procstats[0][current[0]]);
	else
		return 0.0;
};

#endif

bool handler_statbw (user_t *u, char *data)
{
	//char str[128];
// 	sprintf (str, "Bandwidth usage\nUpload: %s/sec\nDownload: %s/sec", 
// 		 share_tostring (stats.traffic_rx), share_tostring (stats.traffic_tx));

	char str[256];
	sprintf (str, "Bandwidth usage\nUpload: %s/sec\nDownload: %s/sec", share_tostring (stats.traffic_rx), share_tostring (stats.traffic_tx));

	bot_to_user (u, str);

	return 0;
}

extern int cont;
bool handler_shutdown (user_t *u, char *data)
{
	cont = 0;	// break loop

	user_t *user;

	/* close all connections */
	for (user = user_list.next; user != &user_list; user = user->next)	
		proto_nmdc_state_disconnect (user);

	return 0;
}

bool handler_cfgload (user_t *u, char *data)
{
	config_load (filename);

	bot_to_user (u, "Config loaded.");

	return 0;
}

bool handler_cfgsave (user_t *u, char *data)
{
	config_save (filename);

	bot_to_user (u, "Config saved.");

	return 0;
}

bool handler_rebuild (user_t *u, char *data)
{
	cache_rebuild ();

	unsigned int n = 0;

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
		if (user->online && user->state == PROTO_STATE_ONLINE && user->socket)
    	    		n ++;
	}

	stats.currusers = n;

	bot_to_user (u, "OK");

	return 0;
}

bool handler_uptime (user_t *u, char *data)
{
	char str[128];
	sprintf (str, "UpTime: %s.", time_tostring (sec));

	bot_to_user (u, str);

	return 0;
}

bool handler_kick (user_t *u, char *data)
{
	int x = 0;
	char a[16], b[64], c[512]; 
	a[0] = '\0'; b[0] = '\0'; c[0] = '\0';
	
	sscanf (data, "%s %s %s", a, b, c);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick> [reason]", a);
		bot_to_user (u, use);
		return 0;
	}

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
    	    if (!strcmp (b, user->nick)) {
		x ++;
		break;
	    }
 	}

	/* user is on hub ? */
	if (!x) {
		char err[128];
		sprintf (err, "User with ip address '%s' is offline.", b);

		bot_to_user (u, err);
		
		return 0;
	}

	/* check rights */
	int acc = account_find (b);

	if (acc > u->iclass) {
		char c[2];
		sprintf (c, "%d", acc);
		
		char err[128];
		sprintf (err, "You have'nt rights to kick user with class '%s'.", c);
	
		bot_to_user (u, err);

		return 0;
	}

	/* drop user from hub */
	char str[128];
	sprintf (str, "You have been kicked by '%s', because: %s.", u->nick, strlen (c) < 1 ? DEFAULT_KICK_MESSAGE : c);

	char t[6];
	t[0] = '\0';
	sprintf (t, "%d", DEFAULT_KICKTIME);

	ban_add_bynick (user, t);

	bot_to_user (user, str);

	proto_nmdc_user_redirect (user);

	stats.num_kick ++;

	return 0;
}

bool handler_bannick (user_t *u, char *data)
{
	int x = 0;
	char a[16], b[64], c[512]; 
	a[0] = '\0'; b[0] = '\0'; c[0] = '\0';
	
	sscanf (data, "%s %s %s", a, b, c);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick> [timeinseconds]", a);
		bot_to_user (u, use);
		return 0;
	}

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
    	    if (!strcmp (b, user->nick)) {
		x ++;
		break;
	    }
 	}

	/* user is on hub ? */
	if (!x) {
		char err[128];
		sprintf (err, "User with ip address '%s' is offline.", b);

		bot_to_user (u, err);
	}

	/* check rights */
	int acc = account_find (b);

	if (acc > u->iclass) {
		char c[2];
		sprintf (c, "%d", acc);
		char err[128];
		sprintf (err, "You have'nt rights to ban user with class '%s'.", c);

		bot_to_user (u, err);

		return 0;
	}

	/* ban for this user already exists */
	if (ban_find_bynick (b)) {
		bot_to_user (u, "Failed. Ban exists");
		return 0;
	}

	if (strlen (c) < 1)
	       sprintf (c, "%d", DEFAULT_BANTIME);

	/* drop user from hub */
	char str[128];
	sprintf (str, "You have been banned by '%s', time: %s.", u->nick, c);

	bot_to_user (user, str);

	ban_t *ban;
	ban = ban_add (c);

	strcpy(ban->nick, b);

	ban_add_bynick (user, c);

	bot_to_user (u, "OK");

	proto_nmdc_user_redirect (user);

	stats.num_ban ++;

	return 0;
}

bool handler_banip (user_t *u, char *data)
{
	int x = 0;
	char a[16], b[64], c[512]; 
	a[0] = '\0'; b[0] = '\0'; c[0] = '\0';
	
	sscanf (data, "%s %s %s", a, b, c);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <ip> [timeinseconds]", a);
		bot_to_user (u, use);
		return 0;
	}

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
    	    if (!strcmp (b, user->ip)) {
		x ++;
		break;
	    }
 	}

	/* user is on hub ? */
	if (!x) {
		char err[128];
		sprintf (err, "User with ip address '%s' is offline.", b);

		bot_to_user (u, err);
		
		return 0;
	}

	/* check rights */
	/*int acc = account_find (b);

	if (acc > u->iclass) {
		char c[2];
		sprintf (c, "%d", acc);
		string err;
		       err += "You have'nt rights to ban user with class '";
		       err += c;
		       err += "'.";

		bot_to_user (u, err);

		return 0;
	}*/

	/* ban for this user already exists */
	if (ban_find_byip (b)) {
		bot_to_user (u, "Failed. Ban exists");
		return 0;
	}

	if (strlen (c) < 1)
	       sprintf (c, "%d", DEFAULT_BANTIME);

	/* drop user from hub */
	char str[128];
	sprintf (str, "You have been banned by '%s', time: %s.", u->nick, c);

	bot_to_user (user, str);

	ban_t *ban;
	ban = ban_add (c);

	strcpy (ban->ip, b);

	bot_to_user (u, "OK");

	proto_nmdc_user_redirect (user);

	stats.num_ban ++;

	return 0;
}

bool handler_banperm (user_t *u, char *data)
{
	int x = 0;
	char a[16], b[64];
	a[0] = '\0'; b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <ip>", a);
		bot_to_user (u, use);
		return 0;
	}

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
    	    if (!strcmp (b, user->ip)) {
		x ++;
		break;
	    }
 	}

	/* user is on hub ? */
	if (!x) {
		char err[128];
		sprintf (err, "User with ip address '%s' is offline.", b);

		bot_to_user (u, err);
	}

	/* check rights */
	/*int acc = account_find (b);

	if (acc > u->iclass) {
		char c[2];
		sprintf (c, "%d", acc);
		string err;
		       err += "You have'nt rights to ban user with class '";
		       err += c;
		       err += "'.";

		bot_to_user (u, err);

		return 0;
	}*/

	/* ban for this user already exists */
	if (ban_find_byip (b)) {
		bot_to_user (u, "Failed. Ban exists");
		return 0;
	}

	/* drop user from hub */
	char str[128];
	sprintf (str, "You have been banned by '%s' permanently.", u->nick);

	bot_to_user (user, str);

	ban_t *ban;
	ban = ban_add ((char *) "0");

	strcpy (ban->ip, b);

	bot_to_user (u, "OK");

	proto_nmdc_user_redirect (user);

	stats.num_ban ++;

	return 0;
}

bool handler_bandel (user_t *u, char *data)
{
	char a[16], b[64];
	a[0] = '\0'; b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick/ip>", a);
		return 0;
	}

	ban_t *ban;
	unsigned short i = 0;

	for (ban = ban_list.next; ban != &ban_list; ban = ban->next) {
		if (!strcmp (ban->nick, b)) {
			ban_del (ban);
			i ++;
			break;
		}
		if (!strcmp (ban->ip, b)) {
			ban_del (ban);
			i ++;
			break;
		}
	}

	if (i)
		bot_to_user (u, "OK");
	else {
		char err[128];
		sprintf (err, "User with nick or ip address '%s' not found.", b);

		bot_to_user (u, err);
	}

	return 0;
}

bool handler_banlist (user_t *u, char *data)
{
	unsigned int i = 0;

	/*string str ("\n------- [ BAN-LIST ] -------\nId\tNick / IP Address\tTime [s]\n");

	ban_t *ban;

	for (ban = ban_list.next; ban != &ban_list; ban = ban->next) {
		char line[96];
		line [0] = '\0';

		if (strlen (ban->nick) > 1) {
			i ++;
			sprintf (line, "\n%d.\t%s\t%lu", i, ban->nick, ban->banlong);
		} else if (strlen (ban->ip) > 1) {
			i ++;
			sprintf (line, "\n%d.\t%s\t%lu", i, ban->ip, ban->banlong);
		}

		str += line;
	}

	if (!i) {
		bot_to_user (u, "Banned user not found.");
		return 0;
	}

	bot_to_user (u, str);*/

	return 0;
}

bool handler_hubinfo (user_t *u, char *data)
{
	char usr[12];
	sprintf (usr, "%d / ", stats.currusers);	

	char str[512];
	
	strcpy (str, "\n--------------------- [ HUBINFO ] ---------------------");
	strcat (str, "\nHub name: "); strcat (str, config_get ((char *) "hubname"));
	strcat (str, "\nHub description: "); strcat (str, config_get ((char *) "hubdesc"));
	strcat (str, "\nOwner: "); strcat (str, config_get ((char *) "hubowner"));
	strcat (str, "\nSearch interval: "); strcat (str, config_get ((char *) "search.interval.guest"));
	strcat (str, "\nMin. share: "); strcat (str, config_get ((char *) "minshare.guest"));
	strcat (str, "\nMin. slots: "); strcat (str, config_get ((char *) "minslots.guest"));
	strcat (str, "\nCurr. / Max. users: "); strcat (str, usr); strcat (str, config_get ((char *) "userlimit"));
	strcat (str, "\nRedirect hosts: "); strcat (str, config_get ((char *) "redirect"));

#ifdef SELECT
	strcat (str, "\n\n" HUBSOFT " is based on Select function.");
#else
# ifndef POLL
	strcat (str, "\n\n" HUBSOFT " is based on Send/Recv function.");
# else
	strcat (str, "\n\n" HUBSOFT " is based on Poll function.");
# endif
#endif
	strcat (str, "\nPowered by " PLATFORM ".");

	bot_to_user (u, str);

	return 0;
}

bool handler_redirect (user_t *u, char *data)
{
	if(!class_right_check (u, CAP_REDIRECT)) {
		bot_to_user (u, "You have'nt right for redirecting.");
		return 0;
	}

	char who[DEFAULT_MAX_NICKLENGTH];
	who[0] = '\0';
		
	char where[DEFAULT_MAX_ADDRESSLENGTH];
	where[0] = '\0';

	char cmd[64];
	cmd[0] = '\0';
		
		
	sscanf (data, "%s %s %s", cmd, who, where);

	int x = strlen (who);
	int y = strlen (where);

	if (x < 1 || y < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick> <address[:port]>", cmd);
		bot_to_user (u, use);
		return 0;
	}

		
	user_t *user;

	for (user = user_list.next; user != &user_list; user = user->next) {
		if (!strcmp(who, user->nick)) {

			char str[256];
			sprintf (str, "You have been moved to '%s' by %s.", where, u->nick);

			bot_to_user (user, str);

			proto_nmdc_user_forcemove (user, where);

			return 0;
		}
	}

	return 0;
}

bool handler_statnmdc (user_t *u, char *data)
{
	char str[1024];

	sprintf (str, "\n--------------------- [ NMDC Stats ] ---------------------"
		      "\n\ncache rebuild: %lld"
		      "\ncache mainchat message: %ld"
		      "\ncache command: %ld"
		      "\ncache private message: %ld"
		      "\ncache zpipe users: %ld"
		      "\nsearch: %ld"
		      "\nconnect: %ld"
		      "\nredirect: %ld"
		      "\nkick: %ld"
		      "\nban: %ld"
		      "\ndrop: %ld"
		      "\nbotinfo: %ld",
		      sec, stats.num_chats, stats.num_commands, stats.num_pm, stats.cache_zpipers, stats.num_search, stats.num_connect,
		      stats.num_forcemove, stats.num_kick, stats.num_ban, stats.num_drop, stats.botinfo
		);

	bot_to_user (u, str);

	return 0;
}

bool handler_statbuffer (user_t *u, char *data)
{
	char str[1024];

	sprintf (str, "\n--------------------- [ Buffers Stats ] ---------------------"
		      "\n\nZPipe users: %ld"
		      "\nZPipe compression (Last/Avarage): %.2f/%.2f %%",
		      stats.cache_zpipers, stats.buf_zpipe_compression, stats.buf_zpipe_compression_avg/stats.buf_zpipe_count
		);

	bot_to_user (u, str);

	return 0;
}

bool handler_mass (user_t *u, char *data)
{
	char cmd[192];
	strcpy (cmd, strarg (1, data, strlen (data)));

	if (cmd[0] != '0') {
		char use[80];
		sprintf (use, "Use:\n%s <message>", "mass");
		bot_to_user (u, use);
		return 0;
	}

	char bot[128];
	strcpy (bot, config_get ((char *) "bot.name"));

	unsigned long l = strlen (data);
	unsigned int bl = strlen (bot);
	unsigned int cl = strlen (cmd);

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
		unsigned int nl = user->nick_len;

		char *str = (char *) calloc (16+nl+bl+bl+l-cl, sizeof (char));
	
		if (!str)
			return 0;
		
		memcpy (str, "$To: ", 5);
		memcpy (str+5, user->nick, nl);
		memcpy (str+5+nl, " $From: ", 8);
		memcpy (str+13+nl, bot, bl);
 		memcpy (str+13+nl+bl, " $<", 3);
		memcpy (str+16+nl+bl, bot, bl);
		memcpy (str+16+nl+bl+bl, "> ", 2);
		memcpy (str+18+nl+bl+bl, data+cl+4, l-cl-4);
		memcpy (str+14+nl+bl+bl+l-cl, "|", 1);

		str[15+nl+bl+bl+l-cl] = '\0';

		printf ("oj: '%s'\n", str);
		
		send_to_user_len (user, str, 15+nl+bl+bl+l-cl);

		free (str);
	}

	return 0;
}

bool handler_say (user_t *u, char *data)
{
	char cmd[192];
	strcpy (cmd, strarg (1, data, strlen (data)));

	if (cmd[0] != '0') {
		char use[80];
		sprintf (use, "Use:\n%s <message>", "say");
		bot_to_user (u, use);
		return 0;
	}

	char bot[128];
	bot[0] = '\0';
	strcpy (bot, config_get ((char *) "bot.name"));

	unsigned long l = strlen (data);

	char *str = (char *) calloc (l, sizeof (char));

	memcpy (str, data+4, l-4);

	user_t *user;

	for (user = user_list.next; user != &user_list; user = user->next)
		bot_to_user (user, str);

	free (str);

	return 0;
}

bool handler_me (user_t *u, char *data)
{
	char cmd[192];
	strcpy (cmd, strarg (1, data, strlen (data)));

	if (cmd[0] != '0') {
		char use[80];
		sprintf (use, "Use:\n%s <message>", "me");
		bot_to_user (u, use);
		return 0;
	}

	unsigned long l = strlen (data);

	char *str = (char *) calloc (l+64, sizeof (char));

	memcpy (str, "*", 1);
	memcpy (str+1, u->nick, u->nick_len);
	memcpy (str+u->nick_len+1, " ", 1);
	memcpy (str+u->nick_len+2, data+3, l-3);
	memcpy (str+u->nick_len+l-1, "|", 1);

	send_to_all_len (str, u->nick_len+l);

	free (str);

	return 0;
}

bool handler_chatlock (user_t *u, char *data)
{
	if (config.deflood.mainchat_lock) {
		config.deflood.mainchat_lock = false;
		
		char *str = (char *) calloc (28, sizeof (char));
	
		memcpy (str, "** Mainchat was unlocked **", 27);
		str[27] = '\0';

		user_t *user;
		for (user = user_list.next; user != &user_list; user = user->next)
			bot_to_user (user, str);
	
		free (str);

		return 0;
	} else {
		config.deflood.mainchat_lock = true;
		
		char *str = (char *) calloc (26, sizeof (char));
	
		memcpy (str, "** Mainchat was locked **", 25);
		str[25] = '\0';

		user_t *user;
		for (user = user_list.next; user != &user_list; user = user->next)
			bot_to_user (user, str);
	
		free (str);

		return 0;
	}

	return 0;
}

bool handler_pmlock (user_t *u, char *data)
{
	if (config.deflood.pm_lock) {
		config.deflood.pm_lock = false;
		
		char *str = (char *) calloc (24, sizeof (char));
	
		memcpy (str, "** PMs were unlocked **", 23);
		str[27] = '\0';

		user_t *user;
		for (user = user_list.next; user != &user_list; user = user->next)
			bot_to_user (user, str);
	
		free (str);

		return 0;
	} else {
		config.deflood.pm_lock = true;
		
		char *str = (char *) calloc (22, sizeof (char));
	
		memcpy (str, "** PMs were locked **", 21);
		str[25] = '\0';

		user_t *user;
		for (user = user_list.next; user != &user_list; user = user->next)
			bot_to_user (user, str);
	
		free (str);

		return 0;
	}

	return 0;
}

bool handler_iplog (user_t *u, char *data)
{
	int iplog = atoi (config_get ((char *) "log.ip"));

	if (!iplog) {
		bot_to_user (u, "No ip in database, please change value of variable 'log.ip' to higher number.");
		return 0;
	}

	int i = 0, d = 0;
	char *str = strdup ("Last connected ip address:\n");
	
	if (!str)
		return 0;
	
	unsigned l = strlen (str);
	
	/* check count of the clients */
	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next)
		i ++;

	for (user = user_list.next; user != &user_list; user = user->next) {
		d ++;

		if (d > (i - iplog)) {
			unsigned ip = strlen (user->ip);
			
			str = (char *) realloc ((void *) str, l + ip + user->nick_len + 4);
			
			if (!str)
				return 0;
	
			memcpy (str+l, user->ip, ip);
			
			if (user->nick_len > 0) {
				memcpy (str+l+ip, " - ", 3);
				memcpy (str+l+ip+3, user->nick, user->nick_len);
				l += ip+3+user->nick_len;
			}
			
			memcpy (str+l, "\n", 1);
			l ++;
		}
	}

	bot_to_user_len (u, str, l);
	
	free (str);

	return 0;
}

bool handler_gag (user_t *u, char *data)
{
	char a[16], b[64];
	a[0] = '\0'; b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick>", a);
		bot_to_user (u, use);
		return 0;
	}

	user_t *target = user_findbynick (b);

	if (!target) {
		char err[128];
		sprintf (err, "User '%s' is offline.", b);
	
		bot_to_user (u, err);
		return 0;
	}

	if (!user_flag_add (target, USER_FLAG_GAG)) {
		char str[128];
		sprintf (str, "User '%s' is already gagged.", b);
	
		bot_to_user (u, str);
		return 0;
	}

	char str[128];
	sprintf (str, "User '%s' was gagged.", b);

	bot_to_user (u, str);

	return 1;
}

bool handler_ungag (user_t *u, char *data)
{
	char a[16], b[64];
	a[0] = '\0'; b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[80];
		sprintf (use, "Use:\n%s <nick>", a);
		bot_to_user (u, use);
		return 0;
	}

	user_t *target = user_findbynick (b);

	if (!target) {
		char err[128];
		sprintf (err, "User '%s' is offline.", b);
	
		bot_to_user (u, err);
		return 0;
	}

	if (!user_flag_del (target, USER_FLAG_GAG)) {
		char str[128];
		sprintf (str, "User '%s' is'nt gagged.", b);
	
		bot_to_user (u, str);
		return 0;
	}

	char str[128];
	sprintf (str, "User '%s' was ungagged.", b);

	bot_to_user (u, str);

	return 1;
}

/*************************************************************\
|		        COMMANDS PARSER	  		      |
\*************************************************************/

int command_parser (user_t *u, char *command, char *data, unsigned int len)
{
	command_t *cmd;

	int l = u->nick_len + 4;

	char *arg = data+l;
	
	arg[len-l-1] = '\0';

	int i = 0;

	for (cmd = command_list.next; cmd != &command_list; cmd = cmd->next) {
    	    if (!strcmp (command+1, cmd->name) && u->iclass >= cmd->iclass) {
		cmd->handler (u, arg);

		printf ("> Command -> %s -> %d -> %s\n", arg, u->socket, u->nick);

		i ++;
	    }
	}

	if (i == 0) {
		char err[128];
		sprintf (err, "Comannd '%s' not found.", command+1);
		
		bot_to_user (u, err);	
		return 0;
	}

	return 1;
}

/*************************************************************\
|		        COMMAND's UTILS	  		      |
\*************************************************************/

int command_register (char *name, char *desc, command_handler_t *handler, unsigned int iclass)
{
	command_t *c;	
	/* we dont want commands with same name */
	for (c = command_list.next; c != &command_list; c = c->next)
		if (!strcmp(c->name, name))
			return 0;

	command_t *ctx;	

	// alloc and init context
	ctx = (command_t*) malloc (sizeof (command_t));

	strcpy (ctx->name, name);
	strcpy (ctx->desc, desc);
	ctx->handler = handler;
	ctx->iclass  = iclass;

	ctx->next = &command_list;
	ctx->prev = command_list.prev;
	ctx->prev->next = ctx;
	ctx->next->prev = ctx;

	return 1;
}

int command_unregister (char *name)
{
	command_t *ctx;	

	for (ctx = command_list.next; ctx != &command_list; ctx = ctx->next) {
		if (!strcmp(ctx->name, name)) {
			ctx->next->prev = ctx->prev;
			ctx->prev->next = ctx->next;

			return 1;
		}
	}
	
	return 0;
}

/*************************************************************\
|		  INITIALIZATION OF COMMANDS		      |
\*************************************************************/

int init_commands ()
{
	command_list.next = &command_list;
	command_list.prev = &command_list;
	command_list.iclass = 0;

	strcpy (command_list.name, "N/A");
	strcpy (command_list.desc, "N/A");

#ifndef __zexos__
	cpu_init ();
#endif
	// reg commands
	command_register ((char *) "regnew", (char *) "Add a user account.", &handler_regnew , 3);
	command_register ((char *) "regdel", (char *) "Delete a user account.", &handler_regdel , 3);
	command_register ((char *) "regclass", (char *) "Change class of user account.", &handler_regclass , 3);
	command_register ((char *) "reglist", (char *) "Displays all registered users.", &handler_reglist , 0);
	command_register ((char *) "regme", (char *) "Ask for account.", &handler_regme , 0);
	command_register ((char *) "passwd", (char *) "Set password for account.", &handler_passwd , 0);
	command_register ((char *) "gag", (char *) "User can't type on mainchat.", &handler_gag , 3);
	command_register ((char *) "ungag", (char *) "Disable gag.", &handler_ungag , 3);
	command_register ((char *) "set", (char *) "Set config variable.", &handler_set , 4);
	command_register ((char *) "=", (char *) "Set config variable.", &handler_set , 4);
	command_register ((char *) "getconfig", (char *) "Get values of variables.", &handler_getconfig , 4);
	command_register ((char *) "gc", (char *) "Get values of variables.", &handler_getconfig , 4);
	command_register ((char *) "statnmdc", (char *) "Displays use of nmdc commands.", &handler_statnmdc , 0);
	command_register ((char *) "statbuffer", (char *) "Displays use of buffers.", &handler_statbuffer , 0);
#ifndef __zexos__
	command_register ((char *) "statcpu", (char *) "Get info about cpu.", &handler_cpuinfo , 0);
#endif
	command_register ((char *) "statbw", (char *) "Displays usege of bandwidth.", &handler_statbw , 0);
#ifndef __WIN32__
#ifndef __zexos__
	command_register ((char *) "osinfo", (char *) "Get info about OS.", &handler_linuxinfo , 0);
#endif
#endif
	command_register ((char *) "iplog", (char *) "Log of last connected users ip.", &handler_iplog , 1);
	command_register ((char *) "help", (char *) "Do you know ?", &handler_help , 0);
	command_register ((char *) "myinfo", (char *) "Yourself information.", &handler_myinfo , 0);
	command_register ((char *) "getinfo", (char *) "Information about user.", &handler_getinfo , 2);
	command_register ((char *) "whoip", (char *) "Prints all users who have the specified IP.", &handler_whoip , 2);
	command_register ((char *) "uptime", (char *) "Hub uptime.", &handler_uptime , 0);
	command_register ((char *) "hubinfo", (char *) "Info about hub.", &handler_hubinfo , 0);
	command_register ((char *) "redirect", (char *) "Redirect user on another hub.", &handler_redirect , 3);
	command_register ((char *) "kick", (char *) "Kick user from hub.", &handler_kick , 3);
	command_register ((char *) "drop", (char *) "Drop user from hub.", &handler_drop , 3);
	command_register ((char *) "bannick", (char *) "Ban user by nick.", &handler_bannick , 3);
	command_register ((char *) "banip", (char *) "Ban user by ip address.", &handler_banip , 3);
	command_register ((char *) "banperm", (char *) "Ban user forever.", &handler_banperm , 3);
	command_register ((char *) "bandel", (char *) "Delete ban of user.", &handler_bandel , 3);
	command_register ((char *) "banlist", (char *) "Dispays all banned users.", &handler_banlist , 1);
	command_register ((char *) "mass", (char *) "Send message to all users.", &handler_mass , 4);
	command_register ((char *) "say", (char *) "Send message to all users as bot.", &handler_say , 4);
	command_register ((char *) "me", (char *) "Write message as third person.", &handler_me , 0);
	command_register ((char *) "chatlock", (char *) "Block all messages on mainchat.", &handler_chatlock , 4);
	command_register ((char *) "pmlock", (char *) "Block all private messages.", &handler_pmlock , 4);
	command_register ((char *) "rebuild", (char *) "Refresh data and userlist.", &handler_rebuild , 4);
	command_register ((char *) "save", (char *) "Save config.", &handler_cfgsave , 4);
	command_register ((char *) "load", (char *) "Load config.", &handler_cfgload , 4);
	command_register ((char *) "shutdown", (char *) "Hub go down.", (command_handler_t *) &handler_shutdown , 5);

	return 1;
}
