/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@gmail.com)
 *  (C) Copyright 2009 ZeXx86 (zexx86@gmail.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#ifdef __WIN32__
 #include <windows.h>
#endif
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include "utils.h"
#include "defaults.h"
#include "chatroom.h"
#include "proto.h"
#include "ban.h"
#include "plugin.h"
#ifdef LUA
# include "elua.h"
#endif
extern config_elem_t config_list;
extern config_t config;

int config_find (char *name)
{
	config_elem_t *cfg;

	for (cfg = config_list.next; cfg != &config_list; cfg = cfg->next) {
		if (!strcmp (name, cfg->name))
			return 1;
	}	

	return 0;
}

int config_save (char *filename)
{

  FILE *fp;
  config_elem_t *cfg;

  fp = fopen (filename, "w+");
  if (!fp)
    return 0;
	
  for (cfg = config_list.next; cfg != &config_list; cfg = cfg->next)
	fprintf (fp, "%s %s\n", cfg->name, cfg->value);

  fclose (fp);

  chatroom_save ();
  ban_save ();
  plugin_save ();
#ifdef LUA
  luascript_save ();
#endif
  return 1;
}

int config_load (char *filename)
{
  unsigned int l;
  FILE *fp;

  char *buffer, *c;

  fp = fopen (filename, "r");
  
  if (!fp)
	return config_save (filename);

  buffer = (char*) calloc (4097, sizeof (char));

  while (!feof (fp)) {
    fgets (buffer, 4096, fp);
    for (c = buffer; *c && *c != ' '; c++);
    if (!*c)
      continue;
    *c++ = '\0';

    if (!config_find (buffer))
      continue;

	l = strlen (c);
	if (c[l - 1] == '\n')
	  c[l-- - 1] = '\0';	

	//c[l - 1] = '\0';

	config_set (buffer, c);

#ifdef DEBUG
	printf ("Config: '%s' : '%s'\n", buffer, c);
#endif
  }
 
  free (buffer);
  fclose (fp);

  get_motd ();								// FIXME memleak

  cache_rebuild ();							// rebuild all needed strings

  return 1;

}

int config_register (char *name, char *desc, int type, char *value)
{
	if (config_find (name))
		return 0;

	config_elem_t *ctx;

	// alloc and init context
	ctx = (config_elem_t *) malloc (sizeof (config_elem_t));

	if (!ctx)
		return 0;
	
	strcpy (ctx->name, name);
	strcpy (ctx->desc, desc);
	ctx->type = type;

	unsigned int value_len = strlen (value);
	ctx->value = (char *) malloc (value_len + 1);
	
	if (!ctx->value) {
		free (ctx);
		return 0;
	}
	
	memcpy (ctx->value, value, value_len);
	ctx->value[value_len] = '\0';

	ctx->next = &config_list;
	ctx->prev = config_list.prev;
	ctx->prev->next = ctx;
	ctx->next->prev = ctx;
	
	return 1;
}

char *config_get (char *name)
{
	config_elem_t *cfg;

	for (cfg = config_list.next; cfg != &config_list; cfg = cfg->next) {
		if (!strcmp (name, cfg->name))
			return cfg->value;
	}

	return "N/A";
}

int config_set (char *name, char *value)
{
	config_elem_t *cfg;

	for (cfg = config_list.next; cfg != &config_list; cfg = cfg->next) {
		if (!strcmp (name, cfg->name)) {
			if (cfg->value && *cfg->value)
				free (cfg->value);
			
			unsigned int value_len = strlen (value);
			cfg->value = (char *) calloc (value_len + 1, sizeof (char));

			memcpy (cfg->value, value, value_len);
			cfg->value[value_len] = '\0';

			return 1;
		}
	}

	return 0;
}

int init_config ()
{
	config_list.next = &config_list;
	config_list.prev = &config_list;

	config.autosave = 0;		// interval of saving config (sec)
	config.userlimit = 0;
	config.motd = 0;
	config.motd_len = 0;

	config_register ("hubname", "Name of hub", 0, "Lama Hub");
	config_register ("hubdesc", "Description of hub", 0, "Created by ZeXx86");
	config_register ("hubaddress", "Address of hub without port", 0, "255.255.255.255");
	config_register ("hubowner", "Owner of hub", 0, "Unnamed");
	config_register ("hublist.hosts", "Hublist autoreg address", 0, "");
	config_register ("hublist.interval", "Interval of hublist update", 0, "0");
	config_register ("bot.name", "Name of bot", 0, "LamaBot");
	config_register ("autosave.interval", "Interval of autosaving config", 0, "600");
	config_register ("search.interval.guest", "Search interval of guests", 0, "64");
	config_register ("search.interval.reg", "Search interval of regs", 0, "32");
	config_register ("search.interval.vip", "Search interval of vips", 0, "16");
	config_register ("search.interval.op", "Search interval of ops", 0, "10");
	config_register ("minslots.guest", "Min. slots of guests", 0, "2");
	config_register ("minslots.reg", "Min. slots of regs", 0, "1");
	config_register ("minslots.vip", "Min. slots of vips", 0, "0");
	config_register ("minslots.op", "Min. slots of ops", 0, "0");
	config_register ("minshare.guest", "Min. share in bytes of guests", 0, "0");
	config_register ("minshare.reg", "Min. share in bytes of regs", 0, "0");
	config_register ("minshare.vip", "Min. share in bytes of vips", 0, "0");
	config_register ("minshare.op", "Min. share in bytes of ops", 0, "0");
	config_register ("redirect", "Address of hub for bad users", 0, "all-hosting.biz:4111");
	config_register ("listenport", "Listening port of hub", 0, "4111");
	config_register ("userlimit", "Max. limit of connected users", 0, "1010");
	config_register ("deflood.mainchat", "Deflood of mainchat", 0, "3");
	config_register ("log.ip", "Remember count ip address of users.", 0, "16");
	config_register ("len.pm", "Length of private message.", 0, "768");
	config_register ("len.chat", "Length of chat message.", 0, "512");
	config_register ("buffering.hardlimit", "Hard limit of user send buffer.", 0, "1024");
	config_register ("buffering.softlimit", "Limit, which descrease hard limit.", 0, "1024");

	cache_rebuild ();							// rebuild all needed strings
	
	return 1;
}
