/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@gmail.com)
 *  (C) Copyright 2009 ZeXx86 (zexx86@gmail.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#ifdef __WIN32__
 #include <windows.h>
#endif
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
//#include <math.h>
#include "chatroom.h"
#include "commands.h"
#include "defaults.h"
#include "config.h"
#include "utils.h"
#include "proto.h"
#include "user.h"
#include "elua.h"

#ifdef LUA
#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"

/* Lua script context */
unsigned int lua_context_count, lua_context_peak;
typedef struct lua_context {
	struct lua_context *next, *prev;

	lua_State *l;
	char *name;
	unsigned long events;
} lua_context_t;

lua_context_t lua_list;

const char *lua_eventnames[] = {
  "EventNewClient",
  "EventUserConnected",
  "EventGetNickList",
  "EventLock",
  "EventSearch",
  "EventQuit",
  "EventChat",
  "EventPM",
  "EventCommand",
  "EventOpConnected",
  "EventTimer",
  "EventChatIn",
  NULL
};

extern user_t user_list;
extern command_t command_list;
extern config_t config;
extern timer_lua_t timer;
extern ban_t ban_list;

extern long long int sec;     // Uptime
extern int cont;

/* constructor */
bool handler_lua_func_cmd (user_t *u, char *data);
int lua_close_script (user_t *u, lua_context_t *ctx);

/*************************************************************\
|		    LUA REGISTERED FUNCTION		      |
\*************************************************************/

int lua_func_shutdown (lua_State *lua)
{
	cont = 0;

	return 1;
}

int lua_func_sendtoall (lua_State *lua)
{
	char *message = (char *) luaL_checkstring (lua, 1);

	if (message)
		send_to_all (message);

	return 1;
}

int lua_func_sendtouser (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);
	char *message = (char *) luaL_checkstring (lua, 2);

	if (nick && message) {
		user_t *u = user_findbynick (nick);

		if (u)
			send_to_user (u, message);
	}

	return 1;
}

int lua_func_bottouser (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);
	char *message = (char *) luaL_checkstring (lua, 2);

	if (nick && message) {
		user_t *u = user_findbynick (nick);

		if (u)
			bot_to_user (u, message);
	}

	return 1;
}

int lua_func_bottoall (lua_State *lua)
{
	char *message = (char *) luaL_checkstring (lua, 1);

	if (!message)
		return 0;

	bot_to_all (message);

	return 1;
}

int lua_func_getconfig (lua_State *lua)
{
	char *var = (char *) luaL_checkstring (lua, 1);

	if (!var)
		return 0;

	char *ret = config_get (var);

	if (strcmp (ret, "N/A")) {
		lua_pushstring (lua, ret);

		return 1;
	}

	lua_pushnil (lua);

	return 0;
}

int lua_func_setconfig (lua_State *lua)
{
	char *var = (char *) luaL_checkstring (lua, 1);
	char *val = (char *) luaL_checkstring (lua, 2);

	if (!var || !val)
		return 0;

	lua_pushboolean (lua, config_set (var, val));

	return 1;
}

int lua_func_bannick (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);
	char *time = (char *) luaL_checkstring (lua, 2);

	user_t *u = user_findbynick (nick);

	if (u) {
		ban_add_bynick (u, time);

		proto_nmdc_user_redirect (u);

		lua_pushboolean (lua, 1);
	} else
		lua_pushboolean (lua, 0);


	return 1;
}

int lua_func_banip (lua_State *lua)
{
	char *ip = (char *) luaL_checkstring (lua, 1);
	char *time = (char *) luaL_checkstring (lua, 2);

	user_t *u = NULL;
	u = user_findbyip (ip);

	if (u) {
		ban_add_byip (u, time);

		proto_nmdc_user_redirect (u);

		lua_pushboolean (lua, 1);
	} else
		lua_pushboolean (lua, 0);

	return 1;
}

int lua_func_banperm (lua_State *lua)
{
	char *ip = (char *) luaL_checkstring (lua, 1);

	user_t *u = user_findbyip (ip);

	if (u) {
		ban_add_byip (u, "0");

		proto_nmdc_user_redirect (u);

		lua_pushboolean (lua, 1);
	} else
		lua_pushboolean (lua, 0);

	return 1;
}

int lua_func_bandel (lua_State *lua)
{
	char *val = (char *) luaL_checkstring (lua, 1);

	ban_t *ban;
	unsigned short i = 0;

	for (ban = ban_list.next; ban != &ban_list; ban = ban->next) {
		if (!strcmp (ban->nick, val)) {
			ban_del (ban);
			i ++;
			break;
		}
		if (!strcmp (ban->ip, val)) {
			ban_del (ban);
			i ++;
			break;
		}
	}

	if (i)
		lua_pushboolean (lua, 1);
	else
		lua_pushboolean (lua, 0);

	return 1;
}

int lua_func_kick (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);

	user_t *u = user_findbynick (nick);

	if (u) {
		ban_add_bynick (u, "120");
		proto_nmdc_user_redirect (u);

		lua_pushboolean (lua, 1);
	} else
		lua_pushboolean (lua, 0);

	return 1;
}

int lua_func_userip (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);

	user_t *u = user_findbynick (nick);

	if (u)
		lua_pushstring (lua, u->ip);
	else
		lua_pushnil (lua);

	return 1;
}

int lua_func_userclass (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);

	user_t *u = user_findbynick (nick);

	if (u)
		lua_pushnumber (lua, u->iclass);
	else
		lua_pushnumber (lua, -1);

	return 1;
}

int lua_func_usershare (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);

	user_t *u = user_findbynick (nick);

	if (u)
		lua_pushnumber (lua, u->share);
	else
		lua_pushnumber (lua, -1);

	return 1;
}

int lua_func_usertag (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);

	user_t *u = user_findbynick (nick);

	if (u) {
		lua_pushnil (lua);
		return 1;
	}
	
	char tag[128];
	unsigned int r = 0, tags = u->myinfo_len, tage = 0;

	while (r != (unsigned) u->myinfo_len) {
		if (u->myinfo[r] == '<')
			tags = r;

		if (u->myinfo[r] == '>')
			tage = r;

		r ++;
	}

	if (tags < tage) {
		memcpy (tag, u->myinfo+tags+1, tage-tags-1);
		tag[tage-tags-1] = '\0';
	}

	lua_pushstring (lua, tag);
	
	return 1;
}

int lua_func_drop (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);

	user_t *u = user_findbynick (nick);

	if (u) {
		u->state = PROTO_STATE_DISCONNECTED;

		lua_pushboolean (lua, 1);
	} else
		lua_pushboolean (lua, 0);

	return 1;
}

int lua_func_addbot (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);
	char *desc = (char *) luaL_checkstring (lua, 2);
	char *conn = (char *) luaL_checkstring (lua, 3);
	char *mail = (char *) luaL_checkstring (lua, 4);
	unsigned short iclass = (unsigned short) luaL_checknumber (lua, 5);
	unsigned long long share = (unsigned long long) luaL_checknumber (lua, 6);

	if (!nick || !desc || !conn || !mail)
		return 0;

	user_t *t = user_findbynick (nick);

	if (t) {
		lua_pushboolean (lua, 0);
		return 0;
	}

	/* create virtual context */
	user_t *u;	
		
	// alloc and init context
	u = (user_t*) malloc (sizeof (user_t));

	if (!u)
		return 0;
	
	u->state = PROTO_STATE_ONLINE;
	u->socket = 0;
	u->iclass = iclass;
	u->share = 0;
	u->searchlast = 0;
	u->flag = 0;
	u->online = true;
	u->buf_len = 0;
	u->buf = (char *) malloc (33);
	u->buf_offset = 0;
	u->buf_used = 0;
	
	if (!u->buf) {
		free (u);
		return 0;
	}
	
	u->nick_len = strlen (nick);

	strcpy (u->nick, nick);
	strcpy (u->ip, "127.0.0.1");

	sprintf (u->myinfo, "$MyINFO $ALL %s %s$ $%s$%s$%llu$|", nick, desc, conn, mail, share);
	u->myinfo_len = strlen (u->myinfo);

	// add into list  
	u->next = &user_list;
	u->prev = user_list.prev;
	u->prev->next = u;
	u->next->prev = u;
	
	/* send to all users virtual myinfo */
	send_to_all (u->myinfo);

	lua_pushboolean (lua, 1);

	return 1;
}

int lua_func_addcommand (lua_State *lua)
{
	char *name = (char *) luaL_checkstring (lua, 1);
	char *desc = (char *) luaL_checkstring (lua, 2);
	unsigned short minclass = (unsigned short) luaL_checknumber (lua, 3);

	if (!name || !desc)
		return 0;

	command_register (name, desc, &handler_lua_func_cmd , minclass);

	lua_pushboolean (lua, 1);

	return 1;
}

int lua_func_delcommand (lua_State *lua)
{
	char *name = (char *) luaL_checkstring (lua, 1);

	if (!name)
		return 0;

	lua_pushboolean (lua, command_unregister (name));

	return 1;
}

int lua_func_addchatroom (lua_State *lua)
{
	char *name = (char *) luaL_checkstring (lua, 1);
	char *desc = (char *) luaL_checkstring (lua, 2);
	unsigned short minclass = (unsigned short) luaL_checknumber (lua, 3);

	if (!name || !desc)
		return 0;

	unsigned short l = strlen (name);
	unsigned short d = strlen (desc);

	chatroom_t *ctx;	

	// alloc and init context
	ctx = (chatroom_t*) malloc (sizeof (chatroom_t));

	strcpy (ctx->name, name);
	ctx->name [l] = '\0';
	strcpy (ctx->desc, desc);
	ctx->desc [d] = '\0';
	ctx->minclass = minclass;
	ctx->myinfo_len = 27+l+d;

	ctx->myinfo = (char *) calloc(28+l+d, sizeof(char));

	/* built myinfo */
	memcpy(ctx->myinfo, "$MyINFO $ALL ", 13);
	memcpy(ctx->myinfo+13, name, l);
	memcpy(ctx->myinfo+13+l, " ", 1);
	memcpy(ctx->myinfo+14+l, desc, d);
	memcpy(ctx->myinfo+14+l+d, " $ $ROOM $$$|", 13);

	ctx->next = &chatroom_list;
	ctx->prev = chatroom_list.prev;
	ctx->prev->next = ctx;
	ctx->next->prev = ctx;

	chatroom_rebuild ();	/* now we can rebuild chatroom myinfos */

	lua_pushboolean (lua, 1);

	return 1; 
}

int lua_func_delchatroom (lua_State *lua)
{
	char *name = (char *) luaL_checkstring (lua, 1);

	if (!name)
		return 0;

	chatroom_t *room = NULL;
	room = chatroom_find (name);
	
	if (!room) {
		lua_pushboolean (lua, 0);
		return 0;
	}

	/* remove a chatroom from context */
 	room->next->prev = room->prev;
 	room->prev->next = room->next; 
 
	free (room);

	chatroom_rebuild ();	/* now we can rebuild chatroom myinfos */

	lua_pushboolean (lua, 1);

	return 1;
}

int lua_func_pmtouser (lua_State *lua)
{
	char *target = (char *) luaL_checkstring (lua, 1);
	char *nick = (char *) luaL_checkstring (lua, 2);
	char *data = (char *) luaL_checkstring (lua, 3);

	if (!target || !data)
		return 0;

	user_t *t = NULL;
	t = user_findbynick (target);

	if (!t) {
		lua_pushboolean (lua, 0);
		return 0;
	}

	pm_to_user (t, nick, data);

	lua_pushboolean (lua, 1);

	return 1;
}

int lua_func_pmtoall (lua_State *lua)
{
	char *nick = (char *) luaL_checkstring (lua, 1);
	char *data = (char *) luaL_checkstring (lua, 2);

	if (!data)
		return 0;

	user_t *u;
	for (u = user_list.next; u != &user_list; u = u->next)
		pm_to_user (u, nick, data);

	lua_pushboolean (lua, 1);

	return 1;
}

int lua_func_settimer (lua_State *lua)
{
	unsigned int interval = (unsigned int) luaL_checknumber (lua, 1);

	if (!interval)
		return 0;

	timer.interval = interval;

	return 1;
}

int lua_func_starttimer (lua_State *lua)
{
	timer.active = true;

	return 1;
}

int lua_func_stoptimer (lua_State *lua)
{
	timer.active = false;

	return 1;
}

int lua_func_sharetostring (lua_State *lua)
{
	unsigned long long share =
		(unsigned long long) luaL_checknumber (lua, 1);

	char *str = strdup (share_tostring (share));
	
	lua_pushstring (lua, str);

	free (str);

	return 1;
}

int lua_func_timetostring (lua_State *lua)
{
	unsigned long long time =
		(unsigned long long) luaL_checknumber (lua, 1);

	char *str = strdup (time_tostring (time));
	
	lua_pushstring (lua, str);

	free (str);

	return 1;
}

int lua_func_chatlock (lua_State *lua)
{
	config.deflood.mainchat_lock = true;

	return 1;
}

int lua_func_chatunlock (lua_State *lua)
{
	config.deflood.mainchat_lock = false;

	return 1;
}

int lua_func_pmlock (lua_State *lua)
{
	config.deflood.pm_lock = true;

	return 1;
}

int lua_func_pmunlock (lua_State *lua)
{
	config.deflood.pm_lock = true;

	return 1;
}

int lua_func_uptime (lua_State *lua)
{
	lua_pushnumber (lua, sec);

	return 1;
}

/*************************************************************\
|		     LUA FUNCTION REGISTER		      |
\*************************************************************/

int lua_register_functions (lua_State *l)
{
	// make hub functions available to Lua programs
	/* work with server */
	lua_register (l, "Shutdown", lua_func_shutdown);

	/* socket function - for send data */
	lua_register (l, "SendToAll", lua_func_sendtoall);
	lua_register (l, "SendToUser", lua_func_sendtouser);

	/* send message as bot */
	lua_register (l, "BotToUser", lua_func_bottouser);
	lua_register (l, "BotToAll", lua_func_bottoall);

	/* send private message */
	lua_register (l, "PMToUser", lua_func_pmtouser);
	lua_register (l, "PMToAll", lua_func_pmtoall);

	/* get/set config of hub */
	lua_register (l, "ConfigGet", lua_func_getconfig);
	lua_register (l, "ConfigSet", lua_func_setconfig);
	
	/* ban/kick/drop user */
	lua_register (l, "BanNick", lua_func_bannick);
	lua_register (l, "BanIP", lua_func_banip);
	lua_register (l, "BanPerm", lua_func_banperm);
	lua_register (l, "BanDel", lua_func_bandel);
	lua_register (l, "Kick", lua_func_kick);
	lua_register (l, "Drop", lua_func_drop);

	/* user info */
	lua_register (l, "UserIP", lua_func_userip);
	lua_register (l, "UserClass", lua_func_userclass);
	lua_register (l, "UserShare", lua_func_usershare);
	lua_register (l, "UserTag", lua_func_usertag);

	/* virtual user */
	lua_register (l, "BotAdd", lua_func_addbot);
	lua_register (l, "BotDel", lua_func_drop);

	/* add/del command */
	lua_register (l, "CommandAdd", lua_func_addcommand);
	lua_register (l, "CommandDel", lua_func_delcommand);

	/* add/del chatroom*/
	lua_register (l, "RoomAdd", lua_func_addchatroom);
	lua_register (l, "RoomDel", lua_func_delchatroom);

	/* timer */
	lua_register (l, "TimerSet", lua_func_settimer);
	lua_register (l, "TimerStart", lua_func_starttimer);
	lua_register (l, "TimerStop", lua_func_stoptimer);

	/* utils */
	lua_register (l, "ShareToString", lua_func_sharetostring);
	lua_register (l, "TimeToString", lua_func_timetostring);
	lua_register (l, "ChatLock", lua_func_chatlock);
	lua_register (l, "ChatUnlock", lua_func_chatunlock);
	lua_register (l, "PMLock", lua_func_pmlock);
	lua_register (l, "PMUnlock", lua_func_pmlock);
	lua_register (l, "Uptime", lua_func_uptime);

	return 1;
}

#endif
/*************************************************************\
|		   	   LUA WORK			      |
\*************************************************************/

int event_lua (user_t* u, char *data, unsigned long event, unsigned long len)
{
#ifdef LUA
	lua_context_t *lua;
	int result = 0;

	/* send info to all scripts about event */
	for (lua = lua_list.next; lua != &lua_list; lua = lua->next) {
		if (!(lua->events & (1 << event)))
			continue;

		lua_settop     (lua->l, 0);				// clear stack
	
		lua_pushstring (lua->l, lua_eventnames[event]);
		lua_gettable   (lua->l, LUA_GLOBALSINDEX);
	
		lua_pushstring (lua->l, u->nick);
		lua_pushlstring (lua->l, data, len);
	
		/* call funtion. */
		result = lua_pcall (lua->l, 2, 1, 0);

		if (result) {
			char *error = (char *) luaL_checkstring (lua->l, 1);
	
			char err[2048];
			sprintf (err, "Lua script return ERROR at event (%s): %s", lua_eventnames[event], error);
	
			bot_to_user (u, err);
		}
		
		/* retrieve return value */
		if (lua_isboolean (lua->l, -1)) {
			result = lua_toboolean (lua->l, -1);
			lua_remove (lua->l, -1);
		}
	}
#endif
	return 1;
}

int event_lua_hub (user_t* u, unsigned long event)
{
#ifdef LUA
	lua_context_t *lua;
	int result = 0;

	/* send info to all scripts about event */
	for (lua = lua_list.next; lua != &lua_list; lua = lua->next) {
		if (!(lua->events & (1 << event)))
			continue;

		lua_settop     (lua->l, 0);				// clear stack
	
		lua_pushstring (lua->l, lua_eventnames[event]);
		lua_gettable   (lua->l, LUA_GLOBALSINDEX);
	
		lua_pushstring (lua->l, u->nick);
	
		/* call funtion. */
		result = lua_pcall (lua->l, 1, 1, 0);

		if (result) {
			char *error = (char *) luaL_checkstring (lua->l, 1);
	
			char err[2048];
			sprintf (err, "Lua script return ERROR at event (%s): %s", lua_eventnames[event], error);
	
			bot_to_user (u, err);
		}
		
		/* retrieve return value */
		if (lua_isboolean (lua->l, -1)) {
			result = lua_toboolean (lua->l, -1);
			lua_remove (lua->l, -1);
		}
	}
#endif
	return 1;
}

int event_lua_sys (unsigned long event)
{
#ifdef LUA
	lua_context_t *lua;
	int result = 0;

	/* send info to all scripts about event */
	for (lua = lua_list.next; lua != &lua_list; lua = lua->next) {
		if (!(lua->events & (1 << event)))
			continue;

		lua_settop (lua->l, 0);				// clear stack
	
		lua_pushstring (lua->l, lua_eventnames[event]);
		lua_gettable (lua->l, LUA_GLOBALSINDEX);

	
		/* call funtion. */
		result = lua_pcall (lua->l, 0, 1, 0);

		if (result) {
			char *error = (char *) luaL_checkstring (lua->l, 1);

			printf ("> WARNING -> Lua -> %s\n", error);
		}
		
		/* retrieve return value */
		if (lua_isboolean (lua->l, -1)) {
			result = lua_toboolean (lua->l, -1);
			lua_remove (lua->l, -1);
		}
	}
#endif
	return 1;
}

#ifdef LUA
void open_libs(lua_State *L) {
	luaopen_io (L);
	luaopen_base (L);
	luaopen_table (L);
	luaopen_string (L);
	luaopen_math (L);
	luaL_openlibs (L);
}

inline int execute_program (lua_State *L) {
	// this function makes the code self-documenting
	return lua_pcall (L, 0, LUA_MULTRET, 0);
}

int report_errors (user_t *u, lua_State *L, const int status)
{
	if (status) {
		char *err = (char *) lua_tostring (L, -1);
		
		if (u)
			bot_to_user (u, err);
		else
			printf ("> WARNING -> Lua -> %s\n", err);

		lua_pop (L, 1); // remove error message
	
		return -1;
	}
	return 0;
}


int lua_load_script (user_t* u, char *name)
{
	lua_State *L = lua_open ();	/* initialize lua */
	
	if (!L)
		return 0;
	
	/* older version of Lua use open_libs () */
	luaL_openlibs (L);			/* open lua libraries */

 	lua_register_functions (L);	/* register lua commands */

	int s = luaL_loadfile (L, name);
	
	if (!s)
		s = execute_program (L);
	
	int ret = report_errors (u, L, s);

	lua_context_t *ctx;
    	/* alloc and init context */
	ctx = (lua_context_t *) malloc (sizeof (lua_context_t));
	
	if (!ctx)
		return 0;

	ctx->events = 0;

	int i;
	for (i = 0; lua_eventnames[i] != NULL; i++) {
		lua_pushstring (L, lua_eventnames[i]);
    		lua_gettable (L, LUA_GLOBALSINDEX);

		if(!lua_isnil(L, -1))
			ctx->events |= (1 << i);
	}

	ctx->l = L;
	ctx->name = strdup (name);
	
	if (!ctx->name)
		return 0;
	
	/* add into list */
	ctx->next = &lua_list;
	ctx->prev = lua_list.prev;
	ctx->prev->next = ctx;
	ctx->next->prev = ctx;

	/* add global variables */
	/* hub version */
	lua_pushstring (ctx->l, "LamaHub");
	lua_pushstring (ctx->l, VERSION);
	lua_settable (ctx->l, LUA_GLOBALSINDEX);
	
	/* reset stack */
	lua_settop (L, 0);

        // lua_close(L) automatically called here

	if (ret < 0)
		return lua_close_script (u, ctx);

    	return 1;
}

int lua_close_script (user_t *u, lua_context_t *ctx)
{
	lua_close (ctx->l);
	free (ctx->name);
	
	ctx->next->prev = ctx->prev;
	ctx->prev->next = ctx->next;
	
	free (ctx);

	lua_context_count --;

    	return 1;
}

/*************************************************************\
|		   HANDLERS OF LUA COMMANDS		      |
\*************************************************************/

bool handler_luaload (user_t *u, char *data)
{
	unsigned short x = 0;
	char a[64], b[192]; 
	b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[128];
		sprintf (use, "Use:\n%s <script file>", a);

		bot_to_user (u, use);
		return 0;
	}

	lua_context_t *lua;

	/* lua script is loaded ? */
	for (lua = lua_list.next; lua != &lua_list; lua = lua->next) {
		if (!strcmp (b, lua->name)) {
			char err[128];

			sprintf (err, "Script '%s' is already loaded.", b);

			bot_to_user (u, err);
			x ++;
			break;
		}
	}

	if (!x) {
		int ret = lua_load_script (u, b);
	
		if (ret) {
			char str[128];

			sprintf (str, "Script '%s' was loaded.", b);
	
			bot_to_user (u, str);
		} else {
			char str[128];

			sprintf (str, "Script '%s' contain some errors.", b);

			bot_to_user (u, str);
	
			return 0;
		}
	}

	return 0;
}

bool handler_luaclose (user_t *u, char *data)
{
	unsigned short x = 0;
	char a[64], b[192]; 
	b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[128];
		sprintf (use, "Use:\n%s <script file>", a);
		bot_to_user (u, use);
		return 0;
	}

	lua_context_t *lua;

	/* lua script is loaded ? */
	for (lua = lua_list.next; lua != &lua_list; lua = lua->next) {
		if (!strcmp (b, lua->name)) {
			x ++;
			break;
		}
	}

	if (!x) {
		char str[128];

		sprintf (str, "Script '%s' does not exists.", b);

		bot_to_user (u, str);
	} else {
		int ret = lua_close_script (u, lua);

		if (ret) {
			char str[128];

			sprintf (str, "Script '%s' was closed.", b);
	
			bot_to_user (u, str);
		}
	}

	return 0;
}

bool handler_lualist (user_t *u, char *data)
{
	lua_context_t *lua;
	char str[2048];
	unsigned short i = 0;

	strcpy (str, "\n--------------------- [ LUA LIST ] ---------------------");
	strcat (str, "\nLua version: 5.1\nScripts loaded (count): ");
	char info[8];

	char str2[1024];
	char id[8];
	strcpy (str2, " ");
	
	for (lua = lua_list.next; lua != &lua_list; lua = lua->next) {
		i ++;
		sprintf (id, "%d.", i);
		strcat (str2, "\n"); strcat (str2, id); strcat (str2, " "); strcat (str2, lua->name);
	}

	sprintf (info, "%d\n", i);
	strcat (str, info);

	strcat (str, str2);

	if (!i)
		strcat (str, "\nNo lua script loaded.");

	strcat (str, "\n------------------------------------------------------");

	bot_to_user (u, str);

	return 0;
}

/* handler of lua created commands */
bool handler_lua_func_cmd (user_t *u, char *data)
{
	lua_context_t *lua;
	int result = 0;

	int len = strlen (data);
	while (result != len) {
		if (data[result] == ' ' || data[result] == '\0')
			break;

		result ++;
	}
	len = result;

	char cmd[64];
	memcpy (cmd, "EventCmd", 8);
	memcpy (cmd+8, data, len);
	cmd[8+len] = '\0';

	result = 0;

	/* send info to all scripts about event */
	for (lua = lua_list.next; lua != &lua_list; lua = lua->next) {
		//if (!(lua->events & (1 << event)))
		//	continue;

		lua_settop     (lua->l, 0);				// clear stack
	
		lua_pushstring (lua->l, cmd);
		lua_gettable   (lua->l, LUA_GLOBALSINDEX);
	
		lua_pushstring (lua->l, u->nick);

		lua_pushstring (lua->l, data+len+1);
	
		/* call funtion. */
		result = lua_pcall (lua->l, 2, 1, 0);

		/*if (result) {							// FIXME
			char *error = (char *) luaL_checkstring (lua->l, 1);
	
			string err;
			       err += "Lua script return ERROR at event (";
			       err += lua_eventnames[event];
			       err += "): ";
			       err += error;
	
			bot_to_user (u, err);
		}*/
		
		/* retrieve return value */
		if (lua_isboolean (lua->l, -1)) {
			result = lua_toboolean (lua->l, -1);
			lua_remove (lua->l, -1);
		}
	}

	return 0;
}

/*************************************************************\
|		      LUA LOAD/SAVE SCRIPTS		      |
\*************************************************************/

int lua_list_load (char *file)
{
	unsigned int i;
	FILE *fp;
	char buffer[10240];
	
	/* load scripts */
	fp = fopen (file, "r+");

	if (!fp) {
		printf ("> ERROR -> Lua -> File %s dont found\n", file);
		return 0;
	}
	
	fgets (buffer, sizeof (buffer), fp);
	while (!feof (fp)) {
		for (i = 0; buffer[i] && buffer[i] != '\n' && (i < sizeof (buffer)); i++);
			if (i == sizeof (buffer))
				break;
	
		if (buffer[i] == '\n')
			buffer[i] = '\0';
		
		lua_load_script (NULL, buffer);
	
		fgets (buffer, sizeof (buffer), fp);
	}

	fclose (fp);

	lua_context_count ++;
	lua_context_peak ++;

	return 1;
}

int lua_list_save (char *file)
{
	FILE *fp;
	lua_context_t *ctx;
	
	fp = fopen (file, "w+");

	if (!fp) {
		printf ("> ERROR -> Lua -> Can't save file %s\n", file);
		return 0;
	}
	
	for (ctx = lua_list.next; (ctx != &lua_list); ctx = ctx->next)
		fprintf (fp, "%s\n", ctx->name);
	
	fclose (fp);	

	return 1;
}

/*************************************************************\
|		   	  REMOTE CALL			      |
\*************************************************************/

int luascript_save ()
{
	char filename [] = (DEFAULT_LUAFILE);

	int ret = lua_list_save (filename);

	if (!ret)
		return 0;

	return 1;
}

/*************************************************************\
|		   	    LUA INIT			      |
\*************************************************************/

int init_lua ()
{
	char filename [] = (DEFAULT_LUAFILE);

	lua_list.next = &lua_list;
	lua_list.prev = &lua_list;
	lua_list.l = NULL;
	lua_context_count = 0;
	lua_context_peak = 0;

	timer.active = false;

	/* load last loaded scripts at start */
	lua_list_load (filename);

	command_register ("luaload", "Load Lua script.", &handler_luaload , 4);
	command_register ("luaclose", "Close Lua script.", &handler_luaclose , 4);
	command_register ("lualist", "List of active Lua scripts.", &handler_lualist , 2);

	return 1;
}
#endif
