/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2009 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2010 ZeXx86 (zexx86@zexos.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#ifdef __WIN32__
 #include <windows.h>
 #include <wininet.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#endif
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "defaults.h"
#include "utils.h"
#include "config.h"
#include "commands.h"
#include "proto.h"

extern config_t config;
extern config_elem_t config_list;
extern stats_t stats;

/* the number of the next char which is the char c */
/* Maybe it's better with just strchr (buf) - buf */
int cut_string (char *buf, char c)
{
	int i = 0;

	while ((buf[i] != c) && (buf[i] != '\0'))
		i ++;

	if (buf[i] == c)
		return i;

	return -1;
}

int hublist_update (user_t *u)
{
  unsigned int port;
  int local_port;
#ifndef __WIN32__
  socklen_t host_len; 
#else
  int host_len;
#endif
  int buf_len;
  int bufp;
  int i, j, k;
  int s;
  int erret;
  int err = 0;
  int oldFlag;
  unsigned int uc, x = 1;
  char buf[400];
  char key[400];
  char str[1024];
  memcpy (str, "Hublists server state\n", 22);

  struct hostent *hostnm;
  struct sockaddr_in server;
  struct sockaddr_in host;

  char *lists, *l, *p;
  if (!config.hublist_hosts || !*config.hublist_hosts)
  	return -1;

  lists = strdup (config.hublist_hosts);

  for (l = strtok (lists, ";, "); l; l = strtok (NULL, ";, ")) {
	if (!strlen (l))
		continue;

	int works = 0;
	
	/* extract port */
	p = strchr (l, ':');
	
	if (p) {
		port = atoi (p + 1);
		*p = '\0';
	} else
		port = 2501;
	
	memset (&host, 0, sizeof(struct sockaddr_in));
	memset (&server, 0, sizeof(struct sockaddr_in));
	memset (buf, 0, sizeof(buf));
	memset (key, 0, sizeof(key));

	hostnm = gethostbyname (l); // "hublist.djinncity.net"

	if (hostnm == (struct hostent *) 0)
		goto resume;
	
	server.sin_family = AF_INET;
	server.sin_port = htons (port);
	server.sin_addr.s_addr = *((long unsigned *) hostnm->h_addr);
	
	if ((s = socket (AF_INET, SOCK_STREAM, 0)) < 0) {	
		//logprintf(1, "Error - In hublist_update (): Could not get a socket\n");
		close (s);
		goto resume;
	}

#ifdef __WIN32__
	/* Set to nonblocking socket mode */
	{	// needed, because warning with crosses goto.
		u_long arg = 1;
		// Operation is  FIONBIO. Parameter is pointer on non-zero number.
		if (ioctlsocket(s, FIONBIO, &arg) == SOCKET_ERROR)
		{
			printf ("> ERROR -> Problem with set socket mode\n");
			return -1;
		}
	}
#else
	/* Set to nonblocking socket mode */
	oldFlag = fcntl(s, F_GETFL, 0);
	if (fcntl(s, F_SETFL, oldFlag | O_NONBLOCK) == -1)
	{
		printf ("> ERROR -> Problem with set socket mode\n");
		return -1;
	}
#endif

	connect (s, (struct sockaddr *)&server, sizeof(server));

	/* we can't read/write to socket until is'nt ready */
#ifdef __WIN32__
	Sleep (1);
#else
	usleep (128);						// FIXME, better use select func
#endif

	while (((erret = recv (s, buf, sizeof(buf), 0)) < 0)) {
		err ++;

		if (err > 256)
			goto resume;

#ifdef DEBUG
		perror ("> WARNING -> hublist recv :");
#endif
#ifdef __WIN32__
		Sleep (1);
#else
		usleep (256);
#endif
	}

	if(erret < 0)
	{
		printf ("Error - In hublist_update (): Receive failed.\n");
		close (s);
		goto resume;
	}

	host_len = sizeof(host);
#ifndef __zexos__
	if(getsockname(s, (struct sockaddr*)&host, &host_len)) {
		close (s);
		goto resume;
	}
#endif
	local_port = htons (host.sin_port);
	uc = local_port + (local_port >> 8);
	
	if(strncmp(buf, "$Lock ", 6)) {
		close (s);
		goto resume;
	}
	
	sprintf (key, "$Key ");
	bufp = 6;
	
	buf_len = cut_string (buf+6, ' ') + 5;

	/* Now, compute the key from the lock */
	
	/* The first character is computed differently */
	i = (((unsigned int)(buf[bufp])) & 0xff)
	^ (((unsigned int)(buf[buf_len])) & 0xff)
	^ (((unsigned int)(buf[buf_len-1])) & 0xff)
	^ uc;
	
	j = ((i | (i << 8)) >> 4) & 0xff;
	
	switch(j) {
		case 5:
			sprintfa (key, "/%%DCN005%%/");
			break;
			
		case 36:
			sprintfa (key, "/%%DCN036%%/");
			break;
			
		case 96:
			sprintfa (key, "/%%DCN096%%/");
			break;
			
		default:
			sprintfa (key, "%c", j);
			break;
	}
	
	bufp ++;
	
	for (k = bufp; k <= buf_len; k ++) {
		i = (((unsigned int)(buf[k]     ))&0xff)
		^ (((unsigned int)(buf[k-1]   ))&0xff);
		
		j = ((i | (i << 8)) >> 4)&0xff;
		
		switch (j) {
			case 5:
				sprintfa(key, "/%%DCN005%%/");
				break;
			
			case '$':
				sprintfa(key, "/%%DCN036%%/");
				break;
		
			case 96:
				sprintfa(key, "/%%DCN096%%/");
				break;
			
			default:
				sprintfa(key, "%c", j);
				break;
		}
	}
	
	sprintfa (key, "|");
	
	/* The listening port only needs to be uploaded if it's not 411 because
	 * the Windoze client defaults to port 411 */

	sprintfa(key, "%s|%s:%d|%s|%d|%llu|", config.hubname, config.hubaddress,
		config.port, config.hubdesc, stats.currusers, config.minshare[0]);

	/* And finally, upload the key */
	int size, key_len;
	key_len = strlen (key) + 1; 

	if ((size = send (s, key, key_len, 0)) != key_len) {
#ifndef __WIN32__ 
		printf ("> WARNING -> %s : %d\n", strerror (errno), s);
#endif
		close(s);
		goto resume;
	}

	close(s);

	works = 1;

resume:

	if (u == NULL)
		continue;

	char line[128];

	if (works)
		sprintf (line, "%d.\t%s  :  OK\n", x, l);
	else
		sprintf (line, "%d.\t%s  :  Failed\n", x, l);

	strcat (str, line);

	x ++;
   }

   if (u)
   	bot_to_user (u, str);

   free (lists);

   return 1;
}

/*************************************************************\
|		     COMMANDS OF HUBLIST		      |
\*************************************************************/

bool handler_hublist (user_t *u, char *data)
{
	if (!hublist_update (u))
		bot_to_user (u, "No hublist hosts configured\n");

	return 0;
}

/*************************************************************\
|		  INITIALIZATION OF HUBLIST		      |
\*************************************************************/

int init_hublist ()
{
	command_register ("hublist", "Update data to hublist", &handler_hublist , 4);

	return 1;
}
