/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2009 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2010 ZeXx86 (zexx86@zexos.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#ifdef __WIN32__
 #include <windows.h>
#else
 #include <netinet/in.h>
 #ifndef __zexos__
 #ifndef __android__
  #include <crypt.h>
 #else
  #include <jni.h>
  #include <android/log.h>
 #endif
 #endif
#endif
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>

#include "commands.h"
#include "user.h"
#include "utils.h"
#include "config.h"
#include "sockets.h"
#include "trigger.h"
#include "defaults.h"
#include "ban.h"
#include "chatroom.h"
#include "elua.h"
#include "hublist.h"
#include "plugin.h"
#include "user_rights.h"

int cont;		      // 1 = run / 0 = shutdown

extern user_t user_list;
extern command_t command_list;
extern config_t config;
extern account_t accounts_list;
extern config_elem_t config_list;

extern int port;                     // ID of port
extern int mainSocket;               // Main socket


/*************************************************************\
|		   	ARGUMENTS		              |
\*************************************************************/

#ifndef __WIN32__
#ifndef __zexos__
#ifndef __android__
/* This function was taken from aquila hubsoft and was modified */
void daemonize ()
{
	int i, lockfd;
  	char pid[16];

	int ipid = getpid ();

  	/* detach if asked */
  	if (ipid == 1)
		return;			/* already a daemon */

    	/* fork to guarantee we are not process group leader */
    	i = fork ();
    	if (i < 0)
      		exit (1);			/* fork error */
    	if (i > 0)
      		exit (0);			/* parent exits */

    	/* child (daemon) continues */
    	setsid ();			/* obtain a new process group */
	ipid = getpid ()+1;

	printf ("> started with pid -> %d\n", ipid);
    	/* fork again so we become process group leader 
     	 * and cannot regain a controlling tty 
     	 */
    	i = fork ();
    	if (i < 0)
      		exit (1);			/* fork error */
    	if (i > 0)
      	exit (0);			/* parent exits */

    	/* close all fds */
    	for (i = getdtablesize (); i >= 0; --i)
      		close (i);		/* close all descriptors */



    	/* close parent fds and send output to fds 0, 1 and 2 to bitbucket */
    	i = open ("/dev/null", O_RDWR);
    	if (i < 0)
      		exit (1);
    	dup (i);
    	dup (i);			/* handle standart I/O */

  	/* create local lock */
  	lockfd = open ("lamahub.pid", O_RDWR | O_CREAT, 0640);
  	if (lockfd < 0) {
    		perror ("lock: open");
    		exit (1);
  	}
	#ifndef __CYGWIN__
	/* lock the file */
	if (lockf (lockfd, F_TLOCK, 0) < 0) {
		perror ("lock: lockf");
		printf (HUBSOFT " is already running.\n");
		exit (0);
	}
	#else
	/* lock the file */
	{
	struct flock lock;
		lock.l_type = F_RDLCK;
		lock.l_start = 0;
		lock.l_whence = SEEK_SET;
		lock.l_len = 0;
		
		if (fcntl (lockfd, F_SETLK, &lock) < 0) {
			printf (HUBSOFT " is already running.\n");
			exit (0);
		}
	}
	#endif
	/* write to pid to lockfile */
	snprintf (pid, 16, "%d\n", getpid ());
	write (lockfd, pid, strlen (pid));
	


	/* restrict created files to 0750 */
	umask (027);
}
#endif
#endif
#endif

int arguments (int argc, char *argv[])
{
	unsigned int i = 0;
	char str[256];
	
	if (argc == 1)
		return 0;

	if (argc == 2) {
#ifndef __WIN32__
#ifndef __zexos__
#ifndef __android__
		if (!strcmp (argv[1], "-d")) {
			daemonize ();
			return 0;
		}
#endif
#endif
#endif
		if (!strcmp (argv[1], "-a")) {
			char name [128], passwd [128];
			printf ("Owner account\nYour owner name: ");
			scanf ("%s", name);
			printf ("Password: ");
			scanf ("%s", passwd);

			sprintf (str, "regnum 1\n43520 44544 306689 306701 323149 323165\n%s %s 5", name, crypt (passwd, HUBSOFT));

			i ++;
		}
	}
	if (argc == 3) {
		sprintf (str, "regnum 1\n43520 44544 306689 306701 323149 323165\n%s %s 5", argv[1], crypt (argv[2], HUBSOFT));

		i ++;
	}

   	if (!i) {
		printf ("> ERROR -> Syntax (create new owner account):\n%s <nick> <password>\n", argv[0]);

        	return -1;
        }

	FILE *fp;

	char filename[80];
	strcpy (filename, DEFAULT_ACCOUNTSFILE);

	fp = fopen (filename, "w+");

	if (!fp)
		return -2;
	
	fprintf (fp, "%s", str);

  	fclose (fp);

	printf ("\n** Your Owner account was created. You can start hub now ! **\n");

	return 1;
}

/*************************************************************\
|		   	APP SIGNALS		              |
\*************************************************************/

void alarm_signal (int z)
{
	//printf ("> WARNING -> alarm signal -> type %d\n", z);
	//usleep (1000);
}

void term_signal (int z)
{
	printf ("> WARNING -> term signal -> type %d\n", z);

	usleep (1024);

	// close all sockets
	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
		close (user->socket);
		//free (user);
	}

	/*config_elem_t *cfg;		// FIXME memleak in exit (calloc not free'ied)
	for (cfg = config_list.next; cfg != &config_list; cfg = cfg->next)
		if (cfg->value)
			free (cfg->value);*/

	cont = 0;	// Bye Bye :)
}

/*************************************************************\
|		   INITIALIZATION OF APP		      |
\*************************************************************/

int init_sig (void)
{
#ifndef __WIN32__
#ifndef __zexos__
   struct sigaction sv;  
   
   memset(&sv, 0, sizeof(struct sigaction));
   sv.sa_flags = 0;
   sigemptyset(&sv.sa_mask);
/*#ifdef SA_NOCLDWAIT
   sv.sa_flags |= SA_NOCLDWAIT;
#endif
#ifdef SA_NOCLDSTOP
   sv.sa_flags |= SA_NOCLDSTOP;
#endif*/
   
   sv.sa_handler = SIG_IGN;
   /* Don't want broken pipes to kill the hub.  */
   sigaction(SIGPIPE, &sv, NULL);
   
   /* ...or any defunct child processes.  */
   sigaction(SIGCHLD, &sv, NULL);
   
   sv.sa_handler = term_signal;
   
   /* Also, shut down properly.  */
   sigaction(SIGTERM, &sv, NULL);
   sigaction(SIGINT, &sv, NULL);
   
   sv.sa_handler = alarm_signal;
   
   /* And set handler for the alarm call.  */
   sigaction(SIGALRM, &sv, NULL);
#endif
#endif
   return 1;
}


int init ()
{
	int ret = 0;

	ret += init_config ();
	ret += init_sig ();
	ret += init_commands ();
	ret += init_accounts ();
	ret += init_trigger ();
	ret += init_ban ();
	ret += init_chatroom ();
	ret += init_hublist ();
	ret += init_plugin ();
	ret += init_rights ();
#ifdef LUA
	ret += init_lua ();
#endif	

	char configfile [] = (DEFAULT_HUBFILE);

	/* We need load config at last, after initialization of all */
	config_load (configfile);

	/* Create main socket */
	if (init_socket () < 1)
		return 0;

	if (ret <= 0) {		// ERROR in initialization ??
		printf ("> ERROR -> init ()\n");
		ret = 1;
        }
        
        cont = 1;

	return ret;
}

/*************************************************************\
|		   	MAIN THREAD		              |
\*************************************************************/

int main (int argc, char *argv[])
{
   if (arguments (argc, argv))
	exit (1);

   if (init () <= 0)
	cont = 0;

   /* MAIN LOOP */
   for ( ; cont ; ) {
	loop ();
#ifdef __zexos__
	schedule ();
#endif
   }

   printf ("> Shutdown\n");
   
   close (mainSocket);

   return 0;
}


/*************************************************************\
|		   	ANDROID BINDINGS	              |
\*************************************************************/

#ifdef __android__
JNIEXPORT jstring JNICALL Java_lamahub_zexx86_lamahub_getconfig (JNIEnv *env, jclass c, jstring jvar)
{
	jboolean iscopy;
	const jbyte *var = (*env)->GetStringUTFChars (env, jvar, &iscopy);
	
	return (*env)->NewStringUTF (env, config_get ((char *) var));
}

JNIEXPORT int JNICALL Java_lamahub_zexx86_lamahub_init (JNIEnv *env, jclass c)
{
	__android_log_print (ANDROID_LOG_INFO, "LamaHub", "init");

	/* create necessary directory for lamahub files for Android */
	mkdir ("/sdcard/lamahub");

	/* check for accounts.conf */
	int fd = open (DEFAULT_ACCOUNTSFILE, 0);
	
	if (fd < 0) {
		char *argv[] = {
			0,
			"owner",
			"owner"
		};
		
		__android_log_print (ANDROID_LOG_INFO, "LamaHub", "creating owner account '%s' with password '%s'", argv[1], argv[2]);

		if (arguments (3, argv) != 1)
			__android_log_print (ANDROID_LOG_INFO, "LamaHub", "problem in creating account");
	} else
		close (fd);
	
	return init ();
}

JNIEXPORT void JNICALL Java_lamahub_zexx86_lamahub_loop (JNIEnv *env, jclass c)
{
	__android_log_print (ANDROID_LOG_INFO, "LamaHub", "loop");

	for ( ; cont ; )
		loop ();
	
	close (mainSocket);
	
	__android_log_print (ANDROID_LOG_INFO, "LamaHub", "end of loop");
}

JNIEXPORT void JNICALL Java_lamahub_zexx86_lamahub_exit (JNIEnv *env, jclass c)
{
	cont = 0;

	__android_log_print (ANDROID_LOG_INFO, "LamaHub", "exit");
}

JNIEXPORT int JNICALL Java_lamahub_zexx86_lamahub_status (JNIEnv *env, jclass c)
{
	__android_log_print (ANDROID_LOG_INFO, "LamaHub", "status");
	
	return cont;
}

int JNI_OnLoad (JavaVM *vm, void *reserved)
{
	__android_log_print (ANDROID_LOG_INFO, "LamaHub", "start");
	
	JNIEnv *env;
	if ((*vm)->GetEnv (vm, (void **) &env, JNI_VERSION_1_4) != JNI_OK) {
		__android_log_print (ANDROID_LOG_ERROR, "LamaHub", "Failed to get the environment using GetEnv()");
		return -1;
	}
	
	return JNI_VERSION_1_4;
}
#endif
