/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2009 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2010 ZeXx86 (zexx86@zexos.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifdef __WIN32__
 #include <windows.h>
#else
 #ifndef __zexos__
  #include <dlfcn.h>
 #endif
#endif
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "commands.h"
#include "plugin.h"
#include "config.h"
#include "utils.h"
#include "proto.h"
#include "ban.h"

plugin_t plugin_list;

extern proto_t nmdc_proto;
extern config_t config;
extern user_t user_list;
extern ban_t ban_list;
extern ban_t ban_list;
extern long long int sec;		     // Uptime


/*************************************************************\
|		       PLUGIN FUNCTIONS	  		      |
\*************************************************************/


// FIXME crash
user_t *next = NULL;
user_t *plugin_user_next ()
{
  next = next->next;

  if (!next) {
	next = NULL;
	return NULL;
  }

  return next;
}

void event_plugin (user_t *u, char *txt, int state, unsigned len)
{
	plugin_t *ctx;
	for (ctx = plugin_list.next; ctx != &plugin_list; ctx = ctx->next) {
		if (ctx->event_handler_e)
			ctx->event_handler (u->nick, txt, state, len);
	}
}

int plugin_botadd (char *nick, char *desc, char *conn, char *mail, unsigned short iclass, unsigned long long share)
{
	user_t *t = NULL;
	t = user_findbynick (nick);

	if (t)
		return 0;

	/* create virtual context */
	user_t *u;	
		
	// alloc and init context
	u = (user_t*) malloc (sizeof (user_t));

	u->state = PROTO_STATE_ONLINE;
	u->socket = 0;
	u->iclass = iclass;
	u->share = 0;
	u->searchlast = 0;
	u->flag = 0;
	u->online = 1;
	u->buf_len = 32;
	u->buf = (char *) calloc (32, sizeof(char));
	u->buf_offset = 0;
	u->buf_used = 0;
	u->nick_len = strlen (nick);

	strcpy (u->nick, nick);
	strcpy (u->ip, "127.0.0.1");

	sprintf (u->myinfo, "$MyINFO $ALL %s %s$ $%s$%s$%llu$|", nick, desc, conn, mail, share);
	u->myinfo_len = strlen (u->myinfo);

	// add into list  
	u->next = &user_list;
	u->prev = user_list.prev;
	u->prev->next = u;
	u->next->prev = u;
	
	/* send to all users virtual myinfo */
	send_to_all (u->myinfo);

	return 1;
}

short plugin_userclass (char *nick)
{
	user_t *u = NULL;
	u = user_findbynick (nick);

	if (!u)
		return 0;

	return (signed) u->iclass;
}

char *plugin_userip (char *nick)
{
	user_t *u = NULL;
	u = user_findbynick (nick);

	if (!u)
		return NULL;

	return u->ip;
}

int plugin_sendtouser (char *nick, char *txt)
{
	user_t *u = NULL;
	u = user_findbynick (nick);

	if (!u)
		return 0;

	send_to_user (u, txt);

	return 1;
}

int plugin_bottouser (char *nick, char *txt)
{
	user_t *u = NULL;
	u = user_findbynick (nick);

	if (!u)
		return 0;

	bot_to_user (u, txt);

	return 1;
}

int plugin_pmtouser (char *target, char *nick, char *txt)
{
	user_t *tgt = NULL;
	tgt = user_findbynick (target);

	if (!tgt)
		return 0;

	pm_to_user (tgt, nick, txt);

	return 1;
}


int plugin_bannick (char *nick, char *time)
{
	user_t *u = NULL;
	u = user_findbynick (nick);

	if (!u)
		return 0;

	ban_add_bynick (u, time);
	proto_nmdc_user_redirect (u);

	return 1;
}

int plugin_banip (char *ip, char *time)
{
	user_t *u = NULL;
	u = user_findbyip (ip);

	if (!u)
		return 0;

	ban_add_byip (u, time);
	proto_nmdc_user_redirect (u);

	return 1;
}

int plugin_banperm (char *ip)
{
	user_t *u = NULL;
	u = user_findbyip (ip);

	if (!u)
		return 0;

	ban_add_byip (u, "0");
	proto_nmdc_user_redirect (u);

	return 1;
}

int plugin_bandel (char *val)
{
	ban_t *ban;
	unsigned short i = 0;

	for (ban = ban_list.next; ban != &ban_list; ban = ban->next) {
		if (!strcmp (ban->nick, val)) {
			ban_del (ban);
			i ++;
			break;
		}
		if (!strcmp (ban->ip, val)) {
			ban_del (ban);
			i ++;
			break;
		}
	}

	if (i)
		return 1;

	return 0;
}

int plugin_kick (char *nick)
{
	user_t *u = NULL;
	u = user_findbynick (nick);

	if (!u)
		return 0;

	ban_add_bynick (u, "120");
	proto_nmdc_user_redirect (u);

	return 1;
}

int plugin_chatlock ()
{
	if (!config.deflood.mainchat_lock) {
		config.deflood.mainchat_lock = 1;
		return 1;
	} else
		config.deflood.mainchat_lock = 0;

	return 0;
}

int plugin_pmlock ()
{
	if (!config.deflood.pm_lock) {
		config.deflood.pm_lock = 1;
		return 1;
	} else
		config.deflood.pm_lock = 0;

	return 0;
}

unsigned long long plugin_uptime ()
{
	return sec;
}

/*************************************************************\
|		       PLUGIN LOAD/DEL	  		      |
\*************************************************************/

int plugin_add (char *name, void *handle)
{
	// first verify, that plugin is'nt already loaded
	plugin_t *ctx;
	for (ctx = plugin_list.next; ctx != &plugin_list; ctx = ctx->next) {
		if (!strcmp (ctx->name, name))
			return 0;
	}

	// alloc and init context
	ctx = (plugin_t*) malloc (sizeof (plugin_t));
	memset (ctx, 0, sizeof (ctx));

	strcpy (ctx->name, name);

	ctx->handle = handle;

	ctx->next = &plugin_list;
	ctx->prev = plugin_list.prev;
	ctx->prev->next = ctx;
	ctx->next->prev = ctx;

#ifndef __WIN32__
#ifndef __zexos__
	typedef int (*plugin_event_t) (char *nick, char *txt, int state, unsigned len);
		
	plugin_event_t event_handler = (plugin_event_t) dlsym(handle, "event_handler");
	const char *dlsym_error = dlerror ();
	
	if (dlsym_error) {	// it is'nt needed
		printf ("> Plugin -> WARNING -> Cannot load symbol 'event_handler': %s\n", dlsym_error);

		ctx->event_handler_e = 0;
	} else {
		ctx->event_handler = *(event_handler);
		ctx->event_handler_e = 1;
	}
#endif
#endif

	return 1;
}

int plugin_del (char *name)
{
#ifndef __WIN32__
#ifndef __zexos__
	// first verify, that plugin is loaded
	plugin_t *ctx;
	for (ctx = plugin_list.next; ctx != &plugin_list; ctx = ctx->next) {
		if (!strcmp (ctx->name, name)) {
			// load the symbol
			typedef void (*plugin_t) ();
		
			plugin_t destroy = (plugin_t) dlsym(ctx->handle, "destroy");
			const char *dlsym_error = dlerror ();

			if (dlsym_error)	// it is'nt needed
				printf ("> Plugin -> WARNING -> Cannot load symbol 'destroy': %s\n", dlsym_error);
			else
				destroy ();

			ctx->next->prev = ctx->prev;
			ctx->prev->next = ctx->next;

			printf ("> Plugin -> %s closed\n", name);

			return 1;
		}
	}
#endif
#endif
	return 0;
}

unsigned short plugin_load (char *name)
{
#ifndef __WIN32__
#ifndef __zexos__
	// open the library
	void* handle = dlopen (name , RTLD_LAZY);
	
	if (!handle) {
		printf ("> Plugin -> WARNING -> Cannot open library: %s\n", dlerror ());
		return 0;
	}
	
	// load the symbol
	typedef int (*plugin_load_t) (proto_t func);
	
	// reset errors
	dlerror();
	plugin_load_t init = (plugin_load_t) dlsym(handle, "init");
	const char *dlsym_error = dlerror();

	if (dlsym_error) {
		printf ("> Plugin -> WARNING -> Cannot load symbol 'init': %s\n", dlsym_error);

		dlclose (handle);
		return 2;
	}

	// does plugin already exists ?
	int ret = plugin_add (name, handle);

	if (!ret)
		return 3;

	/* call init function in library */
	if (init (nmdc_proto) == 0) {
		dlclose (handle);
		return 0;
	}

	printf ("> Plugin -> %s loaded\n", name);

	//dlclose (handle);
#endif
#endif

	return 1;
}

/*************************************************************\
|		      PLUGIN LOAD/SAVE FILE		      |
\*************************************************************/

int plugin_list_load (char *file)
{
	unsigned int i;
	FILE *fp;
	char buffer[10240];
	
	/* load scripts */
	fp = fopen (file, "r+");

	if (!fp) {
		printf ("> ERROR -> Plugin -> File %s dont found\n", file);
		return 0;
	}
	
	char *ret = fgets (buffer, sizeof (buffer), fp);
	
	if (!ret) {
		fclose (fp);
		return 0;
	}
	
	while (!feof (fp)) {
		for (i = 0; buffer[i] && buffer[i] != '\n' && (i < sizeof (buffer)); i++);
			if (i == sizeof (buffer))
				break;
	
		if (buffer[i] == '\n')
			buffer[i] = '\0';

		if (strlen (buffer) < 3)
			break;
		
		if (plugin_load (buffer) != 1)
			continue;
	
		ret = fgets ((char *) buffer, sizeof (buffer), fp);
	}

	fclose (fp);

	return 1;
}

int plugin_list_save (char *file)
{
	FILE *fp;
	plugin_t *ctx;

	fp = fopen (file, "w+");

	if (!fp) {
		printf ("> ERROR -> Lua -> Can't save file %s\n", file);
		return 0;
	}
	
	for (ctx = plugin_list.next; ctx != &plugin_list; ctx = ctx->next)
		fprintf (fp, "%s\n", ctx->name);
	
	fclose (fp);

	return 1;
}

int plugin_save ()
{
	char filename [] = (DEFAULT_PLUGINFILE);

	return plugin_list_save (filename);
}


/*************************************************************\
|		    PLUGIN COMMAND HANDLER  		      |
\*************************************************************/

bool handler_plugin_load (user_t *u, char *data)
{
	char a[48], b[256];
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 3) {
		char use[128];
		
		if (strlen (a) > 110)
			return 0;

		sprintf (use, "Use:\n%s <plugin>", a);

		bot_to_user (u, use);

		return 0;
	}

	unsigned short ret = plugin_load (b);

	char str[256];
	str[0] = '\0';

	if (ret == 1)
		sprintf (str, "Plugin '%s' was loaded.", b);
	else if (ret == 3)
		sprintf (str, "Plugin '%s' is already loaded.", b);
	else if (ret == 0)
		sprintf (str, "Plugin '%s' does'nt exist or contain error.", b);
	else if (ret == 2)
		sprintf (str, "Plugin '%s' is incompatibile.", b);

	bot_to_user (u, str);

	return 1;
}

bool handler_plugin_close (user_t *u, char *data)
{
	char a[16], b[strlen(data)];
	a[0] = '\0'; b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen (b) < 1) {
		char use[128];
		
		if (strlen (a) > 110)
			return 0;

		sprintf (use, "Use:\n%s <plugin>", a);

		bot_to_user (u, use);

		return 0;
	}

	int ret = plugin_del (b);		// "./plugin/plugin_test.so"

	char str[strlen (b)+32];
	str[0] = '\0';

	if (ret) {
		//close the library
		printf ("Closing library...\n");

		sprintf (str, "Plugin '%s' was closed.", b);
	} else
		sprintf (str, "Plugin '%s' is'nt loaded.", b);

	bot_to_user (u, str);

	return 1;
}

bool handler_plugin_list (user_t *u, char *data)
{
	char *str = strdup ("Plugin list");

	unsigned short i = 0;
	unsigned p = 11;

	plugin_t *ctx;
	for (ctx = plugin_list.next; ctx != &plugin_list; ctx = ctx->next) {
		i ++;

		char line[128];

		sprintf (line, "\n%d. %s", i, ctx->name);

		unsigned l = strlen (line);

		str = (char *) realloc ((void *) str, l+1);

		if (!str)
			return 0;

		memcpy (str+p, line, l);
		str[p] = '\0';
		
		p += l;
	}

	if (!i) {
		bot_to_user_len (u, (char *) "No plugin loaded.", 17);

		return 1;
	}

	bot_to_user_len (u, str, p);

	free (str);

	return 1;
}

/*************************************************************\
|		    	 PLUGIN INIT     	 	      |
\*************************************************************/

int init_plugin ()
{
	plugin_list.next = &plugin_list;
	plugin_list.prev = &plugin_list;

	char filename [] = (DEFAULT_PLUGINFILE);

#ifndef __WIN32__
#ifndef __zexos__
	plugin_list_load (filename);

	command_register ("pluginload", "Load a plugin.", &handler_plugin_load , 4);
	command_register ("pluginclose", "Close a plugin.", &handler_plugin_close , 4);
	command_register ("pluginlist", "Display a list of plugin.", &handler_plugin_list , 3);
#endif
#endif

	return 0;
}
