/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2009 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2010 ZeXx86 (zexx86@zexos.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "defaults.h"
#include "commands.h"
#include "user.h"
#include "utils.h"
#include "config.h"
#include "sockets.h"
#include "config.h"
#include "proto.h"
#include "ban.h"
#include "chatroom.h"
#include "ezlib.h"
#include "elua.h"
#include "hublist.h"
#include "plugin.h"
#include "user_rights.h"
#include "plugin.h"

/* create structures */
user_t user_list;
command_t command_list;
config_t config;
account_t accounts_list;
config_elem_t config_list;
stats_t	stats;
ban_t ban_list;
chatroom_t chatroom_list;
chatroom_public_t chatroom_data;
#ifdef LUA
timer_lua_t timer;
#endif
/* variables with other src files */
extern int cont;		     // 1 = run / 0 = shutdown
extern long long int autosave;	     // Autosave interval (sec)
extern long long int search_interval;// Search interval (sec)

/* global variables */
long long int sec = 0;		     // Uptime
long long int lastsave = 0;	     // Last time of save config
long long int lastupdate = 0;	     // Last time of update hublist

char *savefile = DEFAULT_HUBFILE;

/* constructor */
int proto_nmdc_state_disconnect (user_t *user);

/* NMDC Protocol <=> Plugin */
proto_t nmdc_proto = {
	/* send raw data */
	sendtoall:		send_to_all,
	sendtoall_len:		send_to_all_len,
	sendtouser:		plugin_sendtouser,
	sendtouser_len:		send_to_user_len,
	sendtosocket:		send_to_socket,

	/* nmdc data */
	bottouser:		plugin_bottouser,
	bottoall:		bot_to_all,
	pmtouser:		plugin_pmtouser,

	/* config's variable manipulation */
	configget:		config_get,
	configset:		config_set,
	configreg:		config_register,

	/* ban/kick */
	bannick:		plugin_bannick,
	banip:			plugin_banip,
	banperm:		plugin_banperm,
	bandel:			plugin_bandel,
	kick:			plugin_kick,

	/* message control */
	chatlock:		plugin_chatlock,
	pmlock:			plugin_pmlock,

	uptime:			plugin_uptime,

	/* redirect user */
	forcemove:		proto_nmdc_user_forcemove,

	/* command add/del */
	commandadd:		command_register,
	commanddel:		command_unregister,

	/* user's info */
	userip:			plugin_userip,
	userclass:		plugin_userclass,

	/* bot */
	botadd:			plugin_botadd,

	name:			"NMDC"
};

/********************************************************************/

short curr_redirect = 0;
int proto_nmdc_user_redirect (user_t *u)
{
	curr_redirect ++;
	if (curr_redirect >= stats.redirect_hosts)
		curr_redirect = 0;

	char str[128];
	sprintf (str, "$ForceMove %s|", config.redirect [curr_redirect]);

	send_to_user (u, str);

	u->state = PROTO_STATE_DISCONNECTED;

	stats.num_forcemove ++;

	return 1;
}

int proto_nmdc_user_forcemove (user_t *u, char *destination)
{
	char str[256];
	sprintf (str, "$ForceMove %s|", destination);

	send_to_user (u, str);

	u->state = PROTO_STATE_DISCONNECTED;

	stats.num_forcemove ++;

	return 1;
}

/********************************************************************/

int proto_nmdc_token_getnicklist (user_t *u, char *data)
{
	char *sNickList = (char *) malloc (1025);

	if (!sNickList)
		return 0;
	
	unsigned long iNickListSize = 1024;
	unsigned long iNickListLen = 0;
	
	char *sMyINFOs = (char *) malloc (4097);
	
	if (!sMyINFOs)
		return 0;

	unsigned long iMyINFOsSize = 4096;
	unsigned long iMyINFOsLen = 0;
	
	char *sOpList = (char *) malloc (513);

	if (!sOpList)
		return 0;


	unsigned long iOpListSize = 512;
	unsigned long iOpListLen = 0;

	memcpy (sNickList, "$NickList ", 10);
	memcpy (sNickList+10, config.botname, config.botname_len);
	memcpy (sNickList+10+config.botname_len, "$$", 2);
	iNickListLen = 12+config.botname_len;
	sNickList[iNickListLen] = '\0';

	memcpy (sOpList, "$OpList ", 8);
	memcpy (sOpList+8, config.botname, config.botname_len);
	memcpy (sOpList+8+config.botname_len, "$$", 2);
	
	if (chatroom_data.l)
		memcpy (sOpList+10+config.botname_len, chatroom_data.lists, chatroom_data.l);
	
	iOpListLen = 10+config.botname_len+chatroom_data.l;
	sOpList[iOpListLen] = '\0';

	user_t *user;
	for(user = user_list.next; user != &user_list; user = user->next) {
		unsigned int iLen = user->nick_len + 2;

		if (iLen < 5 || user->state != PROTO_STATE_ONLINE)			/* drop, nick is too short */
			continue;

		if(iNickListLen + iLen > iNickListSize) {
			sNickList = (char *) realloc ((void *) sNickList, iNickListSize+1025);
			if (!sNickList) {	 /* out of memory */
				free (sNickList);
				free (sMyINFOs);
				free (sOpList);
				return 0;
			}

			iNickListSize = iNickListSize+1024;
		}
		
		memcpy(sNickList+iNickListLen, user->nick, iLen-2);
		memcpy(sNickList+iNickListLen+(iLen-2), "$$", 2);
		iNickListLen += iLen;
		sNickList[iNickListLen] = '\0';

		if (u != user) {
			if (class_right_check (user, CAP_KEY)) {
				if(iOpListLen + iLen > iOpListSize) {
					sOpList = (char *) realloc ((void *) sOpList, iOpListSize+513);
					if (!sOpList) {  /* out of memory */
						free (sNickList);
						free (sMyINFOs);
						free (sOpList);
						return 0;
					}
				
					iOpListSize = iOpListSize+512;
				}
				memcpy (sOpList+iOpListLen, user->nick, iLen-2);
				memcpy (sOpList+iOpListLen+(iLen-2), "$$", 2);
				iOpListLen += iLen;
				sOpList[iOpListLen] = '\0';
			}

               		iLen = user->myinfo_len;
               		if (iMyINFOsLen + iLen > iMyINFOsSize) {
				sMyINFOs = (char *) realloc ((void *)sMyINFOs, iMyINFOsSize+4097);
				if (!sMyINFOs) {	/* out of memory */
					free (sNickList);
					free (sMyINFOs);
					free (sOpList);
					return 0;
				}

				iMyINFOsSize = iMyINFOsSize+4096;
			}

			memcpy (sMyINFOs+iMyINFOsLen, user->myinfo, iLen);
	
			iMyINFOsLen += iLen;
			sMyINFOs[iMyINFOsLen] = '\0';
			
		}
			
	}

	unsigned long iLen = iNickListLen+39+iMyINFOsLen+iOpListLen+config.botname_len+chatroom_data.len;

	char *sData = (char *) malloc (iLen+1);
	
	if (!sData) {
		free (sNickList);
		free (sMyINFOs);
		free (sOpList);
		return 0;
	}

	memcpy (sData, sNickList, iNickListLen);
	memcpy (sData+iNickListLen, "|", 1);
	memcpy (sData+iNickListLen+1, sMyINFOs, iMyINFOsLen);
	memcpy (sData+iNickListLen+1+iMyINFOsLen, sOpList, iOpListLen);
	memcpy (sData+iNickListLen+1+iMyINFOsLen+iOpListLen, "|", 1);
	// bot myinfo
	memcpy (sData+iNickListLen+2+iMyINFOsLen+iOpListLen, "$MyINFO $ALL ", 13);
	memcpy (sData+iNickListLen+15+iMyINFOsLen+iOpListLen, config.botname, config.botname_len);
	memcpy (sData+iNickListLen+15+iMyINFOsLen+iOpListLen+config.botname_len, " Hub-Security$ $BOT $$$|", 24);
	// chatrooms
	if (chatroom_data.len)
		memcpy (sData+iNickListLen+39+iMyINFOsLen+iOpListLen+config.botname_len, chatroom_data.myinfos, chatroom_data.len);

	// in the end we put null char
        sData[iLen] = '\0';

#ifdef ZPIPE
	if (u->flag & NMDC_SUPPORTS_ZPipe) {	// Client supports Zpipe
		char *out = (char *) malloc (iLen + 65);

		if (!out)
			return 0;

		/* send compressed data to user */
		unsigned long len = zdef (sData, iLen, out, iLen + 64);

		send_to_user_len (u, (char *) out, len);

		free (out);
	} else
		send_to_user_len (u, sData, iLen);
#else
	send_to_user_len (u, sData, iLen);
#endif

	/* now we can free buffers */
	free (sData);
	free (sNickList);
	free (sMyINFOs);
	free (sOpList);

	event_lua_hub (u, EVENT_GETNICKLIST);

	return 1;
}


/*************************************************************\
|		     	NMDC STATE - INIT		      |
\*************************************************************/

int proto_nmdc_state_init (user_t *u)
{
	/* send init text in connect */
	send_to_user_len (u, config.initstr, config.initlen);

	event_lua_hub (u, EVENT_NEWCLIENT);

	unsigned long bantime = 0;
	if ((bantime = ban_find_byip (u->nick)) != 0) {
		char str[64];
		if (bantime != 1)
			sprintf (str, "You are banned for %lu seconds !", bantime);
		else
			strncpy (str, "You are banned permanently !", 28);

		bot_to_user (u, str);

		proto_nmdc_user_redirect (u);

		return 0;
	}

	u->state = PROTO_STATE_SENDLOCK;

	return 1;
}

/*************************************************************\
|		     NMDC STATE - SENDLOCK		      |
\*************************************************************/

int proto_nmdc_state_sendlock (user_t *u, char *data, unsigned int len)
{
	char *k, *l;

	// check a nmdc command
	if (!strncmp (data, TOKEN_SUPPORTS, 9)) {
		/* move after nmdc cmd */
		data += 9;

		k = data;
		
		for (k = data; *k; k = l) {
		
			while (*k && (*k == ' '))
				k++;
		
			if (!*k)
				break;

			l = k + 1;
		
			while (*l && (*l != ' '))
				l++;
			
			/* check length too? */
			switch (*k) {
				/*case 'N':
					if (!strncmp (k, "NoGetINFO", 9)) {
						u->flag |= NMDC_SUPPORTS_NoGetINFO;
						continue;
					}
					if (!strncmp (k, "NoHello", 7)) {
						u->flag |= NMDC_SUPPORTS_NoHello;
						continue;
					}
					break;
				case 'U':
					if (!strncmp (k, "UserIP2", 7)) {
						u->flag |= NMDC_SUPPORTS_UserIP2;
						continue;
					}
					if (!strncmp (k, "UserCommand", 11)) {
						u->flag |= NMDC_SUPPORTS_UserCommand;
						continue;
					}
					break;
				case 'Q':
					if (!strncmp (k, "QuickList", 9)) {
						u->flag |= NMDC_SUPPORTS_QuickList;
						continue;
					}
					break;
				case 'T':
					if (!strncmp (k, "TTHSearch", 9)) {
						u->flag |= NMDC_SUPPORTS_TTHSearch;
						continue;
					}
					break;*/
				case 'Z':
					if (!strncmp (k, "ZPipe", 5)) {
						u->flag |= NMDC_SUPPORTS_ZPipe;
						stats.cache_zpipers ++;
						continue;
					}
					/*if (!strncmp (k, "ZLine", 5)) {
						u->flag |= NMDC_SUPPORTS_ZLine;
						continue;
					}*/
					break;
			}
		}

		send_to_user_len (u, "$Supports NoGetINFO BotINFO ZPipe|", 34);

		event_lua (u, data, EVENT_LOCK, len);

		return 1;
	} 
	else if (!strncmp(data, TOKEN_KEY, 4)) {
		// Key decode (not supported)
		u->state = PROTO_STATE_WAITNICK;
		return 1;
	}

	return 0;
}

/*************************************************************\
|		     NMDC STATE - WAITNICK		      |
\*************************************************************/

int proto_nmdc_state_waitnick (user_t *u, char *data, unsigned int len)
{
	if (len < 13)
		return 0;
	
	// check a nmdc command
	if (strncmp (data+2, TOKEN_VALIDATENICK, 11))
		return 0;

	char token[80], nick[80];
	
	sscanf (data, "%s %s", token, nick);

	// cut of last character '|' with nickname
	unsigned int l = len-15;

	if (l < 4 || l >= DEFAULT_MAX_NICKLENGTH) {
		//bot_to_user (u, "Your nick is corrupted !");
		return proto_nmdc_state_disconnect (u);
	}

	// assign nick to user
	memcpy (u->nick, nick, l);
	u->nick[l] = '\0';

	u->nick_len = l;

#ifdef DEBUG
	printf ("nick: '%s' : length: %d\n", u->nick, u->nick_len);
#endif

	// User is already on hub ?
	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
    	    if (!strcmp (u->nick, user->nick) && user->online && user != u) {
	       send_to_user_len (u, "$ValidateDenide|", 16);
	       u->state = PROTO_STATE_DISCONNECTED;
	       return 0;
	    }
 	}

	unsigned long bantime = 0;
	if ((bantime = ban_find_bynick (u->nick)) != 0) {
		char str[64];
		sprintf (str, "You are banned for %lu seconds !", bantime);

		bot_to_user (u, str);
		u->state = PROTO_STATE_DISCONNECTED;
		return 0;
	}

	// Send hub name and description to user	
	char str[128];
	sprintf (str, "$HubName %s - %s|", config_get ("hubname"), config_get ("hubdesc"));

	send_to_user (u, str);

	/* Is user registered ? */
	u->iclass = account_find (u->nick);
	
	// Assign rights
	if (u->iclass > 0) {
		u->state = PROTO_STATE_WAITPASS;
		send_to_user_len (u, "$GetPass|", 9);			// Ask for passwd
		return 1;
	}

	if (stats.currusers > (signed) config.userlimit) {
		char str[48];
		sprintf (str, "Hub is currently full ! (%d users)", config.userlimit);
		bot_to_user (u, str);

		proto_nmdc_user_redirect (u);
		return 0;
	}

	char b[64];
	sprintf (b, "$Hello %s|", u->nick);
	
	send_to_user (u, b);

	u->state = PROTO_STATE_HELLO;
	u->online = true;

	return 1;
}

/*************************************************************\
|		     NMDC STATE - WAITPASS		      |
\*************************************************************/

int proto_nmdc_state_waitpass (user_t *u, char *data, unsigned int len)
{
	if (len < 7)
		return 0;
  
	// check a nmdc command
	if (strncmp (data, TOKEN_MYPASS, 7))	
		return 0;
	
	char token[256], pass[128];

	sscanf (data, "%s %s", token, pass);

	int l = len-8;

	pass[l-1] = '\0';

	if (!account_check (u, pass)) {
		bot_to_user (u, "Bad password.");
		send_to_user (u, "$BadPass|");
		u->state = PROTO_STATE_DISCONNECTED;
		return 0;
	}

	char b[9 + DEFAULT_MAX_NICKLENGTH];
	sprintf (b, "$Hello %s|", u->nick);
	
	send_to_user (u, b);

	u->state = PROTO_STATE_HELLO;	
	u->online = true;

	return 1;
}

/*************************************************************\
|		      NMDC STATE - HELLO		      |
\*************************************************************/

int proto_nmdc_state_hello (user_t *u, char *data, unsigned int len)
{
	// check a nmdc command
	if (len >= 12)
	if (!strncmp (data, TOKEN_GETNICKLIST, 12))
		return proto_nmdc_token_getnicklist (u, data);

	/* myinfo parse */

	if (len <= 13+u->nick_len)
		return 0;
	
	// check a nmdc command
	if (strncmp (data, "$MyINFO", 7))
		return 0;

	if (strncmp (data+13, u->nick, u->nick_len))
		return 0;

	unsigned short d = 14, dn = 0;

	while (d != len) {
		if (data[d] == '$')
			dn ++;

		d ++;
	}
	
	if (dn != 5)
		return 0;

	unsigned int l = 14 + u->nick_len;

	char b[256];
	b[0] = '\0';

	/* parse desctag */
	char desc[DEFAULT_MAX_DESCLENGTH];
	desc[0] = '\0';
	
	int e = l;
	
	while (data[e] != '$' && e <= len)
		e ++;

	unsigned short dl = e-l;

	//if (dl >= DEFAULT_MAX_DESCLENGTH)			// FIXME tag + desc calculate is incorrect 
	//	dl = DEFAULT_MAX_DESCLENGTH - 1;
	if (dl > 1) {
		if (dl >= DEFAULT_MAX_DESCLENGTH)
			dl = DEFAULT_MAX_DESCLENGTH;
		
		memcpy (desc, data+l, dl);
		desc[dl] = '\0';
	}

	/* parse tag */
	unsigned int r = 0, tags = len, tage = 0;

	while (r != (unsigned) dl) {
		if (desc[r] == '<')
			tags = r;

		if (desc[r] == '>')
			tage = r;

		r ++;
	}

	if (tags < tage) {	// user have tag
		char *tag = desc+tags;

		char slots[16];
		// parse mode, hubs, slots
		r = 0;
		while (r != tage) {
			// mode
			if (tag[r] == 'M')
				if (tag[r+1] == ':')
					u->mode = tag[r+2];

			// slots
			if (tag[r] == 'S')
				if (tag[r+1] == ':') {
					memcpy (slots, tag+r+2, tage-tags-2-r);
					slots[tage-tags-2-r] = '\0';

					int islots = atoi (slots);
					if (islots < 0)
						islots = 0;

					u->slots = islots;

					if (islots < (signed) config.minslots[u->iclass]) {
						char str[64];
						sprintf (str, "Min. slots on this hub is %d, you have only %d !", config.minslots[u->iclass], islots);
						bot_to_user (u, str);

						proto_nmdc_user_redirect (u);
						return 0;
					}
			}

			r ++;
		}
	}
	else if (!class_right_check (u, CAP_TAG)) {
		bot_to_user (u, "This hub requests tag of client !");
		proto_nmdc_user_redirect (u);
		return 0;
	}

	/* parse connection type / speed */
	char conn[64];
	conn[0] = '\0';

	int t = e + 3;

	while (data[t] != '$')
		t ++;

	memcpy (conn, data+e+3, t-e-3);
	conn[t-e-3] = '\0';

	/* parse email */
	char email[DEFAULT_MAX_EMAILLENGTH];
	email[0] = '\0';

	int m = t + 1;

	while (data[m] != '$')
		m ++;

	if (m-t-1 < DEFAULT_MAX_EMAILLENGTH) {
		memcpy (email, data+t+1, m-t-1);
		email[m-t-1] = '\0';
	}

	/* parse share */
	char shared[64];
	shared[0] = '\0';

	int p = m + 1;

	while (data[p] != '$')
		p++;

	short mp = p-m-1;
	if (mp > 0 && (unsigned) p < len) {
		memcpy (shared, data+m+1, mp);
		shared[mp] = '\0';
		u->share = atoll (shared);
	}

	/* check min. share with user share */
	if (!class_right_check (u, CAP_SHARE))
	if (u->share < config.minshare[u->iclass]) {
		char str[80];
		sprintf (str, "Min. share is %s! Your share is only %s.",
			 share_tostring (config.minshare[u->iclass]), share_tostring (u->share));

		bot_to_user (u, str);

		proto_nmdc_user_redirect (u);
			
		return 0;
	}

	strncpy (u->myinfo, data, len);		// save data of myinfo

	u->myinfo_len = len;			// save length of myinfo
	
	memcpy (b, data, len);

	if (class_right_check (u, CAP_KEY)) {
		memcpy (b+len, "$OpList ", 8);
		memcpy (b+len+8, u->nick, u->nick_len);
		memcpy (b+len+8+u->nick_len, "$$|", 3);
		len += 11+u->nick_len;

		event_lua (u, data, EVENT_OPCONNECTED, len);
	}

	// File with motd must exist
	if (config.motd)
		bot_to_user_len (u, config.motd, config.motd_len);	/* send motd to new user */

	u->state = PROTO_STATE_ONLINE;
	
	// send new myinfo to all online (logged) clients
	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next)
		if (user->state == PROTO_STATE_ONLINE)
			send_to_user_len (user, b, len);

        stats.currusers ++;

	event_lua (u, data, EVENT_USERCONNECTED, len);
	event_plugin (u, data, EVENT_USERCONNECTED, len);

	return 1;
}

/*************************************************************\
|		   NMDC STATE - DISCONECTED		      |
\*************************************************************/

int proto_nmdc_state_disconnect (user_t *user)
{
	event_lua_hub (user, EVENT_QUIT);
	event_plugin (user, NULL, EVENT_QUIT, 0);

	unsigned int id = user->socket;

	if (user->state > PROTO_STATE_SENDLOCK)
		if (user->flag & NMDC_SUPPORTS_ZPipe)
			stats.cache_zpipers --;

	printf ("> disconnected -> %d", id);

	if (user->state >= PROTO_STATE_ONLINE && user->online) {
		if (user->nick_len > 1) {
			printf (" -> %s", user->nick);
	
			char str[64];	
			sprintf (str, "$Quit %s|", user->nick);
			
			send_to_all_len (str, 7+user->nick_len);
		}

		stats.currusers --;
	}

	putchar ('\n');

	if (user->online)
		user->online = false;

	user->socket = 0;		// user is disconnected

	if (user->state >= PROTO_STATE_ONLINE)
		if (user->buf)
			free (user->buf);
	
	user->buf_len = 0;
	user->buf_used = 0;
	user->buf_offset = 0;

 	user->next->prev = user->prev;
 	user->prev->next = user->next; 

	close (id);

	user->state = PROTO_STATE_OUT;

	free (user);
	
	stats.num_drop ++;

	return -1;
}

/*************************************************************\
|		     NMDC STATE - ONLINE		      |
\*************************************************************/

int proto_nmdc_state_online (user_t *u, char *data, unsigned int len)
{
	// check a nmdc command
	switch (data[0]) {
		case TOKEN_CHAT:
		{
			int p = 1;
			while (data[p] != '|')
				p ++;
	
			if (u->nick_len + 3 >= len)
				return 0;
			
			char *cmd = data + u->nick_len + 3;
	
			if (cmd[0] == '+' || cmd[0] == '!') {	// Command
				int e = 0;
				int m = 0;
				
				while (e != p) {
					if (!m)
					if (cmd[e] == ' ')
						m = e;
				  
					if (cmd[e] == '|')
						break;
					e ++;
				}	
	
				cmd[e] = '\0';

				event_lua (u, data, EVENT_COMMAND, len);
				
				char c[32];
				
				if (!m)
					m = e;
				
				if (m < 32) {
					memcpy (c, cmd, m);
					c[m] = '\0';
				}
				
				command_parser (u, c, data, len);
	
				stats.num_commands ++;
				
			} else {				// Chat message
				
				if (!class_right_check (u, CAP_CHAT))
					return 0;

				if (strncmp (data+1, u->nick, u->nick_len))
					return 0;
	
				if (len < config.chat_lenght) {
					/* Deflood */
					stats.def_mainchat ++;

					event_lua (u, data, EVENT_CHATIN, len);
					event_plugin (u, data, EVENT_CHATIN, len);

					if (!class_right_check (u, CAP_SPAM)) {
						if (config.deflood.mainchat_lock)
							return 0;

						if (user_flag_check (u, USER_FLAG_GAG))
							return 0;
	
						if (stats.def_mainchat > config.deflood.mainchat) {
							if (stats.def_mainchat == config.deflood.mainchat+1) {
								user_t *user;
								for (user = user_list.next; user != &user_list; user = user->next)
									bot_to_user (user, "** Flood detected **");
								
								ban_add_byip (u, "300");
	
								stats.def_mainchat += 2;
	
								u->state = PROTO_STATE_DISCONNECTED;
							}
		
							/* after flood we need turn on mainchat in quickly time */
							if (stats.def_mainchat > (config.deflood.mainchat*2))
								stats.def_mainchat = config.deflood.mainchat*2;
		
							return 0;
						}
					}
	#ifdef ZPIPE
					if (len > ZPIPE_THRESHOLD && stats.cache_zpipers > 0) {
						char *out = (char *) calloc (len + 65, sizeof (char));
				
						unsigned long l = zdef (data, len, out, len + 64);
			
						user_t *user;
						for (user = user_list.next; user != &user_list; user = user->next) {
							if (user->online) {
								if (user->flag & NMDC_SUPPORTS_ZPipe)
									send_to_user_len (user, (char *) out, l);
								else
									send_to_user_len (user, data, len);
							}
						}	
	
						free (out);	
					} else
	#endif
						send_to_all_len (data, len);				// Message

					event_lua (u, data, EVENT_CHAT, len);

					stats.num_chats ++;
				} else
					bot_to_user (u, "Your message is too long.");
			}
	
			return 1;
		
			break;
		}
		case '$':
		{
			switch (data[1]) 
			{
				case 'C':
					if (!class_right_check (u, CAP_DL)) {
						bot_to_user (u, "You are not allowed to downloading.");
						return 0;
					}

					if (!strncmp(data, TOKEN_CONNECTTOME, 12)) {
				
						char target[128], ip[128], tkn[64];
				
						sscanf (data, "%s %s %s", tkn, target, ip);
				
						user_t *user = NULL;
						user = user_findbynick (target);
				
						if (!user)	// user dont exists
							return 0;
				
						if (class_right_check (user, CAP_SHAREBLOCK)) {
							bot_to_user (u, "You can't download from this user.");
							return 0;
						}

						send_to_user_len (user, data, len);
				
						stats.num_connect ++;
				
						return 1;	
					}
					break;
				case 'T':
					if (!class_right_check (u, CAP_PM))
						return 0;

					if (!strncmp(data, TOKEN_TO, 4)) {
						if (len >= config.pm_lenght) {
							bot_to_user (u, "Private message is too long.");
							return 0;
						}
				
						unsigned int s = 0, x = 0;
						while (x <= len) {
							if (data[x] == '$')
								s ++;
				
							if (s == 2)
								break;
				
							x ++;
						}
				
						if (s != 2) {
							bot_to_user (u, "Invalid private message.");
							return 0;
						}
				
						char a[64], b[64], c[64], d[64], msg[config.pm_lenght+1];
				
						sscanf (data, "%s %s %s %s", a, b, c, d);		
				
						// b komu to mam poslat
						// d kdo to posila
				
						/* user is chatroom */
						chatroom_t *room = NULL;
						room = chatroom_find (b);
						if (room)
							return chatroom_msghandler (u, data, len, room);

						event_lua (u, data, EVENT_PM, len);

						if (u->iclass < 2)
							if (config.deflood.pm_lock)
								return 0;
				
						/* we can find user where is message sent */
						user_t *user = user_findbynick (b);
				
						/* user dont exists */
						if (!user) {
							bot_to_user (u, "User is offline.");
							return 0;
						}
				
						int l = strlen (a) + strlen (b) + strlen (c) + strlen (d) + u->nick_len + 8;
				
						msg[0] = '\0';
						memcpy (msg, data + l, len - l -1);		// parse text
						msg[len - l -1] = '\0';
				
						pm_to_user (user, d, msg);

				
						stats.num_pm ++;
				
						return 1;
					}	
					break;
				case 'S':
					if (!class_right_check (u, CAP_SEARCH))
						return 0;

					if (!strncmp(data, TOKEN_SEARCH, 7)) {
						long long int l = sec - u->searchlast;
						int search_interval = config.search_interval[u->iclass];
				
						if (l >= search_interval) {
							user_t *user;
							
							for (user = user_list.next; user != &user_list; user = user->next) {
								if (user->online && user->state == PROTO_STATE_ONLINE)
									send_to_user_len (user, data, len);
							}
				
							u->searchlast = sec;
				
							event_lua (u, data, EVENT_SEARCH, len);
				
							stats.num_search ++;
						} else {
						
							if (u->searchlast == 0) {
								send_to_all_len (data, len);
				
								u->searchlast = sec;
								return 1;
							}
				
							long long int w = search_interval - l;
				
							char str[64];
							sprintf (str, "Search is limited to %ds. Please wait %lld sec.",
								search_interval, w);
				
							bot_to_user (u, str);
						}
				
						return 1;
					}
					break;
				case 'O':
					if(!class_right_check (u, CAP_REDIRECT)) {
						bot_to_user (u, "You have'nt right for redirecting.");
						return 0;
					}

					if (!strncmp(data, TOKEN_OPFORCEMOVE, 12)) {
				
						if (u->iclass < 3)
							return 0;

						unsigned short i = 12, a = 0, b = 0;
						while (data[i] != '\0') {
							if (data[i] == '$') {
								if (!a)
									a = i;
								else if (!b)
									b = i;
								else
									break;
							}

							i ++;
						}

						char who[b-4];

						memcpy (who, data+a+5, b-6);
						who[b-a-5] = '\0';

						i = 15;
						a = 0; b = 0;
						while (data[i] != '\0') {
							if (data[i] == '$') {
								if (!a)
									a = i;
								else if (!b)
									b = i;
								else
									break;
							}

							i ++;
						}

						char where[b-4];

						memcpy (where, data+a+7, b-7);
						where[b-a-7] = '\0';
				
						int x = strlen (who);
						int y = strlen (where);
					
						if (x < 1 || y < 1)
							return 0;

						user_t *user = NULL;
						user = user_findbynick (who);
					
						if (!user)
							return 0;
					
						char msg[len-16];
						memcpy (msg, data + b + 5, len - b - 6);
						msg [len - b - 6] = '\0';
				
						char str[len-b+32];
						sprintf (str, "You have been moved to '%s' because: %s", where, msg);
				
						bot_to_user (user, str);
				
						proto_nmdc_user_forcemove (user, where);	
				
						return 1;
					}
					break;
				case 'G':
					if (!strncmp(data, TOKEN_GETNICKLIST, 12))
						return proto_nmdc_token_getnicklist (u, data);
					break;
				case 'B':
					if (!strncmp(data, TOKEN_BOTINFO, 8)) {
						
						/* $HubINFO <hub name>$<hub address:port>$<hub description>$<max users>$<min share in bytes>$<min slots>$<max hubs>$<hub type>$[<hubowner login>]| */
				
						char str[256];
						sprintf (str, "$HubINFO %s$%s:%d$%s$%u$%llu$%u$0$%s$|",
							config.hubname, config.hubaddress,
							config.port, config.hubdesc, config.userlimit,
							config.minshare[0], config.minslots[0], HUBSOFT);
				
						send_to_user (u, str);
				
						stats.botinfo ++;
				
						return 1;
					}
					break;
			}
			break;
		}
	}

	return 0;
}


/*************************************************************\
|		   NMDC PROTOCOL HANDLING		      |
\*************************************************************/

int proto_state_handler (user_t *u, char *data, unsigned int len)
{
	switch (u->state) {
		case PROTO_STATE_INIT:		// new user connected
		     return proto_nmdc_state_init (u);
		case PROTO_STATE_SENDLOCK:	// waiting for user $Key
		     return proto_nmdc_state_sendlock (u, data, len);
		case PROTO_STATE_WAITNICK:	// waiting for user $ValidateNick
		     return proto_nmdc_state_waitnick (u, data, len);
		case PROTO_STATE_WAITPASS:	// waiting for user $GetPass
	             return proto_nmdc_state_waitpass (u, data, len);
		case PROTO_STATE_HELLO:		// waiting for user $MyINFO
		     return proto_nmdc_state_hello (u, data, len);
		case PROTO_STATE_ONLINE:	// user is avaible now
		     return proto_nmdc_state_online (u, data, len);
		case PROTO_STATE_DISCONNECTED:	// user gone out    $Quit
		     return proto_nmdc_state_disconnect (u);
	}	

	return 1;
	
}


/*************************************************************\
|		  NMDC PROTOCOL CACHE FLUSH		      |
\*************************************************************/

int proto_flush_cache ()
{
	sec ++;			// uptime calculation

	/* deflood flush */
	if (stats.def_mainchat < 2)
		stats.def_mainchat = 0;
	else
		stats.def_mainchat --;

	/* cpu usage */
#ifndef __zexos__
	cpu_measure ();		// cpu calculation
#endif
	/* null traffic */
	stats.traffic_rx = 0;
	stats.traffic_tx = 0;

	/* Save config in interval 'autosave' */
	if (config.autosave)
		if ((sec - lastsave) >= config.autosave) {
			config_save (savefile);
			lastsave = sec;
		}

	/* Hublist update interval */
	if (config.hublist_interval)
		if ((sec - lastupdate) >= config.hublist_interval) {
			hublist_update (NULL);
			lastupdate = sec;
		}

#ifdef LUA
	/* Lua script timer */  				// FIXME only for one script
	if (timer.active)
		if (((unsigned) sec - timer.sec) >= timer.interval) {
			event_lua_sys (EVENT_TIMER);
			timer.sec = sec;
		}
#endif

	return 1;
}
