/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2009 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2010 ZeXx86 (zexx86@zexos.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#define FD_SETSIZE 				16384
#define MAX_FD					16384

#ifdef __WIN32__
 #include <windows.h>
 #include <wininet.h>
#else
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>
#endif
#include <string.h>
#include <stdio.h>

#include <stdlib.h>
#include <sys/types.h>

#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>

#include "defaults.h"
#include "user.h"
#include "utils.h"
#include "config.h"
#include "proto.h"


extern user_t user_list;
extern int bufsize;
extern stats_t stats;

#ifdef __WIN32__
WORD wVersionRequested = MAKEWORD(1,1); // WinSock versin
WSADATA data;           		// Info about WS
long long int boottime = 0;
long long int tlast = 0;
#else
struct timeval to, tnow, tnext, boottime;
#endif

/* global variables for select work */
#ifdef POLL
# undef SELECT
# undef EPOLL
# include <poll.h>
# define TIMEOUT 1000
# define CAPACITY_INCREMENT 60240
unsigned int cnt = 0;
unsigned int maxAlloc = 0, changed, p;
struct pollfd *ufds;
struct pollfd *temp;
#endif

#ifdef EPOLL
# undef SELECT
# undef POLL
# include <sys/epoll.h>
# define ESOCKET_MAX_FDS 	16384
# define ESOCKET_ASK_FDS 	2048
   struct epoll_event ev, events[ESOCKET_ASK_FDS];
   int n, nfds, kdpfd;
   struct timeval epoll_to;
#endif

#ifdef SELECT
# undef POLL
# undef EPOLL
# ifndef __WIN32__
#  include <sys/select.h>
# endif
int fdmax = 1;
fd_set socketRead;
fd_set socketWrite;
struct timeval tv;
#endif

char buf[BUFSIZE];            		// Recv buffer
struct sockaddr_in sockName;         		// "Jm�no" portu
struct sockaddr_in clientInfo;       		// Klient, kter� se pripojil 
SOCKET mainSocket;               	// Sokt
int size = 0;                 		// Pocet prijat�ch a odeslan�ch bytu

/************************************/
#ifndef __WIN32__
socklen_t addrlen;
#else
int addrlen;
#endif

extern int cont;			// we want continue in main thread ?
SOCKET server_add_port (int port);


/*************************************************************\
|		   INITIALIZATION OF SOCKET		      |
\*************************************************************/

int init_socket ()
{
	char *sPort = strdup (config_get ("listenport"));

#ifdef __WIN32__
	// P�iprav�me sokety na pr�ci
	if (WSAStartup(wVersionRequested, &data) != 0) {
		printf ("> ERROR -> I can't init sockets\n");
		return -1;
	}
#endif

	// Create mainsocket
	if ((mainSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		printf ("> ERROR -> I can't create main socket\n");
		return -1;
	}

#ifdef __WIN32__
	/* Set to nonblocking socket mode */
	u_long arg = 1;
	// Operation is  FIONBIO. Parameter is pointer on non-zero number.
	if (ioctlsocket(mainSocket, FIONBIO, &arg) == SOCKET_ERROR) {
		printf ("> ERROR -> Problem with set socket mode\n");
		return -1;
	}
#else
	/* Set to nonblocking socket mode */
	int oldFlag = fcntl(mainSocket, F_GETFL, 0);
	if (fcntl(mainSocket, F_SETFL, oldFlag | O_NONBLOCK) == -1) {
		printf ("> ERROR -> Problem with set socket mode\n");
		return -1;
	}

	setsockopt (mainSocket, SOL_SOCKET, SO_REUSEADDR, (char *) &bufsize, sizeof (bufsize));
#endif
	//Put data to structure sockaddr_in
	// 1) Family of protocol
	sockName.sin_family = AF_INET;
	// 2) Number of listen port
	sockName.sin_port = htons(atoi (sPort));
	// 3) Set local ip address. You can connect in with whatever IP
	sockName.sin_addr.s_addr = INADDR_ANY;
	// Name of socket
	if (bind(mainSocket, (struct sockaddr *)&sockName, sizeof(sockName)) == -1) {
		printf ("> ERROR -> Port is used or danied\n");
		free (sPort);
		cont = 0;
		return -1;
	}

	// Crete front (max 100 sessions in one time)
	if (listen(mainSocket, 100) == -1) {
		printf ("> ERROR -> I can't create front\n");
		cont = 0;
		return -1;
	}

#ifdef SELECT
	// Clean FD
	FD_ZERO(&socketRead);
	FD_ZERO(&socketWrite);	
	// Nastavim soket do mnozin - mnozina ma zatim jen jeden prvek.
	FD_SET(mainSocket, &socketRead);
	// Soket s maximalnim identifikacnim cislem je timpadem mainSocket.
	fdmax = mainSocket;
	// Nastavim cekaci dobu 3 minuty
	tv.tv_sec = 1;
	tv.tv_usec = 0;
#endif

#ifdef POLL
	if ((ufds = (struct ollfd *) realloc(ufds, CAPACITY_INCREMENT)) == NULL) {
		// Zavru hlavni soket
		close(mainSocket);
		printf ("> ERROR -> POLL -> out of memory\n");
		exit (1);
		return -1;
	}
#endif

#ifdef EPOLL

	epoll_to.tv_sec = 0;
	epoll_to.tv_usec = 1000;

	kdpfd = epoll_create (ESOCKET_MAX_FDS);
	
	if (!kdpfd)
		perror ("epoll_create");

	nfds = mainSocket;

	events[0].events = EPOLLIN;
	events[0].data.fd = nfds;
	
	if (epoll_ctl(kdpfd, EPOLL_CTL_ADD, nfds, events) < 0) {
		fprintf(stderr, "> ERROR -> EPOLL -> set insertion error: fd=%d\n", nfds);
		return -1;
	}
#endif

#ifndef __WIN32__ 
#ifndef __zexos__
	/* add lowest member of the statistices */
	gettimeofday (&boottime, NULL);
		
	/* initialize the random generator */
	srandom (boottime.tv_sec ^ boottime.tv_usec);
			
	gettimeofday (&tnext, NULL);
	tnext.tv_sec += 1;
#endif
#else
//	boottime
#endif

	printf ("> init () -> OK\n> started on port -> %s\n", sPort);

	free (sPort);

	//server_add_port (4112);	// ### In progress

	return 1;
}


SOCKET server_add_port (int port)
{
	SOCKET sock;
#ifdef __WIN32__
	// Prepare socket api
	if (WSAStartup (wVersionRequested, &data) != 0)  {
		printf ("> ERROR -> I can't init sockets\n");
		return -1;
	}
#endif
	// Create mainsocket
	if ((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		printf ("> ERROR -> I can't create main socket\n");
		return -1;
	}

#ifdef __WIN32__
	/* Set to nonblocking socket mode */
	u_long arg = 1;
	// Operation is  FIONBIO. Parameter is pointer on non-zero number.
	if (ioctlsocket(sock, FIONBIO, &arg) == SOCKET_ERROR)
	{
		printf ("> ERROR -> Problem with set socket mode\n");
		return -1;
	}
#else
	/* Set to nonblocking socket mode */
	int oldFlag = fcntl(sock, F_GETFL, 0);
	if (fcntl(sock, F_SETFL, oldFlag | O_NONBLOCK) == -1)
	{
		printf ("> ERROR -> Problem with set socket mode");
		return -1;
	}

	setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (char *) &bufsize, sizeof (bufsize));
#endif
	//Put data to structure sockaddr_in
	// 1) Family of protocol
	sockName.sin_family = AF_INET;
	// 2) Number of listen port
	sockName.sin_port = htons(port);
	// 3) Set local ip address. You can connect in with whatever IP
	sockName.sin_addr.s_addr = INADDR_ANY;
	// Name of socket
	if (bind (sock, (struct sockaddr *) &sockName, sizeof (sockName)) == -1) {
		printf ("> ERROR -> Port is used or danied");;
		cont = 0;
		return -1;
	}
	// Crete front (max 100 sessions in one time)
	if (listen(sock, 100) == -1) {
		printf ("> ERROR -> I can't create front");;
		cont = 0;
		return -1;
	}

	return sock;
}

/*************************************************************\
|		       ACCEPT NEW SOCKET		      |
\*************************************************************/

int accept_socket ()
{
#ifdef EPOLL
	addrlen = sizeof (clientInfo);

	int client = accept (mainSocket, (struct sockaddr *) &clientInfo, &addrlen);
	
	if (client < 0){
		perror("accept");
		return -1;
	}

	// Set client socket to nonblocking mode
	int oldFlag = fcntl (client, F_GETFL, 0);
	if (fcntl (client, F_SETFL, oldFlag | O_NONBLOCK) == -1) {
		printf ("**ERROR** Problem with set socket mode");;
		return 0;
	}

	ev.events = EPOLLIN | EPOLLOUT | EPOLLET;
	ev.data.fd = client;
	
	events[n].data.fd = client;

	if (epoll_ctl (kdpfd, EPOLL_CTL_ADD, client, &ev) < 0) {
		fprintf (stderr, "epoll set insertion error: fd=%d\n", client);
		return -1;
	}

	/* Set send buffer size */
	setsockopt (client, SOL_SOCKET, SO_SNDBUF, &bufsize, sizeof (bufsize));
		
	printf ("> new client -> %s -> %d\n", inet_ntoa((struct in_addr) clientInfo.sin_addr), client);
	
	return client;
#endif

#ifdef POLL
	if (cnt != sockets.size() + 1) {
		cnt = sockets.size() + 1;

		ufds[0].fd = mainSocket;
		ufds[0].events = POLLIN;
		ufds[0].revents = 0;
		for (i = sockets.begin(), p = 1; i != sockets.end (); i++, p ++) {
			ufds[p].fd = *i;
			ufds[p].events = POLLIN;
			ufds[p].events |= POLLOUT;
			ufds[p].revents = 0;
		}
	}

	if ((changed = poll (ufds, cnt, TIMEOUT)) < 0) {
		// Close listen socket
		close (mainSocket);
		// Free ufds structure
		free (ufds);
		printf ("Problem s poll");;
		return -1;
	}

	// Something changed on listen socket ?
	if (ufds[0].revents & POLLIN != 0) {
		socklen_t addrlen = sizeof (clientInfo);
		// Accept new connection
		int client = accept (mainSocket, (struct sockaddr *) &clientInfo, &addrlen);
	
		if (client < 0)
			return 0;
		
		sockets.push_back(client);

		// Set client socket to nonblocking mode
		int oldFlag = fcntl(client, F_GETFL, 0);
		if (fcntl (client, F_SETFL, oldFlag | O_NONBLOCK) == -1) {
			printf ("> ERROR -> Problem with set non-block socket mode -> %s\n", client);
			return 0;
		}

		/* Set send buffer size */
		setsockopt (client, SOL_SOCKET, SO_SNDBUF, &bufsize, sizeof (bufsize));

		printf ("> new client -> %s -> %d\n", inet_ntoa ((struct in_addr) clientInfo.sin_addr), client);
		
		return client;
	}

#else
#ifndef EPOLL
     addrlen = sizeof (clientInfo);

     /* accept new socket */
     SOCKET client = accept (mainSocket, (struct sockaddr *) &clientInfo, &addrlen);

#ifdef __WIN32__ 
     if(client != INVALID_SOCKET) {
#else
     if(client != -1 || errno != EAGAIN) {
#endif
	if (client < 0) {
	  	//printf ("> ERROR -> I can't accept socket -> %d\n", client);
	  	//close (client);
	  	return -1;
	}

#ifdef __WIN32__
	/* set socket to nonblocking mode */
	u_long arg = 1;
	// Operace je FIONBIO. Parametr je ukazatel na nenulov� ��slo.
	if (ioctlsocket (client, FIONBIO, &arg) == SOCKET_ERROR) {
		printf ("> ERROR -> Problem with set socket mode");;
		return -1;
	}
#else
	/* set socket to nonblocking mode */
	int oldFlag = fcntl(client, F_GETFL, 0);
	if (fcntl (client, F_SETFL, oldFlag | O_NONBLOCK) == -1) {
		printf ("> ERROR -> Problem with set non-block socket mode -> %d\n", client);
		return 0;
	}

     	setsockopt (client, SOL_SOCKET, SO_SNDBUF, &bufsize, sizeof (bufsize));
/*	int sock_buf_size = BDP;
	setsockopt( client, SOL_SOCKET, SO_SNDBUF, (char *)&sock_buf_size, sizeof(sock_buf_size) );
	setsockopt( client, SOL_SOCKET, SO_RCVBUF, (char *)&sock_buf_size, sizeof(sock_buf_size) );

	int flag = 1;
	setsockopt( client, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag) );*/

#endif
	printf ("> new client -> %s -> %d\n", inet_ntoa ((struct in_addr) clientInfo.sin_addr), client);

	return client;
     }
#endif
#endif

     return 0;
}


/*************************************************************\
|		   CREATE NEW STRUCT OF USER		      |
\*************************************************************/

extern long long int sec;
int user_connect ()
{
        SOCKET client = accept_socket ();

	if (client > 0) {
		// For communication is used socket "client"
		user_t *u;	
		
		// alloc and init context
		u = (user_t *) malloc (sizeof (user_t));

		if (!u) {
		  	printf ("user_connect () -> malloc () : ERROR\n");
			return 0;
		}

		u->state = PROTO_STATE_INIT;
#ifdef POLL
		u->socket = client;
		u->fd = p;
#else
		u->socket = client;
#endif
		u->iclass = 0;
		u->share = 0;
		u->searchlast = -3600;
		u->flag = 0;
		u->online = false;
		u->nick_len = 0;
		u->myinfo_len = 0;
		u->myinfo[0] = '\0';
		u->buf_offset = 0;
		u->buf_used = 0;
#ifndef __zexos__
		u->buf_len = 32;
		u->buf = (char *) malloc (u->buf_len + 1);

		if (!u->buf) {
			printf ("user_connect () -> malloc () : ERROR\n");
			free (u);
			return 0;
		}
#else
		u->buf = 0;
		u->buf_len = 0;
#endif
		u->nick[0] = '\0';

		unsigned char *ip = (unsigned char *) &clientInfo.sin_addr;
		
		sprintf (u->ip, "%u.%u.%u.%u", ip[0], ip[1], ip[2], ip[3]);

		// add into list  
		u->next = &user_list;
		u->prev = user_list.prev;
		u->prev->next = u;
		u->next->prev = u;

		return proto_state_handler (u, NULL, 0);
	}

	return 0;
}

/*************************************************************\
|		     	    L O O P			      |
\*************************************************************/

int r = 0;
int loop ()
{
     int ret = 0;

#ifndef POLL
#ifndef EPOLL

#ifndef SELECT
        user_connect ();
# ifndef __WIN32__
# ifndef __zexos__
        usleep (20);		// Its needed for socket function send ()
# endif
# endif
#else
     ret = select (fdmax+1, &socketRead, NULL, NULL, &tv);
#ifdef DEBUG
     if (ret < 0) 
	perror ("select :");
#endif
    // Anything change on mainSocket ?
    if (FD_ISSET (mainSocket, &socketRead))
	user_connect ();
# ifndef SELECT
#  ifdef __zexos__
     usleep (20);
#  endif
# endif
#endif
	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {

		int id = user->socket;
#ifdef SELECT
		if (FD_ISSET(id, &socketRead)) {		// Is possible send to socket ? (SELECT)
#endif

		buf[0] = '\0';
#ifdef __zexos__
		if ((size = recv(id, buf, BUFSIZE - 1, 0)) < 0) {
#else
		if ((size = recv(id, buf, BUFSIZE - 1, 0)) == 0) {
#endif
			// printf (strerror(errno) << ;
			// Prijato 0 bytu - druha strana uzavrela spojeni

			proto_nmdc_state_disconnect (user);

			// Ukoncim cyklus. Neni dobre pokracovat v prochazeni kontejneru, kdyz jsem vymazal prvek.
			break;
		}
		buf[size] = '\0';

		if (user->state == PROTO_STATE_DISCONNECTED) {
			if (!user->online) {
				strncpy (buf, "$Quit|", 6);
				size = 6;
			} else
				user->online = false;
		}

       /* MAIN HANDLERS */
       if (size)
      		if (parse_token (user, buf, size) == -1)
			break;

#ifdef SELECT
       }
#endif
		socket_buf (user);		// Send/Recv

		stats.traffic_rx += size;	// Traffic calculation
	}


#ifdef SELECT
	// Znovu vyplnim prvky mnoziny
	// Vyprazdnim mnozinu
	FD_ZERO (&socketRead);
	//FD_ZERO(&socketWrite);
	// Nastavim hlavni soket
	FD_SET (mainSocket, &socketRead);

	fdmax = 0;

	// Nastavim vsechny ostatni sokety.
	for (user = user_list.next; user != &user_list; user = user->next) {
		if (user->socket > (unsigned int) fdmax)
			fdmax = user->socket;

		FD_SET(user->socket, &socketRead);
		//FD_SET(user->socket, &socketWrite);
	}

	if (fdmax < mainSocket)
		fdmax = mainSocket;

	// Nastavim cekaci dobu 3 minuty
	tv.tv_sec = 1;
	tv.tv_usec = 0;
#endif


#endif
#endif

#ifdef	POLL		/** POLL **/
	user_connect ();
	usleep (20);
	user_t *u;
	// Projdu vsechna spojeni a zkontroluji, jestli na nich nedoslo k zmene

	for (u = user_list.next; u != &user_list; u = u->next) {
		if (ufds[u->fd].revents & POLLIN) {		// Is possible send to socket ? (POLL)
			char buf[BUFSIZE];
			int size;

			if ((size = recv (ufds[u->fd].fd, (void *)buf, BUFSIZE - 1, 0)) == 0) {	
				// Prijato 0 bytu - druha strana uzavrela spojeni
				proto_nmdc_state_disconnect (u);
				// close(ufds[p].fd);
				//sockets.erase(find(sockets.begin(), sockets.end(), ufds[u->fd].fd));		// FIXME crash
				// Odectu jeden prvek ze struktury pollfd
				//p --;
				// Ukoncim cyklus. Neni dobre pokracovat v prochazeni kontejneru, kdyz jsem vymazal prvek.
				break;
			} else {
				buf[size] = '\0';
		
				if (u->state == PROTO_STATE_DISCONNECTED) {
					strcpy (buf, "$Quit|");
					size = 6;
				}
						
				/* MAIN HANDLERS */
				if (size > 0)
					if (parse_token (u, buf, size) == -1)
						break;
			}
		}
	
		if (ufds[u->fd].revents & POLLOUT)	// Is possible write to socket ? (POLL)
			socket_buf (u);

		ufds[u->fd].revents = 0;
	}
#endif

#ifdef	EPOLL		/** Extended POLL **/
	uint32_t sock_events;
	nfds = epoll_wait(kdpfd, events, ESOCKET_ASK_FDS, 1000);

	for(n = 0; n < nfds; n++) {

		unsigned int id = events[n].data.fd;
		sock_events = events[n].events;

		if(id == mainSocket) {
			user_connect ();
			usleep (20);
		} else {
			user_t *user;
			for (user = user_list.next; user != &user_list; user = user->next)
				if (user->socket == id)
					break;

			if (sock_events & EPOLLIN) {
				buf[0] = '\0';
				if ((size = recv(id, buf, BUFSIZE - 1, 0)) == 0) {
					// printf (strerror(errno) << endl;
					// Prijato 0 bytu - druha strana uzavrela spojeni
					proto_nmdc_state_disconnect (user);
			
					//user->state = PROTO_STATE_DISCONNECTED;
					// Ukoncim cyklus. Neni dobre pokracovat v prochazeni kontejneru, kdyz jsem vymazal prvek.
					break;
				}
				buf[size] = '\0';
			}
				
			/* MAIN HANDLERS */
			if ( size )
				parse_token (user, buf, size);

				stats.traffic_rx += size;	// Traffic calculation
		}
	
			//if (sock_events & EPOLLOUT)	
	}

	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next) {
		if (user->state == PROTO_STATE_DISCONNECTED) {
			if (!user->online) {
				strncpy (buf, "$Quit|", 6);
				size = 6;
			} else
				user->online = false;
		}

		socket_buf (user);		// Send/Recv
	}

#endif

/* 		1sec period timer 		*/
#ifndef __WIN32__
	/* 1s periodic cache flush */
	gettimeofday (&tnow, NULL);
	if (timercmp (&tnow, &tnext, >=)) {
		tnext = tnow;
		tnext.tv_sec += 1;
# ifdef POLL
		tnext.tv_usec -= (TIMEOUT + 20);
# endif
# ifdef SELECT
		tnext.tv_usec -= 20;
# endif
		proto_flush_cache ();
	}
#else
	long long int tnow = GetTickCount () / 1000;

	if ((tnow - tlast) >= 1) {
		tlast = tnow;
		proto_flush_cache ();
	}
#endif

	return ret;
}
