/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2009 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2010 ZeXx86 (zexx86@zexos.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "trigger.h"
#include "defaults.h"
#include "commands.h"
#include "user.h"
#include "utils.h"
#include "config.h"

trigger_t trigger_list;

extern config_t config;

int trigger_add (char *name, int type, char *data)
{
	// alloc and init context
	trigger_t *ctx = (trigger_t*) malloc (sizeof (trigger_t));
	
	if (!ctx)
		return 0;
	
	ctx->data = (char *) malloc (sizeof (char) * strlen (data) + 1);
	
	if (!ctx->data) {
		free (ctx);
		return 0;
	}

	strcpy (ctx->name, name);
	strcpy (ctx->data, data);
	ctx->type = type;

	ctx->next = &trigger_list;
	ctx->prev = trigger_list.prev;
	ctx->prev->next = ctx;
	ctx->next->prev = ctx;	

	return 1;
}

int trigger_del (char *name)
{
	trigger_t *ctx;
	for (ctx = trigger_list.next; ctx != &trigger_list; ctx = ctx->next) {
		if (!strcmp(ctx->name, name))
		{
			ctx->next->prev = ctx->prev;
			ctx->prev->next = ctx->next;

			return 1;
		}
	}

	return 0;
}

/* command handlers */
bool handler_cmd (user_t *u, char *data)
{
	char name[128];
	char text[128];

	name[0] = '\0';
	text[0] = '\0';

	sscanf (data, "%s %s", name, text);

	trigger_t *t;
	for (t = trigger_list.next; t != &trigger_list; t = t->next) {
    	    if (!strcmp (name, t->name)) {

		int l = strlen (t->data);

		char str[10240];
		str[0] = '\0';

		memcpy (str, t->data + strlen (name) + 1, l - strlen (name) - 1);

		int x = l - strlen (name) - 1;

		str[x] = '\0';

		bot_to_user (u, str);
	    }
	}

	return 1;
}

bool handler_file (user_t *u, char *data)
{
	

	return 1;
}

bool handler_triggeradd (user_t *u, char *data)
{
	char a[64], b[64], c[64]; 
	a[0] = '\0'; b[0] = '\0'; c[0] = '\0';
	
	sscanf (data, "%s %s %s", a, b, c);

	int data_len = strlen (data);

	if (strlen(c) < 1 || strlen (b) < 1 || data_len < 1) {
		char use[128];
		sprintf (use, "Use:\n%s <name> <type> <data>\n\ntype:\n\t- motd\n\t- forbid\n\t- command <name>", a);

		bot_to_user (u, use);
		return 0;
	}

	if (!strcmp (c, "command")) {
		char text[data_len];
		text[0] = '\0';

		int l = 0, x = 0;
		while (x < 3) {
			l ++;

			if (data[l] == ' ')
				x ++;
		}
		x = 0;
		while (data[x] != '\0')
			x++;

		memcpy (text, data + l-4, x);
		data[x] = '\0';

		trigger_add (b, 3, text);

		command_register (b, "Information command", &handler_cmd , 0);

		bot_to_user (u, "OK");
	} 
	else if (!strcmp (c, "motd")) {
		char text[data_len];
		text[0] = '\0';

		int l = 0, x = 0;
		while (x < 2) {
			l ++;

			if (data[l] == ' ')
				x ++;
		}
		x = 0;
		while (data[x] != '\0')
			x++;

		memcpy (text, data + l-4, x);
		data[x] = '\0';

		trigger_add (b, 1, text);

		set_motd (text);

		strcpy (config.motd, text);

		bot_to_user (u, "OK");
	} 


	return 1;
}

bool handler_triggerdel (user_t *u, char *data)
{
	char a[64], b[64];
	a[0] = '\0'; b[0] = '\0';
	
	sscanf (data, "%s %s", a, b);

	if (strlen(b) < 1) {
		char use[80];

		sprintf (use, "Use:\n%s <name>", a);

		bot_to_user (u, use);
		return 0;
	}

	trigger_t *ctx;

	int x = 0;

	for (ctx = trigger_list.next; ctx != &trigger_list; ctx = ctx->next) {
		if (!strcmp(ctx->name, b)) {
			switch (ctx->type)
			{
				case 0:
					// zatim nic
					break;
				case 1:
					strcpy (config.motd, "0");
					break;
				case 2:
					if (!command_unregister (b))
						goto result;
					break;
			}

			x = trigger_del (b);
		}
	}

result:
	if (x) {
		char str[80];

		sprintf (str, "Trigger %s is succefully deleted.", b);

		bot_to_user (u, str);
	} else {
		char str[80];

		sprintf (str, "Trigger %s don't exists.", b);

		bot_to_user (u, str);
	}

	return 1;
}

bool handler_triggerlist (user_t *u, char *data)
{
	trigger_t *ctx;
	unsigned int i = 0;

	char str[1024];
	sprintf (str, "Trigger list:\n");

	for (ctx = trigger_list.next; ctx != &trigger_list; ctx = ctx->next) {
		strcat (str, "\nName: "); strcat (str, ctx->name); strcat (str, " Type: ");

		switch (ctx->type) {
			case 0:
				strcat (str, "forbid");
				break;
			case 1:
				strcat (str, "motd");
				break;
			case 2:
				strcat (str, "command");
				break;
		}

		i ++;
	}

	if (!i)
		strcat (str, "\n trigger not found");

	bot_to_user (u, str);	

	return 1;
}

int init_trigger ()
{
	trigger_list.next = &trigger_list;
	trigger_list.prev = &trigger_list;

	command_register ("triggeradd", "Add trigger to hub", &handler_triggeradd , 4);
	command_register ("triggerdel", "Delete trigger from hub", &handler_triggerdel , 4);
	command_register ("triggerlist", "List of triggers", &handler_triggerlist , 4);

	return 1;
}
