/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@gmail.com)
 *  (C) Copyright 2009 ZeXx86 (zexx86@gmail.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#ifdef __WIN32__
 #include <windows.h>
#endif
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "defaults.h"
#include "user.h"
#include "utils.h"
#include "config.h"
#include "proto.h"
#include "ban.h"
#include "user_rights.h"

/* create structures */
extern user_t user_list;
extern config_t config;
extern account_t accounts_list;


flag_t class_rights[] = {
	{ CAP_CHAT, "chat", "User can write messages to chat." },
	{ CAP_PM, "pm", "User can write private messages." },
	{ CAP_DL, "dl", "User can download." },
	{ CAP_SEARCH, "search", "Right for searching on hub." },
	{ CAP_SHARE, "share", "User dont need to share nothing." },
	{ CAP_SPAM, "spam", "User can spam." },
	{ CAP_TAG, "tag", "User dont need tag in client." },
	{ CAP_SHAREBLOCK, "shareblock", "Uploading is blocked for this class." },
	{ CAP_REDIRECT, "redirect", "User can redirect user to another hubs." },
	{ CAP_KEY, "key", "User is keymaster." },
	{ CAP_BAN, "ban", "User can ban and kick." },
	{ CAP_CONFIG, "config", "User can change configuration fo hub." },

	{ 0, 0, 0 }
};


/* play with rights */
int class_right_add (unsigned short iclass, unsigned long long flag)
{
	if (config.rights [iclass] & flag)
		return 0;

	config.rights [iclass] |= flag;

	return 1;
}

int class_right_del (unsigned short iclass, unsigned long long flag)
{
	if (config.rights [iclass] & flag) {
		config.rights [iclass] &= ~flag;
		return 1;
	}
	
	return 0;
}


int class_right_check (user_t *u, unsigned long long flag)
{
	if (config.rights [u->iclass] & flag)
		return 1;

	return 0;
}

/* user flags work */
int user_flag_add (user_t *u, unsigned long flag)
{
	if (u->flag & flag)
		return 0;

	u->flag |= flag;

	return 1;
}

int user_flag_del (user_t *u, unsigned long flag)
{
	if (u->flag & flag) {
		u->flag &= ~flag;
		return 1;
	}

	return 0;
}

int user_flag_check (user_t *u, unsigned long flag)
{
	if (u->flag & flag)
		return 1;

	return 0;
}

/* we are play with rights here :P */
bool handler_rightadd (user_t *u, char *data)
{
	int c = 0;
	char a[64], b[64];
	b[0] = '\0';
	
	sscanf (data, "%s %d %s", a, &c, b);

	if (strlen (b) < 1 && c <= 5 && c >= 0) {
		int l = strlen (a);

		if (l >= 60)
			return 0;

		char str[80];
		sprintf (str, "Use:\n%s <class> <flag>", a);

		bot_to_user (u, str);
		return 0;
	}

	if (c > u->iclass) {
		bot_to_user (u, "You are not allowed to edit rights in this class !");
		return 0;
	}
	int i;
	for (i = 0; class_rights[i].flag != 0; i++) {
		if (!strcmp (class_rights[i].name, b)) {
			if (class_right_add (c, class_rights[i].flag)) {
				char str[80];
				sprintf (str, "Right '%s' was added to class (%d).", b, c);
				bot_to_user (u, str);
				return 1;
			}

			char str[80];
			sprintf (str, "Right '%s' is already in class (%d).", b, c);
			bot_to_user (u, str);
			return 0;
		}
	}

	char str[80];
	sprintf (str, "Right '%s' does not exist.", b);
	bot_to_user (u, str);

	return 0;
}

bool handler_rightdel (user_t *u, char *data)
{
	int c = 0;
	char a[64], b[64];
	b[0] = '\0';
	
	sscanf (data, "%s %d %s", a, &c, b);

	if (strlen (b) < 1 && c <= 5 && c >= 0) {
		int l = strlen (a);

		if (l >= 60)
			return 0;

		char str[80];
		sprintf (str, "Use:\n%s <class> <flag>", a);

		bot_to_user (u, str);
		return 0;
	}

	if (c > u->iclass) {
		bot_to_user (u, "You are not allowed to edit rights in this class !");
		return 0;
	}
	
	int i;
	for (i = 0; class_rights[i].flag != 0; i++) {
		if (!strcmp (class_rights[i].name, b)) {
			if (class_right_del (c, class_rights[i].flag)) {
				char str[80];
				sprintf (str, "Right '%s' was deleted from class (%d).", b, c);
				bot_to_user (u, str);
				return 1;
			}

			char str[80];
			sprintf (str, "Right '%s' is'nt in class (%d).", b, c);
			bot_to_user (u, str);
			return 0;
		}
	}

	char str[80];
	sprintf (str, "Right '%s' does not exist.", b);
	bot_to_user (u, str);

	return 0;
}

bool handler_rightlist (user_t *u, char *data)
{
	char *str = (char *) malloc (sizeof (char) * 100); 

	if (!str)
		return 0;

	memcpy (str, "Rights of classes\nClass : Rights", 32);

	unsigned l = 32;

	int iclass = 0;
	while (iclass < 6) {
		char buf[128];
		sprintf (buf, "\n%d : ", iclass);
		int i;
		
		for (i = 0; class_rights[i].flag != 0; i ++) {
			if (config.rights[iclass] & class_rights[i].flag) {
				sprintfa (buf, "%s;", class_rights[i].name);
			}
		}

		unsigned m = strlen (buf);

		memcpy (str+l, buf, m);
		l += m;

		iclass ++;
	}	

	memcpy (str+l, "\n\nFlag : Description\n", 21);
	l += 21;

	unsigned a = 0, b = 0;
	int i;
	
	for (i = 0; class_rights[i].flag != 0; i++) {
		a = strlen (class_rights[i].name);
		b = strlen (class_rights[i].description);

		memcpy (str+l, class_rights[i].name, a);
		memcpy (str+l+a, " : ", 3);
		memcpy (str+l+a+3, class_rights[i].description, b);  
		str[l+a+b+3] = '\n';

		l += a+b+4;
	}

	bot_to_user_len (u, str, l);

	return 0;
}

/*************************************************************\
|		  INITIALIZATION OF COMMANDS		      |
\*************************************************************/

int init_rights ()
{
	config.rights [0] = CAP_GUEST;
	config.rights [1] = CAP_REG;
	config.rights [2] = CAP_VIP;
	config.rights [3] = CAP_OP;
	config.rights [4] = CAP_ADMIN;
	config.rights [5] = CAP_OWNER;

	command_register ("rightadd", "Add some right for class.", &handler_rightadd, 4);
	command_register ("rightdel", "Delete right from class.", &handler_rightdel, 4);
	command_register ("rightlist", "Displays all list of rights.", &handler_rightlist, 4);

	return 1;
}
