/*
 *  (C) Copyright 2006 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2009 ZeXx86 (zexx86@zexos.org)
 *  (C) Copyright 2010 ZeXx86 (zexx86@zexos.org)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#ifdef __WIN32__
 #include <windows.h>
 #include "proto.h"
#else
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
#endif
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "commands.h"
#include "user.h"
#include "utils.h"
#include "config.h"
#include "proto.h"

extern user_t user_list;
extern command_t command_list;
extern config_t config;
extern stats_t stats;

extern int proto_state_handler (user_t *u, char *data, unsigned int len);

char motdfile[] = DEFAULT_MOTDFILE;
unsigned int bufsize = 10240000;

/*************************************************************\
|		   	   UTILITIES		     	      |
\*************************************************************/

/** manipulation with user's send buffer **/
int user_buf_realloc (user_t *u, unsigned long len)
{
     if (len <= (u->buf_len - u->buf_used))
	return 1;
     
     u->buf = (char *) realloc ((void *) u->buf, u->buf_len+len+1);
	
     if (!u->buf) {
	printf ("user_buf_realloc () -> realloc () : ERROR\n");
	return 0;
     }
     
     u->buf_len += len;
#ifdef DEBUG
     printf ("> buffer reallocated to: %lu Bytes (%d)\n", u->buf_len, u->socket);
#endif
     return 1;
}

int user_buf_rebuild (user_t *u)
{
     free (u->buf);
     
     u->buf_len = 32;  
	
     u->buf = (char *) malloc (u->buf_len + 1);
	
     if (!u->buf) {
	printf ("user_buf_rebuild () -> malloc () : ERROR\n");
	return 0;
     }
     
#ifdef DEBUG
     printf ("> buffer rebuilded to: %lu Bytes (%d)\n", u->buf_len, u->socket);
#endif
     return 1;
}

/** send functions - delivers data to user **/
int send_to_socket (int id, char *txt)
{
     unsigned int size;
     unsigned int l = strlen (txt);

     if ((size = send (id, txt, l, 0)) != l) {
	printf ("> WARNING -> %s %d\n", strerror (errno), id);

	return 0;
     }

     return 1;
}

int send_to_user (user_t *u, char *txt)
{
     unsigned long len = strlen (txt);
#ifndef __zexos__
     if (!user_buf_realloc (u, len))
	return 0;
     
     memcpy (u->buf+u->buf_used, txt, len);

     u->buf_used += len;
#else
     send_to_user_len_direct (u, txt, len);
#endif
     return 1;
}

int send_to_user_len (user_t *u, char *txt, unsigned long len)
{
#ifndef __zexos__
     if (!user_buf_realloc (u, len))
	return 0;
     
     memcpy (u->buf+u->buf_used, txt, len);
     u->buf_used += len;
#else
     send_to_user_len_direct (u, txt, len);
#endif
     return 1;
}

int send_to_user_len_direct (user_t *u, char *txt, unsigned long len)
{
     unsigned int id = u->socket;

     unsigned long left = len;

     unsigned long size = 0, ret = 0;
     unsigned long n = 0;

     while (left) {	// loop until all data are out ;)	FIXME This take a moment
	if (ret > DEFAULT_SOCKET_RETTIME)
		break;

	size = send (id, txt+n, left, 0);

	n += size;
	left -= size;
	ret ++;
     }

     return 1;
}


int send_to_all (char *txt)
{	
     user_t *user;
     for (user = user_list.next; user != &user_list; user = user->next)
	  send_to_user (user, txt);

     return 1;
}

int send_to_all_len (char *txt, unsigned long len)
{
     user_t *user;
     for (user = user_list.next; user != &user_list; user = user->next)
	  send_to_user_len (user, txt, len);

     return 1;
}

int bot_to_user (user_t *u, char *txt)
{
	unsigned l = strlen (txt);
	unsigned m = strlen (config.botname);
	
	char *str = (char *) malloc (sizeof (char) * (m + l + 5));

	if (!str) {
		printf ("bot_to_user () -> malloc () : ERROR\n");
		return 0;
	}

	str[0] = '<';
	memcpy (str+1, config.botname, m);
	str[m+1] = '>';
	str[m+2] = ' ';
	memcpy (str+m+3, txt, l);
	str[m+l+3] = '|';

	send_to_user_len (u, (char *) str, m+l+4);

	free (str);

	return 1;
}

int bot_to_user_len (user_t *u, char *txt, unsigned l)
{
	unsigned m = strlen (config.botname);
  
	char *str = (char *) malloc (sizeof (char) * (m + l + 5));

	if (!str) {
		printf ("bot_to_user_len () -> malloc () : ERROR\n");
		return 0;
	}

	str[0] = '<';
	memcpy (str+1, config.botname, m);
	str[m+1] = '>';
	str[m+2] = ' ';
	memcpy (str+m+3, txt, l);
	str[m+l+3] = '|';

	send_to_user_len (u, (char *) str, m+l+4);

	free (str);

	return 1;
}

int bot_to_all (char *txt)
{
	unsigned l = strlen (txt);
	unsigned m = strlen (config.botname);
	
	char *str = (char *) malloc (sizeof (char) * (m + l + 5));

	if (!str) {
		printf ("bot_to_all () -> malloc () : ERROR\n");
		return 0;
	}

	str[0] = '<';
	memcpy (str+1, config.botname, m);
	str[m+1] = '>';
	str[m+2] = ' ';
	memcpy (str+m+3, txt, l);
	str[m+l+3] = '|';

    	user_t *user;
	for (user = user_list.next; user != &user_list; user = user->next)
		send_to_user_len (user, (char *) str, m+l+4);

	return 1;
}

int pm_to_user (user_t *target, char *nick, char *txt)
{
	/* we can use as second param. NULL, pm will be sent by bot */
	//if (!nick)
	//	nick = strdup (config.botname);

	unsigned long  l  = strlen (txt);
	unsigned short bl = strlen (nick);		
	unsigned short nl = target->nick_len;

	char *str = (char *) calloc (20+nl+bl+bl+l, sizeof (char));
	
	if (!str) {
	  printf ("pm_to_user () -> calloc () : ERROR\n");
	  return 0;
	}
	
	memcpy (str, "$To: ", 5);
	memcpy (str+5, target->nick, nl);
	memcpy (str+5+nl, " $From: ", 8);
	memcpy (str+13+nl, nick, bl);
 	memcpy (str+13+nl+bl, " $<", 3);
	memcpy (str+16+nl+bl, nick, bl);
	memcpy (str+16+nl+bl+bl, "> ", 2);
	memcpy (str+18+nl+bl+bl, txt, l);
	memcpy (str+18+nl+bl+bl+l, "|", 1);

	unsigned long len = 19+nl+bl+bl+l;

	str[len] = '\0';

	send_to_user_len (target, str, len);

	free (str);

	return 1;
}

/* handle buffering */
/*int socket_buffering (user_t *u)
{
     SOCKET id = u->socket;

     if (id < 1)
	return 0;

     unsigned int ret = 0;
     long size = 0;
     unsigned long left = u->buf_len;

     if (left > buffering)
	left = buffering;
     else
	u->buffered = false;

     if (left < 1)
	return 0;

     u->buf [left] = '\0';

     size = send(id, u->buf, left, 0);

     if(size == -1)
	if (socket_error (u, size, ret) == -1)
	    return -1;

     stats.traffic_tx += size;

     *u->buf += size;
     u->buf_len -= size;

     return 1;
}*/

/* handling socket errors */
int socket_error (user_t *u, long error, int ret)
{
#ifdef __WIN32__
	if (ret > 10) {
		u->state = PROTO_STATE_DISCONNECTED;
		proto_nmdc_state_disconnect (u);
		return -1; 
        }
	ret += 2;

#else
# ifdef DEBUG
	printf ("> ERROR -> socket_error () -> %s (%d)\n", strerror (errno), errno);
# endif
	if (errno == 9)		// Bad file descriptor
		return -1;

	if (errno == 32)	// Broken pipe
		return -1;

	if (errno == 11) {	// Resource temporarily unavailable
		/*usleep (1000);
		send_to_user_len_direct (u, "<ERROR> Hub is too busy !|", 26);
		u->state = PROTO_STATE_OUT;*/
# ifdef DEBUG
		printf ("BUFFERING -> %lu (%d)\n", u->buf_len, u->socket);	// priblizne na 275839 se to neda odesilat
# endif
		return -1;
	}

	if (errno == 104)
		return -1;

	if (errno == EAGAIN)
		return -1;
#endif
	return 1;
}

/* handling buffers of users */
/*int socket_buf (user_t *u)
{
     //if (u->buffered)
	//return socket_buffering (u);

     SOCKET id = u->socket;

     if (id < 1)
	return 0;

     unsigned int ret = 1;
     long size = 0;
     unsigned long n = 0;
     unsigned long left = u->buf_len;
     long buffering;
     unsigned long limit = 0;

     if (left < 1)
	return 0;

     u->buf [left] = '\0';

     while (u->buf_len)			// opakujeme dokud neodesleme vsechno ;)	FIXME This take a moment
     {
	//if (ret > DEFAULT_SOCKET_RETTIME)
	//	break;

	left = u->buf_len;
	buffering = config.bufferhard - limit;

	if (buffering < 128)
		buffering = 128;

        if (left > buffering)
	   	left = buffering;

	size = send(id, u->buf+n, left, 0);
	//printf ("**DATA: '%s'\n", u->buf+n);
	if(size == -1)
		if (socket_error (u, size, ret) == -1) {
			//u->buf += n;
			u->buf_len -= left;

			char *data = (char *) calloc (sizeof (char), u->buf_len+1);
			memcpy (data, u->buf+n, u->buf_len);
			send_to_user_len (u, data, u->buf_len);

			printf ("**CHYBA** DATA: '%s'\n", data);
			free (data);
			return -1;
		}

	limit = ret*config.buffersoft;

	n += size;	// posuneme ukazatel o delku dat, ktere jsme zvladli poslat
	u->buf_len -= size;

	ret ++;
	stats.traffic_tx += size;
     }

     memset (u->buf, 0, u->buf_len);
     u->buf_len = 0;

     return 1;
}*/

/* handling buffers of users */
int socket_buf (user_t *u)
{
#ifndef __zexos__
     SOCKET id = u->socket;

     if (id < 1)
	return 0;

     unsigned int ret = 0;
     long size = 0;
     unsigned long n = u->buf_offset;
     unsigned long left = u->buf_used - n;

     if (left < 1)
	return 0;
 
     size = send (id, u->buf+n, left, 0);

     if(size == -1)
	if (socket_error (u, size, ret) == -1)
		return -1;

     n += size;
     u->buf_offset += size;
     left -= size;
     ret ++;

     stats.traffic_tx += size;
     
     /* all user's data were sended, let's "clear" user buffer */
     if (u->buf_offset && u->buf_offset == u->buf_used) {
	u->buf_offset = 0;
	u->buf_used = 0;
	
	/* we want to save memory as much as possible,
	   so 1024 Bytes is max. possible send buffer length.
	   (used especially after $NickList) */
	if (u->buf_len > 1024)
		user_buf_rebuild (u);
     }
#endif
     return 1;
}

static char token_buf[BUFSIZE];
int parse_token (user_t *u, char *data, unsigned long int len)
{
	if (data[0] == TOKEN_CHAT)			// MAIN CHAT
		return proto_state_handler (u, data, len);

	long z = -1;
	long x;
	
	int ret = 1;
	
        for (x = 0; x < (signed) len; x++) {	// NMDC Command
		if (data[x] == '$' && x > z) {
			int y = x;
			
			while (data[y] != '|' && y <= (signed) len)
			         y++;

			int l = y-x+1;

			if (!l)
				return 0;

			char *token = (char *) &token_buf;
			
			if (l >= BUFSIZE)
				continue;
			
			memcpy (token, data + x, l);
			token[l] = '\0';

#ifdef DEBUG
			printf ("> x: %ld y: %d = %s   l: %lu\n", x, y, token, y-x);	// output of parse NMDC commands
#endif

			ret = proto_state_handler (u, token, l);

			z = y;
			
			/* user was disconnected */
			if (ret == -1)
				break;
		}
	}

	return ret;
}

int set_motd (char *data)
{
	FILE *fp;
	
	fp = fopen (motdfile, "w+");

	if (!fp)
		return 0;
	
	fprintf (fp, "%s", data);

  	fclose (fp);

	return 1;
}

int get_motd ()
{
	FILE *fp = fopen (motdfile, "r");

	if (!fp)
		return 0;

	fseek (fp, 0, SEEK_END);
	unsigned flen = ftell (fp);
	fseek (fp, 0, SEEK_SET);

	if (!flen)
		return 0;
	
	if (config.motd)
		free (config.motd);
	
	config.motd = (char *) malloc (sizeof (char) * flen + 1);
	
	if (!config.motd)
		return 0;
	
	size_t ret = fread (config.motd, flen, sizeof (char), fp);

	if (!ret)
		free (config.motd);
	
	fclose (fp);
	
	config.motd_len = flen;

	return 1;
}

#ifdef __WIN32__	// not implemented in Win32, we needs link a static library from cygwin
char *crypt (char *data, char *hash)
{
	return data;
}
#endif
#ifdef __zexos__	// same for zexos
char *crypt (char *data, char *hash)
{
	return data;
}
#endif
#ifdef __android__	// same for android
char *crypt (char *data, char *hash)
{
	return data;
}
#endif

const char *units[] = { "Bytes", "Kilobyte", "Megabyte", "Gigabyte", "Terabyte", "Petabyte" };

char *share_tostring (unsigned long long size)
{
  static char buf[128];
  int i = 0;
  double output;

  output = size;
  while (((size >> 10) > 0) && (i < 5)) {
    i++;
    output = size;
    size = size >> 10;
    output = output / 1024;
  };

  snprintf (buf, 128, "%f %s", output, units[i]);

  return buf;
}

char *time_tostring (unsigned long s)
{
  static char buffer[512];
  unsigned int weeks, days, hours, minutes, seconds, total;

  seconds = s % 60;
  minutes = s / 60;
  hours = minutes / 60;
  days = hours / 24;
  weeks = days / 7;
  minutes = minutes % 60;
  hours = hours % 24;
  days = days % 7;

  total = 0;

  if (weeks)
    total += snprintf (buffer, 512, "%u week%s, ", weeks, (weeks > 1) ? "s" : "");
  if (days)
    total += snprintf (buffer + total, 512 - total, "%u day%s, ", days, (days > 1) ? "s" : "");

  if ((hours || minutes || seconds) || (!(weeks || days)))
    total += snprintf (buffer + total, 512 - total, "%2uh %2umin %2usec", hours, minutes, seconds);

  return buffer;
}

/* function for calculate count of spaces in char * */
char arg[192];
char *strarg (short m, char *data, unsigned long int len)
{
	if (m == 0) {
		memcpy (arg, data, len);
		return arg;
	}

	int c = 0, f = 0, s = 0;

	while (c < (signed) len) {
		if (data [c] == ' ') {
			s ++;

			if (s == 1)	// where is end of first word ?
				f = c;
		}

		c ++;
	}

	if (m > s) {
		if (f < 192) {
			if (s != 0) {
				memcpy (arg, data, f);
				arg [f] = '\0';
			} else {
				memcpy (arg, data, len);
				arg [len] = '\0';
			}
			return arg;
		}
	} else
		return (char *) "0";
	
		return NULL;
}

/** BUFFERS **/
unsigned int bf_alloc (bf_buffer *buf, unsigned long len)
{
	buf    = (bf_buffer *) malloc (sizeof (bf_buffer));
	buf->s = (char *) calloc (len+1, sizeof (char));

	if (buf->s == NULL)
		return 0;

	buf->l = 0;

	return 1;
}

unsigned int bf_free (bf_buffer *buf)
{
	free (buf->s);
	free (buf);

	buf->l = 0;

	return 1;
}

unsigned int bf_printf (bf_buffer *buf, char *data, unsigned long len)
{
	if (!bf_alloc (buf, len))
		return 0;

	//if (buf->s == NULL)
		//return 0;

	memcpy (buf->s, data, len);

	buf->l = len;
	printf ("printf: %s - %lu\n", buf->s, buf->l);
	return 1;
}

unsigned int bf_strncat (bf_buffer *buf, char *data, unsigned long len)
{
	printf ("BUF.L je: %lu\n", buf->l);
	printf ("OKvstrncat\n");
	if (buf->s == NULL || len == 0)
		return 0;

	printf ("OKponull, %lu\n", buf->l);

	memcpy (buf->s+buf->l, data, len);

	buf->l += len;
	
	buf->s [buf->l] = '\0';

	return 1;
}

unsigned int bf_strncat_alloc (bf_buffer *buf, char *data, unsigned long len)
{
	buf    = (bf_buffer *) malloc (sizeof (bf_buffer));
	buf->s = (char *) realloc((void *)buf->s, buf->l+len+1);

	if (buf->s == NULL)
		return 0;

	memcpy (buf->s+buf->l, data, len);

	buf->l += len;
	
	buf->s [buf->l] = '\0';

	return 1;
}

/* Appends to the end of a string */
/* NO checking of buf size, so it MUST be big enough */
/* The string usually has to be zeroed before this can be used */
void sprintfa (char *buf, const char *format, ...)
{  
	if(format) {
		va_list args;
		va_start(args, format);
		vsprintf(buf + strlen(buf), format, args);
		va_end(args);
     	}
}

/* rebuild needed buffers */
void cache_rebuild ()
{
	config.port = atoi (config_get ("listenport"));	

	config.autosave = atoll (config_get ("autosave.interval"));		// parse autosave interval

	config.search_interval[0] = atoll (config_get ("search.interval.guest"));// parse search interval
	config.search_interval[1] = atoll (config_get ("search.interval.reg"));// parse search interval
	config.search_interval[2] = atoll (config_get ("search.interval.vip"));// parse search interval
	config.search_interval[3] = atoll (config_get ("search.interval.op"));// parse search interval

	config.userlimit = atoi (config_get ("userlimit"));			// parse user limit

	config.minslots[0] = atoi (config_get ("minslots.guest"));		// parse slots limit
	config.minslots[1] = atoi (config_get ("minslots.reg"));		// parse slots limit
	config.minslots[2] = atoi (config_get ("minslots.vip"));		// parse slots limit
	config.minslots[3] = atoi (config_get ("minslots.op"));			// parse slots limit

	config.minshare[0] = atoll (config_get ("minshare.guest"));		// parse share limit
	config.minshare[1] = atoll (config_get ("minshare.vip"));		// parse share limit
	config.minshare[2] = atoll (config_get ("minshare.reg"));		// parse share limit
	config.minshare[3] = atoll (config_get ("minshare.op"));		// parse share limit

	config.deflood.mainchat = atoll (config_get ("deflood.mainchat"));	// parse deflood mainchat limit

	config.pm_lenght = atoi (config_get ("len.pm"));			// parse max. private message length
	config.chat_lenght = atoi (config_get ("len.chat"));			// parse max. chat message length

	strcpy (config.botname, config_get ("bot.name"));
	config.botname_len = strlen (config.botname);

	strcpy (config.hubname, config_get ("hubname"));
	strcpy (config.hubdesc, config_get ("hubdesc"));

	sprintf (config.initstr, "$Lock EXTENDEDPROTOCOLZeXx86 Pk=%s|<%s> This hub is running version %s of %s %s created by ZeXx86|",
	   	       HUBSOFT, config.botname, VERSION, HUBSOFT, PLATFORM);

	unsigned int l = 60;
	while (l <= 150) {
		if (config.initstr[l] == '|')
			break;
		l ++;
	}
	
	config.initstr[l + 1] = '\0';
	config.initlen = l + 1;

	/* parse redirect hosts */
	char *lists, *lc;
	short i = 0;
	
	lists = strdup (config_get ("redirect"));
	
	if (!lists)
		goto more;

	for (lc = strtok (lists, ";, "); lc; lc = strtok (NULL, ";, ")) {
		int lc_len = 0;
		if (!(lc_len = strlen (lc)))
			continue;

		if (i < MAX_REDIRECTHOST) {
			if (config.redirect[i] && *config.redirect[i])
				free (config.redirect[i]);

			config.redirect[i] = strdup (lc);
			//memcpy (config.redirect[i], lc, lc_len);
			//config.redirect[i] [lc_len] = '\0';	
			//printf ("ve foru: %s\n", config.redirect[i]);
		} else
			break;

		i ++;
	}

	free (lists);
more:
	stats.redirect_hosts = i;
//	printf ("pocet redirect hostu je: %d\n", i);

        strcpy (config.hubaddress, config_get ("hubaddress"));

	config.hublist_hosts = strdup (config_get ("hublist.hosts"));
	config.hublist_interval = atoll (config_get ("hublist.interval"));
	
	config.bufferhard = atoll (config_get ("buffering.hardlimit"));
	config.buffersoft = atoll (config_get ("buffering.softlimit"));

      //bf_printf (config.hubname, config_get ("hubname"), strlen (config_get ("hubname")));		// parse hub name
      //bf_printf (config.hubdesc, config_get ("hubdesc"), strlen (config_get ("hubdesc")));		// parse hub description
}
